
package cz.cuni.mff.corrector;

import static cz.cuni.mff.corrector.corrector.Corrector.*;
import cz.cuni.mff.corrector.corrector.*;
import cz.cuni.mff.corrector.edit.Edit;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.repair.Repair;
import cz.cuni.mff.corrector.testing.AnalysisCalls;
import cz.cuni.mff.corrector.testing.AnalysisGraphs;
import cz.cuni.mff.corrector.testing.AnalysisScaling;
import cz.cuni.mff.corrector.testing.AnalysisTasks;
import cz.cuni.mff.corrector.testing.AnalysisTimes;
import cz.cuni.mff.corrector.tree.Tree;
import java.util.List;

/**
 * Entry point for the corrector implementation.
 *
 * @author Martin Svoboda
 */
public class Main {
    
    // -------------------------------------------------------------------------
    

    /**
     * @param args The command line arguments.
     */
    public static void main(String[] args) {
        
        // Printing corrector introduction
        System.out.println("Corrector version 2013-11-05");
        System.out.println();
        
        // Switching required operations
        if (
            (args.length == 5) && (args[0].equals("correction"))
        ) {
            
            // Executing required corrector
            System.out.println("Correction mode active\n");
            runCorrectionAlgorithm(args[1], args[2], args[3], args[4]);
        
        } else if ((args.length == 1) && (args[0].equals("profiling"))) {
            
            // Executing required profiler
            System.out.println("Profiling mode active\n");
            runProfilingAnalyses();
            
        } else {
            
            // Printing usage description
            System.out.println("Description mode active\n");
            runUsageDescription();
            
        }

    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Runs the selected correction algorithm.
     * @param configCode Code of the selected correction configuration.
     * @param documentFile File name of the XML document to be corrected.
     * @param grammarFile File name of the grammar to be used as a schema.
     * @param outputFile Output file name for exporting found repairs.
     */
    private static void runCorrectionAlgorithm(
        String configCode,
        String documentFile,
        String grammarFile,
        String outputFile
    ) {
        
        // Opening required XML document
        System.out.println("Processing XML document: " + documentFile);
        Tree dataTree = Tree.createFromXmlFile(documentFile);
        if (dataTree == null) {
            System.out.println("  Unable to load required document");
            return;
        }
        
        // Opening required grammar file
        System.out.println("Processing grammar file: " + grammarFile);
        Grammar treeGrammar = Grammar.createFromDataFile(grammarFile);
        if (treeGrammar == null) {
            System.out.println("  Unable to load required grammar");
            return;
        }
        
        // Parsing configuration items
        System.out.println("Parsing corrector configuration");
        if (!configCode.matches("^[A-Z]{3}\\-[A-Z0-9]{2}\\-[A-Z]{1}")) {
            System.out.println("  Unable to parse configuration code");
            return;
        }
        
        // Parsing algorithm strategy
        String algorithmCode = configCode.substring(0, 3);
        AlgorithmStrategy algorithmStrategy = null;
        if (algorithmCode.equals(AlgorithmStrategy.DEFAULT.codeName)) {
            algorithmStrategy = AlgorithmStrategy.DEFAULT;
        } else if (algorithmCode.equals(AlgorithmStrategy.EXPLORE.codeName)) {
            algorithmStrategy = AlgorithmStrategy.EXPLORE;
        } else if (algorithmCode.equals(AlgorithmStrategy.REFINE.codeName)) {
            algorithmStrategy = AlgorithmStrategy.REFINE;
        }
        if (algorithmStrategy == null) {
            System.out.println("  Unable to parse algorithm strategy");
            return;
        }
        
        // Parsing execution approach
        String approachCode = configCode.substring(4, 6);
        ExecutionApproach executionApproach = null;
        if (approachCode.equals(ExecutionApproach.NESTING_SINGLE.codeName)) {
            executionApproach = ExecutionApproach.NESTING_SINGLE;
        } else if (approachCode.equals(ExecutionApproach.INVOKING_SINGLE.codeName)) {
            executionApproach = ExecutionApproach.INVOKING_SINGLE;
        } else if (approachCode.equals(ExecutionApproach.INVOKING_MULTIPLE.codeName)) {
            executionApproach = ExecutionApproach.INVOKING_MULTIPLE;
        } else if (approachCode.equals(ExecutionApproach.SCHEDULING_SINGLE.codeName)) {
            executionApproach = ExecutionApproach.SCHEDULING_SINGLE;
        } else if (approachCode.equals(ExecutionApproach.SCHEDULING_MULTIPLE.codeName)) {
            executionApproach = ExecutionApproach.SCHEDULING_MULTIPLE;
        }
        if (executionApproach == null) {
            System.out.println("  Unable to parse execution approach");
            return;
        }
        
        // Parsing signatures mode
        String signaturesCode = configCode.substring(7, 8);
        SignatureHandling signatureHandling = null;
        if (signaturesCode.equals(SignatureHandling.DISABLED.codeName)) {
            signatureHandling = SignatureHandling.DISABLED;
        } else if (signaturesCode.equals(SignatureHandling.ENABLED.codeName)) {
            signatureHandling = SignatureHandling.ENABLED;
        }
        if (signatureHandling == null) {
            System.out.println("  Unable to parse signatures mode");
            return;
        }
        
        // Printing the configuration label
        System.out.println(
            "Recognised configuration: " +
            Corrector.composeAlgorithmCode(
                algorithmStrategy, executionApproach, signatureHandling
            )
        );
        
        // Creating corrector instance
        System.out.println("Starting correction routine");
        Corrector correctorInstance = Corrector.createCorrectorInstance(
            dataTree, treeGrammar,
            algorithmStrategy, executionApproach, signatureHandling,
            false, false
        );
        
        // Repairing provided data tree
        System.out.println("Correction routine finished");

        // Exporting computed repairs
        System.out.println("Exporting found repairs: " + outputFile);
        Repair treeRepair = correctorInstance.getTreeRepair();
        List<List<Edit>> treeFix = treeRepair.getIntentFix();
        treeRepair.exportIntentFix(outputFile);
        if (treeFix.isEmpty()) {
            System.out.println("Provided document is valid");
        } else {
            System.out.println(
                "Total number of repairs: " + treeFix.size()
            );
            System.out.println(
                "Total correction cost: " + treeRepair.getTotalCost()
            );
        }
        
    }
    
    
    /**
     * Runs all the predefined profiling analyses.
     */
    private static void runProfilingAnalyses() {
        
        AnalysisTasks aTasks = new AnalysisTasks();
        aTasks.generateDataTrees();
        aTasks.run();
        
        AnalysisCalls aCalls = new AnalysisCalls();
        aCalls.generateDataTrees();
        aCalls.run();
        
        AnalysisGraphs aGraphs = new AnalysisGraphs();
        aGraphs.generateDataTrees();
        aGraphs.run();
        
        AnalysisTimes aTimes = new AnalysisTimes();
        aTimes.generateDataTrees();
        aTimes.run();
        
        AnalysisScaling aScale = new AnalysisScaling();
        aScale.generateDataTrees();
        aScale.run();
        
    }
    
    
    /**
     * Prints the corrector usage description.
     */
    private static void runUsageDescription() {
        
        System.out.println(
            "Implemented modes:\n" +
            "  correction - finds corrections for a given XML document\n" +
            "  profiling - conducts predefined evaluation experiments"
        );
        System.out.println(
            "Expected arguments:\n" +
            "  correction algorithm document grammar output\n" +
            "  profiling"
        );
        System.out.println(
            "Correction arguments:\n" +
            "  algorithm:\n" +
            "    specification of the algorithm to be used\n" +
            "    follow the form strategy-approach-signatures\n" +
            "      algorithm strategies: DEF, EXP or RFN\n" +
            "      execution approaches: N1, I1, IN, S1 or SN\n" +
            "      signatures handling: D or E\n" +
            "    for example RFN-N1-E\n" +
            "  document:\n" +
            "    file name of the XML document to be processed\n" +
            "  grammar:\n" +
            "    file name of the grammar definition to be used\n" +
            "    standard DTD or XSD schemata are not accepted\n" +
            "    only proprietary grammar syntax is allowed\n" +
            "  output:\n" +
            "    output file name for exporting found repairs"
        );
        System.out.println(
            "Correction example\n" +
            "  correction RFN-N1-E document.xml grammar.txt repairs.txt"
        );
            
    }
    
    
    // -------------------------------------------------------------------------

}
