
package cz.cuni.mff.corrector.cache;

import cz.cuni.mff.corrector.grammar.Rule;
import cz.cuni.mff.corrector.intent.Intent;
import cz.cuni.mff.corrector.intent.IntentInsert;
import cz.cuni.mff.corrector.tree.Node;
import java.util.*;

/**
 * Hierarchical implementation of the store manager based on signature indexing.
 *
 * @param <Item> Type of object to be stored by the cache manager instance.
 * @author Martin Svoboda
 */
public class SignatureStoreNested<Item> implements Store<Item> {

    // -------------------------------------------------------------------------

    
    /**
     * Internal storage for correct intent.
     */
    private Item correctStorage;

    /**
     * Internal storage for insert intents.
     * Indexing: nonterminal, terminal, chain.
     */
    private Map<String, Map<String, Map<String, Item>>> insertStorage;

    /**
     * Internal storage for delete intents.
     * Indexing: node.
     */
    private Map<Node, Item> deleteStorage;

    /**
     * Internal storage for repair intents.
     * Indexing: node, nonterminal.
     */
    private Map<Node, Map<String, Item>> repairStorage;

    /**
     * Internal storage for rename intents.
     * Indexing: node, nonterminal, terminal.
     */
    private Map<Node, Map<String, Map<String, Item>>> renameStorage;


    // -------------------------------------------------------------------------


    /**
     * Creates a new and empty caching manager with initialized storages.
     */
    public SignatureStoreNested() {
        this.correctStorage =
            null;
        this.insertStorage =
            new HashMap<String, Map<String, Map<String, Item>>>();
        this.deleteStorage =
            new HashMap<Node, Item>();
        this.repairStorage =
            new HashMap<Node, Map<String, Item>>();
        this.renameStorage =
            new HashMap<Node, Map<String, Map<String, Item>>>();
    }


    // -------------------------------------------------------------------------


    public Item getStoredItem(Intent indexingIntent) {
        Rule rule = indexingIntent.getGrammarContext().getBaseRule();
        switch (indexingIntent.getIntentType()) {
            case CORRECT:
                return this.correctStorage;
            case INSERT:
                return fetchInsertSignature(
                    rule.getNonterminalSymbol(),
                    rule.getTerminalSymbol(),
                    ((IntentInsert)indexingIntent).
                        computeChainSignature()
                );
            case DELETE:
                return fetchDeleteSignature(
                    indexingIntent.getBaseNode()
                );
            case REPAIR:
                return fetchRepairSignature(
                    indexingIntent.getBaseNode(),
                    rule.getNonterminalSymbol()
                );
            case RENAME:
                return fetchRenameSignature(
                    indexingIntent.getBaseNode(),
                    rule.getNonterminalSymbol(),
                    rule.getTerminalSymbol()
                );
            default:
                throw new UnsupportedOperationException();
        }
    }

    
    public void setStoredItem(Intent indexingIntent, Item itemRef) {
        Rule rule = indexingIntent.getGrammarContext().getBaseRule();
        switch (indexingIntent.getIntentType()) {
            case CORRECT:
                this.correctStorage = itemRef;
                break;
            case INSERT:
                storeInsertSignature(
                    itemRef,
                    rule.getNonterminalSymbol(),
                    rule.getTerminalSymbol(),
                    ((IntentInsert)indexingIntent).
                        computeChainSignature()
                );
                break;
            case DELETE:
                storeDeleteSignature(
                    itemRef,
                    indexingIntent.getBaseNode()
                );
                break;
            case REPAIR:
                storeRepairSignature(
                    itemRef,
                    indexingIntent.getBaseNode(),
                    rule.getNonterminalSymbol()
                );
                break;
            case RENAME:
                storeRenameSignature(
                    itemRef,
                    indexingIntent.getBaseNode(),
                    rule.getNonterminalSymbol(),
                    rule.getTerminalSymbol()
                );
                break;
        }
    }


    public void removeStoredItem(Intent indexingIntent) {
        Rule rule = indexingIntent.getGrammarContext().getBaseRule();
        switch (indexingIntent.getIntentType()) {
            case CORRECT:
                this.correctStorage = null;
                break;
            case INSERT:
                removeInsertSignature(
                    rule.getNonterminalSymbol(),
                    rule.getTerminalSymbol(),
                    ((IntentInsert)indexingIntent).
                        computeChainSignature()
                );
                break;
            case DELETE:
                removeDeleteSignature(
                    indexingIntent.getBaseNode()
                );
                break;
            case REPAIR:
                removeRepairSignature(
                    indexingIntent.getBaseNode(),
                    rule.getNonterminalSymbol()
                );
                break;
            case RENAME:
                removeRenameSignature(
                    indexingIntent.getBaseNode(),
                    rule.getNonterminalSymbol(),
                    rule.getTerminalSymbol()
                );
                break;
        }
    }


    public Collection<Item> getStoredItems(Intent.Type intentType) {
        switch (intentType) {
            case CORRECT:
                return getCorrectItems();
            case INSERT:
                return getInsertItems();
            case DELETE:
                return getDeleteItems();
            case REPAIR:
                return getRepairItems();
            case RENAME:
                return getRenameItems();
            default:
                throw new UnsupportedOperationException();
        }
    }


    // -------------------------------------------------------------------------


    /**
     * Attempts to fetch the stored item under the insert intent signature.
     * @param nonterminal Indexing nonterminal symbol.
     * @param terminal Indexing terminal symbol.
     * @param chain Serialized contexts chain identifying intent.
     * @return Fetched item reference or <code>null</code> reference.
     */
    private Item fetchInsertSignature(
        String nonterminal,
        String terminal,
        String chain
    ) {
        Map<String, Map<String, Item>> stMap =
            this.insertStorage.get(nonterminal);
        if (stMap == null) {
            return null;
        }
        Map<String, Item> ndMap = stMap.get(terminal);
        if (ndMap == null) {
            return null;
        }
        return ndMap.get(chain);
    }

    /**
     * Attempts to fetch the stored item under the delete intent signature.
     * @param node Indexing base node reference.
     * @return Fetched item reference or <code>null</code> reference.
     */
    private Item fetchDeleteSignature(
        Node node
    ) {
        return this.deleteStorage.get(node);
    }

    /**
     * Attempts to fetch the stored item under the repair intent signature.
     * @param node Indexing base node reference.
     * @param nonterminal Indexing nonterminal symbol.
     * @return Fetched item reference or <code>null</code> reference.
     */
    private Item fetchRepairSignature(
        Node node,
        String nonterminal
    ) {
        Map<String, Item> stMap = this.repairStorage.get(node);
        if (stMap == null) {
            return null;
        }
        return stMap.get(nonterminal);
    }

    /**
     * Attempts to fetch the stored item under the rename intent signature.
     * @param node Indexing base node reference.
     * @param nonterminal Indexing nonterminal symbol.
     * @param terminal Indexing terminal symbol.
     * @return Fetched item reference or <code>null</code> reference.
     */
    private Item fetchRenameSignature(
        Node node,
        String nonterminal,
        String terminal
    ) {
        Map<String, Map<String, Item>> stMap = this.renameStorage.get(node);
        if (stMap == null) {
            return null;
        }
        Map<String, Item> ndMap = stMap.get(nonterminal);
        if (ndMap == null) {
            return null;
        }
        return ndMap.get(terminal);
    }


    // -------------------------------------------------------------------------


    /**
     * Stores the provided item under the insert intent signature.
     * @param item Reference to the object to be stored.
     * @param nonterminal Indexing nonterminal symbol.
     * @param terminal Indexing terminal symbol.
     * @param chain Serialized contexts chain identifying intent.
     */
    private void storeInsertSignature(
        Item item,
        String nonterminal,
        String terminal,
        String chain
    ) {
        Map<String, Map<String, Item>> stMap =
            this.insertStorage.get(nonterminal);
        if (stMap == null) {
            stMap = new HashMap<String, Map<String, Item>>();
            this.insertStorage.put(nonterminal, stMap);
        }
        Map<String, Item> ndMap = stMap.get(terminal);
        if (ndMap == null) {
            ndMap = new HashMap<String, Item>();
            stMap.put(terminal, ndMap);
        }
        ndMap.put(chain, item);
    }

    /**
     * Stores the provided item under the delete intent signature.
     * @param item Reference to the object to be stored.
     * @param node Indexing base node reference.
     */
    private void storeDeleteSignature(
        Item item,
        Node node
    ) {
        this.deleteStorage.put(node, item);
    }

    /**
     * Stores the provided item under the repair intent signature.
     * @param item Reference to the object to be stored.
     * @param node Indexing base node reference.
     * @param nonterminal Indexing nonterminal symbol.
     */
    private void storeRepairSignature(
        Item item,
        Node node,
        String nonterminal
    ) {
        Map<String, Item> stMap = this.repairStorage.get(node);
        if (stMap == null) {
            stMap = new HashMap<String, Item>();
            this.repairStorage.put(node, stMap);
        }
        stMap.put(nonterminal, item);
    }

    /**
     * Stores the provided item under the rename intent signature.
     * @param item Reference to the object to be stored.
     * @param node Indexing base node reference.
     * @param nonterminal Indexing nonterminal symbol.
     * @param terminal Indexing terminal symbol.
     */
    private void storeRenameSignature(
        Item item,
        Node node,
        String nonterminal,
        String terminal
    ) {
        Map<String, Map<String, Item>> stMap = this.renameStorage.get(node);
        if (stMap == null) {
            stMap = new HashMap<String, Map<String, Item>>();
            this.renameStorage.put(node, stMap);
        }
        Map<String, Item> ndMap = stMap.get(nonterminal);
        if (ndMap == null) {
            ndMap = new HashMap<String, Item>();
            stMap.put(nonterminal, ndMap);
        }
        ndMap.put(terminal, item);
    }

    
    // -------------------------------------------------------------------------


    /**
     * Removes the cached item indexed by the insert intent signature.
     * @param nonterminal Indexing nonterminal symbol.
     * @param terminal Indexing terminal symbol.
     * @param chain Serialized contexts chain identifying intent.
     */
    private void removeInsertSignature(
        String nonterminal,
        String terminal,
        String chain
    ) {
        Map<String, Map<String, Item>> stMap =
            this.insertStorage.get(nonterminal);
        if (stMap == null) {
            return;
        }
        Map<String, Item> ndMap = stMap.get(terminal);
        if (ndMap == null) {
            return;
        }
        ndMap.remove(chain);
    }

    /**
     * Removes the cached item indexed by the delete intent signature.
     * @param node Indexing base node reference.
     */
    private void removeDeleteSignature(
        Node node
    ) {
        this.deleteStorage.remove(node);
    }

    /**
     * Removes the cached item indexed by the repair intent signature.
     * @param node Indexing base node reference.
     * @param nonterminal Indexing nonterminal symbol.
     */
    private void removeRepairSignature(
        Node node,
        String nonterminal
    ) {
        Map<String, Item> stMap = this.repairStorage.get(node);
        if (stMap == null) {
            return;
        }
        stMap.remove(nonterminal);
    }

    /**
     * Removes the cached item indexed by the rename intent signature.
     * @param node Indexing base node reference.
     * @param nonterminal Indexing nonterminal symbol.
     * @param terminal Indexing terminal symbol.
     */
    private void removeRenameSignature(
        Node node,
        String nonterminal,
        String terminal
    ) {
        Map<String, Map<String, Item>> stMap = this.renameStorage.get(node);
        if (stMap == null) {
            return;
        }
        Map<String, Item> ndMap = stMap.get(nonterminal);
        if (ndMap == null) {
            return;
        }
        ndMap.remove(terminal);
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the list with all stored items indexed by correct signatures.
     * @return List container with all stored items.
     */
    private List<Item> getCorrectItems() {

        // Creating and output container
        List<Item> outputList = new ArrayList<Item>();

        // Inserting the item reference
        if (this.correctStorage != null) {
            outputList.add(this.correctStorage);
        }

        // Returning the list of found items
        return outputList;

    }

    /**
     * Returns the list with all stored items indexed by insert signatures.
     * @return List container with all stored items.
     */
    private List<Item> getInsertItems() {

        // Creating and output container
        List<Item> outputList = new ArrayList<Item>();

        // Iterating over all cached items
        for (
            Map<String, Map<String, Item>> stMap :
            this.insertStorage.values()
        ) {
            for (Map<String, Item> ndMap : stMap.values()) {
                outputList.addAll(ndMap.values());
            }
        }

        // Returning the list of found items
        return outputList;

    }

    /**
     * Returns the list with all stored items indexed by delete signatures.
     * @return List container with all stored items.
     */
    private List<Item> getDeleteItems() {

        // Creating and output container
        List<Item> outputList = new ArrayList<Item>();

        // Iterating over all cached items
        outputList.addAll(this.deleteStorage.values());

        // Returning the list of found items
        return outputList;

    }

    /**
     * Returns the list with all stored items indexed by repair signatures.
     * @return List container with all stored items.
     */
    private List<Item> getRepairItems() {

        // Creating and output container
        List<Item> outputList = new ArrayList<Item>();

        // Iterating over all cached items
        for (Map<String, Item> stMap : this.repairStorage.values()) {
            outputList.addAll(stMap.values());
        }

        // Returning the list of found items
        return outputList;

    }

    /**
     * Returns the list with all stored items indexed by rename signatures.
     * @return List container with all stored items.
     */
    private List<Item> getRenameItems() {

        // Creating and output container
        List<Item> outputList = new ArrayList<Item>();

        // Iterating over all cached items
        for (
            Map<String, Map<String, Item>> stMap :
            this.renameStorage.values()
        ) {
            for (Map<String, Item> ndMap : stMap.values()) {
                outputList.addAll(ndMap.values());
            }
        }

        // Returning the list of found items
        return outputList;

    }


    // -------------------------------------------------------------------------

}
