
package cz.cuni.mff.corrector.corrector;

import cz.cuni.mff.corrector.cache.SignatureStoreFlat;
import cz.cuni.mff.corrector.cache.Store;
import cz.cuni.mff.corrector.execution.Pool;
import cz.cuni.mff.corrector.execution.Quota;
import cz.cuni.mff.corrector.execution.Task;
import cz.cuni.mff.corrector.execution.Wrapper;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.intent.Intent;
import cz.cuni.mff.corrector.intent.IntentCorrect;
import cz.cuni.mff.corrector.repair.Repair;
import cz.cuni.mff.corrector.testing.Report;
import cz.cuni.mff.corrector.testing.Timer;
import cz.cuni.mff.corrector.tree.Node;
import cz.cuni.mff.corrector.tree.Tree;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * Implementation of the universal corrector.
 *
 * @author Martin Svoboda
 */
public class Corrector {

    // -------------------------------------------------------------------------
    
    
    /**
     * Enumeration of all implemented correction algorithm strategies.
     */
    public enum AlgorithmStrategy {
        
        /**
         * Algorithm strategy: default (direct construction).
         */
        DEFAULT("DEF", "Default"),
        /**
         * Algorithm strategy: explore (dynamic exploration).
         */
        EXPLORE("EXP", "Explore"),
        /**
         * Algorithm strategy: refine (continuous refinement).
         */
        REFINE("RFN", "Refine");
        
        /** Code name of the algorithm strategy. */
        public final String codeName;
        /** Full name of the algorithm strategy. */
        public final String fullName;
        
        /**
         * Creates an enumeration item of an algorithm strategy.
         * @param codeName Algorithm strategy code name.
         * @param fullName Algorithm strategy full name.
         */
        AlgorithmStrategy(String codeName, String fullName) {
            this.codeName = codeName;
            this.fullName = fullName;
        }
        
    }
    
    
    /**
     * Enumeration of all implemented execution approaches.
     * Describes possible ways of nested requests invocation and scheduling.
     */
    public enum ExecutionApproach {
        
        /**
         * Execution approach: direct synchronous nesting, only 1 request.
         */
        NESTING_SINGLE("N1", "Single Recursive Nesting"),
        /**
         * Execution approach: direct workers invocation, only 1 request.
         */
        INVOKING_SINGLE("I1", "Single Worker Invocation"),
        /**
         * Execution approach: direct workers invocation, multiple requests.
         */
        INVOKING_MULTIPLE("IN", "Multiple Workers Invocation"),
        /**
         * Execution approach: indirect tasks scheduling, only 1 request.
         */
        SCHEDULING_SINGLE("S1", "Single Task Scheduling"),
        /**
         * Execution approach: indirect tasks scheduling, multiple requests.
         */
        SCHEDULING_MULTIPLE("SN", "Multiple Tasks Scheduling");
        
        /** Code name of the execution approach. */
        public final String codeName;
        /** Full name of the execution approach. */
        public final String fullName;
        
        /**
         * Creates an enumeration item of an execution approach.
         * @param codeName Execution approach code name.
         * @param fullName Execution approach full name.
         */
        ExecutionApproach(String codeName, String fullName) {
            this.codeName = codeName;
            this.fullName = fullName;
        }
        
    }
    
    
    /**
     * Enumeration of all implemented signature handling behaviours.
     */
    public enum SignatureHandling {
        
        /**
         * Signature handling: enabled (repeated repair evaluation avoided).
         */
        ENABLED("E", "Enabled Signatures"),
        /**
         * Signature handling: disabled (all intents are always evaluated).
         */
        DISABLED("D", "Disabled Signatures");
        
        /** Code name of the signature handling mode. */
        public final String codeName;
        /** Full name of the signature handling mode. */
        public final String fullName;
        
        /**
         * Creates an enumeration item of a signature handling mode.
         * @param codeName Signature handling mode code name.
         * @param fullName Signature handling mode name.
         */
        SignatureHandling(String codeName, String fullName) {
            this.codeName = codeName;
            this.fullName = fullName;
        }
        
    }
    
    
    /**
     * String separator used when composing algorithm code
     */
    private static final String algorithmCodeSeparator = "-";


    // -------------------------------------------------------------------------
    
    
    /**
     * Selected algorithm strategy.
     */
    private final AlgorithmStrategy algorithmStrategy;
    
    /**
     * Selected execution approach.
     */
    private final ExecutionApproach executionApproach;
    
    /**
     * Selected signature handling mode.
     */
    private final SignatureHandling signatureHandling;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Data tree object representing the input XML document.
     */
    protected Tree dataTree;

    /**
     * Single type tree grammar object reference.
     */
    protected Grammar treeGrammar;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Set of selected intent types allowed in the correction.
     * The presence of Intent.Type.CORRECT is ignored.
     */
    protected EnumSet<Intent.Type> enabledIntents;
    
    
    /**
     * Default choice of intent types allowed in the correction.
     */
    public static EnumSet<Intent.Type> defaultIntents = EnumSet.of(
            Intent.Type.CORRECT, 
            Intent.Type.INSERT, Intent.Type.DELETE,
            Intent.Type.REPAIR, Intent.Type.RENAME
        );
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Reference to the repair of the entire data tree.
     */
    protected Repair treeRepair;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Instance of the pool with workers used for the execution of handlers.
     * 
     * This pool is only created and used in case of the following approaches:
     *   INVOKING_SINGLE, INVOKING_MULTIPLE,
     *   SCHEDULING_SINGLE and SCHEDULING_MULTIPLE.
     * 
     * Note that the access to this pool needs not to be synchronised.
     */
    protected Pool workersPool;
    
    
    /**
     * Internal store for managing computed repairs by their signatures.
     * 
     * This store is only created and used if signature handling is enabled!
     * @see SignatureHandling#ENABLED
     * 
     * Indexing in this container is always based on intent signatures.
     * @see SignatureStoreFlat
     * 
     * Note that the access to this container needs not to be synchronised.
     * It is because we are not in fact able to always avoid conflicts,
     * so we need to be aware of their possible existence.
     */
    protected Store<Repair> repairsCache;
    
    
    /**
     * Internal store for managing existing tasks.
     * 
     * This cache is used when multiple requests can be initiated at the same
     * time and we need to avoid concurrency conflicts based on initiating
     * execution of the same intent signatures.
     * 
     * Tasks caching is only relevant in the following execution approaches:
     *   INVOKING_MULTIPLE and SCHEDULING_MULTIPLE
     * and at the same time only when signatures handling is enabled!
     * 
     * As a consequence, this indexing of this container is always based
     * on intent signatures rather than on intent themselves.
     * @see SignatureStoreFlat
     * 
     * Note that the access to this container needs not to be synchronised.
     * It is because we are not in fact able to always avoid conflicts,
     * so we need to be aware of their possible existence.
     */
    protected Store<Task> tasksCache;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Flag of activated profiling of execution times.
     * @see Report
     */
    protected final boolean profilingTimes;
    
    /**
     * Flag of activated profiling of various statistics.
     * @see Report
     */
    protected final boolean profilingStats;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Optional reference to the prepared profiling results.
     * @see #profilingTimes
     * @see #profilingStats
     */
    protected final Report profilingReport;
    
    
    // -------------------------------------------------------------------------


    /**
     * Internal constructor for the universal corrector implementation.
     * @see #createCorrectorInstance
     * @param dataTree Data tree to be corrected.
     * @param treeGrammar Single type tree grammar.
     * @param algorithmStrategy Selected correction algorithm strategy.
     * @param executionApproach Selected execution approach.
     * @param signatureHandling Selected signature handling mode.
     * @param profilingTimes Flag of activated execution time profiling.
     * @param profilingStats Flag of activated statistics profiling.
     */
    protected Corrector(
        Tree dataTree,
        Grammar treeGrammar,
        AlgorithmStrategy algorithmStrategy,
        ExecutionApproach executionApproach,
        SignatureHandling signatureHandling,
        boolean profilingTimes,
        boolean profilingStats
    ) {
        
        // Storing input parameters
        this.dataTree = dataTree;
        this.treeGrammar = treeGrammar;
        
        // Storing behaviour selections
        this.algorithmStrategy = algorithmStrategy;
        this.executionApproach = executionApproach;
        this.signatureHandling = signatureHandling;
        
        // Setting allowed intent types
        this.enabledIntents = Corrector.defaultIntents;
        
        // Initialising repairs store
        this.repairsCache = null;
        if (hasSignaturesEnabled()) {
            this.repairsCache = new SignatureStoreFlat<Repair>();
        }
        
        // Initialising workers pool
        this.workersPool = null;
        if (hasWorkersEnabled()) {
            this.workersPool = new Pool(Pool.POOL_UNLIMITED);
        }
        
        // Initialising tasks cache
        this.tasksCache = null;
        if (hasTasksCaching()) {
            this.tasksCache = new SignatureStoreFlat<Task>();
        }
        
        // Initialising profiling data
        this.profilingTimes = profilingTimes;
        this.profilingStats = profilingStats;
        if (profilingTimes || profilingStats) {
            this.profilingReport = new Report();
        } else {
            this.profilingReport = null;
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prepares and executes the required correction algorithm.
     * @param dataTree Data tree to be corrected.
     * @param treeGrammar Single type tree grammar.
     * @param algorithmStrategy Selected correction algorithm strategy.
     * @param executionApproach Selected execution approach.
     * @param signatureHandling Selected signature handling mode.
     * @param profilingTimes Flag of activated execution time profiling.
     * @param profilingStats Flag of activated statistics profiling.
     * @return Created corrector instance with correction results.
     * 
     * Note that is it not a good idea to activate both execution times
     * and other statistics profiling since the profiling report and all
     * its items have to be always accessed with achieved synchronization,
     * so recording full statistics cause the entire execution becomes slow.
     */
    public static Corrector createCorrectorInstance(
        Tree dataTree,
        Grammar treeGrammar,
        AlgorithmStrategy algorithmStrategy,
        ExecutionApproach executionApproach,
        SignatureHandling signatureHandling,
        boolean profilingTimes,
        boolean profilingStats
    ) {
        
        // Resetting the intent identifiers
        Intent.resetIdProvider();
        
        // Creating a new corrector instance
        Corrector correctorInstance =
            new Corrector(
                dataTree, treeGrammar,
                algorithmStrategy, executionApproach, signatureHandling,
                profilingTimes, profilingStats
            );
        
        // Executing the correction algorithm
        correctorInstance.repairDataTree();
        
        // Returning the corrector reference
        return correctorInstance;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Entry point of the entire data tree correction.
     * This method should not be called directly, see the corrector factory.
     * @see #createCorrectorInstance
     */
    private void repairDataTree() {
        
        // Starting the execution timer
        Timer localTimer = new Timer();
        localTimer.startProfilingTimer();
        
        // Creating the starting intent
        IntentCorrect startingIntent = createStartingIntent();
        
        // Performing data tree correction
        switch (getExecutionApproach()) {
            
            case NESTING_SINGLE:
                this.treeRepair = performNestingCorrection(startingIntent);
                break;
            
            case INVOKING_SINGLE:
            case INVOKING_MULTIPLE:
                this.treeRepair = performInvokingCorrection(startingIntent);
                break;
            
            case SCHEDULING_SINGLE:
            case SCHEDULING_MULTIPLE:
                this.treeRepair = performSchedulingCorrection(startingIntent);
                break;
            
            default:
                this.treeRepair = null;
                throw new UnsupportedOperationException();
            
        }
        
        // Stopping the execution timer
        localTimer.stopProfilingTimer();
        if (this.profilingTimes) {
            this.profilingReport.setExecutionTime(
                localTimer.getElapsedMilis()
            );
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the reference to the profiling report container.
     * Note that a valid report is only returned when the corresponding
     * profiling was activated during the corrector instantiation.
     * @return Computed repair structure or <code>null</code> reference.
     * @see #profilingTimes
     * @see #profilingStats
     */
    public Report getProfilingReport() {
        return this.profilingReport;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Performs the data tree correction according to the nesting mode.
     * @param startingIntent Reference to the starting correction intent.
     * @return Intent repair structure for the entire data tree.
     */
    private Repair performNestingCorrection(
        Intent startingIntent
    ) {
        
        // Creating the starting task
        Task startingTask = createSimpleTask(startingIntent);
        
        // Creating the handler instance
        Handler startingHandler = createHandlerInstance(startingTask);
        
        // Passing execution parameters
        if (startingHandler.hasStrategy(AlgorithmStrategy.REFINE)) {
            startingTask.setQuotaData(Quota.createUnlimitedQuota());
        }
        
        // Executing the entire correction
        startingHandler.run();
        
        // Fetching the computed repair
        Repair primaryRepair = startingTask.getIntentRepair();
        
        // Returning the repair reference
        return primaryRepair;
        
    }
    
    
    /**
     * Performs the data tree correction according to the invoking mode.
     * @param startingIntent Reference to the starting correction intent.
     * @return Intent repair structure for the entire data tree.
     */
    private Repair performInvokingCorrection(
        Intent startingIntent
    ) {
        
        // Creating the starting task
        Task startingTask;
        if (hasMultipleApproach()) {
            // Creating an extended task without dependencies
            startingTask = createExtendedTask(startingIntent);
        } else {
            // Creating a simple task with dependencies
            startingTask = createSimpleTask(startingIntent);
        }
        
        // Creating the handler instance
        Handler startingHandler = createHandlerInstance(startingTask);
        
        // Passing execution parameters
        if (startingHandler.hasStrategy(AlgorithmStrategy.REFINE)) {
            startingTask.setQuotaData(Quota.createUnlimitedQuota());
        }
        
        // The starting intent has to be invoked directly!
        
        // Executing the entire correction
        startingHandler.run();
        
        // Fetching the computed repair
        Repair primaryRepair = startingTask.getIntentRepair();
        
        // Returning the repair reference
        return primaryRepair;
        
    }
    
    
    /**
     * Performs the data tree correction according to the scheduling mode.
     * @param startingIntent Reference to the starting correction intent.
     * @return Intent repair structure for the entire data tree.
     */
    private Repair performSchedulingCorrection(
        Intent startingIntent
    ) {
        
        // Creating the starting task
        Task startingTask;
        if (hasMultipleApproach()) {
            // Creating an extended task without dependencies
            startingTask = createExtendedTask(startingIntent);
        } else {
            // Creating a simple task with dependencies
            startingTask = createSimpleTask(startingIntent);
        }
        
        // Creating the handler instance
        Handler startingHandler = createHandlerInstance(startingTask);
        
        // Passing scheduling parameters
        startingTask.setSchedulingMaster(Boolean.TRUE);
        startingTask.setRequestingTask(null);
        
        // Passing iterative parameters
        if (startingHandler.hasStrategy(AlgorithmStrategy.REFINE)) {
            startingTask.setQuotaData(Quota.createUnlimitedQuota());
        }
        
        // The starting intent must not to be invoked directly!
        
        // Starting the entire correction
        requestHandlerExecution(startingHandler);
        
        // The previous call causes that the entire correction
        // is executed by worker threads from now on
        
        // Waiting for the correction end
        synchronized (this) {
            try {
                // Avoiding spurious but possible wakeups
                while (!startingTask.hasCompletedRapair()) {
                    this.wait();
                }
            } catch (InterruptedException ex) {
                throw new Error(
                    "Waiting for the scheduled execution interrupted!"
                );
            }
        }
        
        // Fetching the computed repair
        Repair primaryRepair = startingTask.getIntentRepair();
        
        // Returning the repair reference
        return primaryRepair;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates the starting correction intent for the provided data tree.
     * @return Constructed starting correction intent structure.
     */
    private IntentCorrect createStartingIntent() {
        
        // Creating a sequence with the root node
        Node rootNode = this.dataTree.getRootNode();
        List<Node> rootNodes = new ArrayList<Node>(1);
        if (rootNode != null) {
            rootNodes.add(rootNode);
        }

        // Creating the starting correction intent
        IntentCorrect startingIntent =
            new IntentCorrect(rootNodes, this.treeGrammar.getStartingContext());
        
        // Returning the created starting intent
        return startingIntent;
        
    }
    
    
    /**
     * Creates a new instance of the correction handler.
     * @param associatedTask Task object to be associated with the handler.
     * @return Instance of the particular handler according to the strategy.
     */
    protected Handler createHandlerInstance(Task associatedTask) {
        
        // Determining the handler type
        switch (getAlgorithmStrategy()) {
            case DEFAULT:
            case EXPLORE:
            case REFINE:
                return new Handler(this, associatedTask);
            default:
                throw new UnsupportedOperationException();
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the selected algorithm strategy.
     * @return Corresponding algorithm strategy enumeration item.
     */
    public final AlgorithmStrategy getAlgorithmStrategy() {
        return this.algorithmStrategy;
    }
    
    /**
     * Returns the selected execution approach.
     * @return Corresponding execution approach enumeration item.
     */
    public final ExecutionApproach getExecutionApproach() {
        return this.executionApproach;
    }
    
    /**
     * Returns the selected signature handling mode.
     * @return Corresponding signature handling enumeration item.
     */
    public final SignatureHandling getSignatureHandling() {
        return this.signatureHandling;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Tests whether the given algorithm strategy is activated.
     * @param selectedStrategy Algorithm strategy to be tested.
     * @return Returns <code>true</code> if the strategy corresponds.
     */
    public boolean hasStrategy(
        Corrector.AlgorithmStrategy selectedStrategy
    ) {
        return (this.algorithmStrategy == selectedStrategy);
    }
    
    /**
     * Tests whether the given execution approach is activated.
     * @param selectedApproach Execution approach to be tested.
     * @return Returns <code>true</code> if the approach corresponds.
     */
    public boolean hasApproach(
        Corrector.ExecutionApproach selectedApproach
    ) {
        return (this.executionApproach == selectedApproach);
    }
    
    /**
     * Tests whether the given signatures handling mode is activated.
     * @param selectedMode Signature handling mode to be tested.
     * @return Returns <code>true</code> if the mode corresponds.
     */
    public boolean hasSignatures(
        Corrector.SignatureHandling selectedMode
    ) {
        return (this.signatureHandling == selectedMode);
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the composed algorithm code.
     * @return Code composed from the strategy, execution and signature modes.
     */
    public final String getAlgorithmCode() {
        return getAlgorithmStrategy().codeName+
            (Corrector.algorithmCodeSeparator) +
            getExecutionApproach().codeName +
            (Corrector.algorithmCodeSeparator) +
            getSignatureHandling().codeName;
    }
    
    /**
     * Composes the algorithm code.
     * @param algorithmStrategy Selected correction algorithm strategy.
     * @param executionApproach Selected execution approach.
     * @param signatureHandling Selected signature handling mode.
     * @return Code composed from the strategy, execution and signature modes.
     */
    public static String composeAlgorithmCode(
        AlgorithmStrategy algorithmStrategy,
        ExecutionApproach executionApproach,
        SignatureHandling signatureHandling
    ) {
        return algorithmStrategy.codeName +
            (Corrector.algorithmCodeSeparator) +
            executionApproach.codeName +
            (Corrector.algorithmCodeSeparator) +
            signatureHandling.codeName;
    }
    
    
    // -------------------------------------------------------------------------


    /**
     * Returns the reference to the repair of the entire data tree.
     * @return Tree repair structure reference.
     */
    public Repair getTreeRepair() {
        return this.treeRepair;
    }
    
    
    // ------------------------------------hasWorkersEnabledAlso--------
    
    
    /**
     * Detects whether the worker pool is required by the corrector.
     * The pool is required by the following execution approaches:
     *   INVOKING_SINGLE and INVOKING_MULTIPLE,
     *   SCHEDULING_SINGLE and SCHEDULING_MULTIPLE.
     * @return Returns <code>true</code> when workers pool is used.
     */
    public final boolean hasWorkersEnabled() {
        return hasInvokingApproach() || hasSchedulingApproach();
    }
    
    
    /**
     * Executes the given prepared correction handler.
     * 
     * This method can only be used when workers pool is enabled!
     * @see #hasWorkersEnabled
     * Also note that this method does not need to be synchronised, since
     * it just invokes the execution but is not responsible to decide
     * whether this execution should or should not be requested.
     * 
     * @param requestedHandler Correction handler instance to be executed.
     */
    public void requestHandlerExecution(Handler requestedHandler) {
        this.workersPool.executeRequestedHandler(requestedHandler);
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Detects whether the signature handling is enabled by the corrector.
     * @return Returns <code>true</code> when signatures are enabled.
     */
    public final boolean hasSignaturesEnabled() {
        return (this.signatureHandling == SignatureHandling.ENABLED);
    }
    
    
    /**
     * Attempts to fetch the required intent repair from the store.
     * 
     * This method can only be used when signature handling is enabled!
     * @see #hasSignaturesEnabled
     * 
     * This method is not synchronised, so it can only be used in approaches
     * that do not suffer from concurrency problems. In particular, only the
     * following execution approaches are suitable for direct calls:
     *   NESTING_SINGLE, INVOKING_SINGLE, and SCHEDULING_SINGLE.
     * Otherwise, more complex and synchronised operations that are able
     * to fetch repairs and tasks depending on their state of existence
     * should be used.
     * 
     * @param requiredIntent Intent for which the repair should be fetched.
     * @return Found repair structure or <code>null</code> on failure.
     */
    public Repair fetchStoredRepair(Intent requiredIntent) {
        return this.repairsCache.getStoredItem(requiredIntent);
    }
    
    /**
     * Inserts the provided computed repair into the repairs cache.
     * 
     * This method can only be used when signature handling is enabled!
     * @see #hasSignaturesEnabled
     * Also note that this method does not need to be synchronised, since
     * repeated insertions cannot be avoided and do not cause difficulties.
     * 
     * @param computedRepair Newly computed repair to be registered.
     */
    public synchronized void storeComputedRepair(Repair computedRepair) {
        this.repairsCache.setStoredItem(
            computedRepair.getCorrectionIntent(),
            computedRepair
        );
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Detects whether tasks need to be used to avoid concurrency conflicts.
     * 
     * Tasks caching is only relevant in the following execution approaches:
     *   INVOKING_MULTIPLE and SCHEDULING_MULTIPLE
     * and at the same time only when signatures handling is enabled!
     * 
     * In other words, we need to cache all existing and not yet fully
     * evaluated tasks by the corrector and when we would like to initiate
     * a new request, we first need to attempt to fetch the corresponding
     * task from the cache and only if it not exists we can create a new one.
     * 
     * Apparently, we do not need to take care (follow this behaviour) when
     * only one request can be initiated at a time and when signature handling
     * is not enabled (because this means that each and every new request
     * is unique and must be really executed and fully evaluated).
     * 
     * @return Returns <code>true</code> when tasks caching is enabled.
     */
    public final boolean hasTasksCaching() {
        return hasMultipleApproach() && hasSignaturesEnabled();
    }
    
    
    /**
     * Attempts to fetch the required task from the store.
     * 
     * This method can only be used when tasks caching is enabled!
     * @see #hasTasksCaching
     * 
     * This method cannot be used directly, since synchronisation over
     * the more complex operation over repairs and tasks is required!
     * 
     * @param requiredIntent Intent for which the repair should be fetched.
     * @return Found repair structure or <code>null</code> on failure.
     */
    private Task fetchCachedTask(Intent requestedIntent) {
        return this.tasksCache.getStoredItem(requestedIntent);
    }
    
    /**
     * Inserts the provided task instance into the tasks cache.
     * 
     * This method can only be used when tasks caching is enabled!
     * @see #hasTasksCaching
     * 
     * This method cannot be used directly, since synchronisation over
     * the more complex operation over repairs and tasks is required!
     * 
     * @param computedRepair Newly computed repair to be registered.
     */
    private void storeCreatedTask(Task createdTask) {
        this.tasksCache.setStoredItem(
            createdTask.getCorrectionIntent(),
            createdTask
        );
    }
    
    /**
     * Removes the provided task from the cache of tasks.
     * 
     * This method can only be used when tasks caching is enabled!
     * @see #hasTasksCaching
     * 
     * @param cachedTask Task to be removed from the tasks cache.
     */
    public void removeCachedTask(Task cachedTask) {
        this.tasksCache.removeStoredItem(cachedTask.getCorrectionIntent());
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new instance of a simple task.
     * Signature profiling is taken into account when enabled.
     * @param correctionIntent Associated correction intent reference.
     * @return Newly created and initialised task instance.
     */
    public Task createSimpleTask(Intent correctionIntent) {
        
        // Profiling signature statistics
        if (this.profilingStats == true) {
            this.profilingReport.increaseTaskCounters(
                correctionIntent.getIntentType()
            );
        }
        
        // Creating a new task instance
        return Task.createSimpleTask(correctionIntent);
        
    }
    
    /**
     * Creates a new instance of an extended task.
     * Signature profiling is taken into account when enabled.
     * @param correctionIntent Associated correction intent reference.
     * @return Newly created and initialised task instance.
     */
    public Task createExtendedTask(Intent correctionIntent) {
        
        // Profiling signature statistics
        if (this.profilingStats == true) {
            this.profilingReport.increaseTaskCounters(
                correctionIntent.getIntentType()
            );
        }
        
        // Creating a new task instance
        return Task.createExtendedTask(correctionIntent);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Fetches the required repair or creates a new simple task.
     * Only suitable for NESTING_SINGLE and INVOKING_SINGLE approaches.
     * As a consequence, no synchronisation is required to be assured.
     * @param requestedIntent Reference to the requested intent.
     * @return Wrapper with the repair or task references.
     */
    protected Wrapper provideSimpleWrapper(
        Intent requestedIntent
    ) {
        
        // Fetching already existing repair
        Repair nestedRepair = null;
        if (hasSignaturesEnabled()) {
            nestedRepair = fetchStoredRepair(requestedIntent);
        }
        
        // Returning the fetched repair
        if (nestedRepair != null) {
            return new Wrapper(nestedRepair);
        }
        
        // Creating the nested task
        return new Wrapper(createSimpleTask(requestedIntent));
        
    }
    
    
    /**
     * Fetches the required repair, existing or newly created extended task.
     * Only suitable for INVOKING_MULTIPLE and SCHEDULING_MULTIPLE approaches.
     * 
     * This method creates extended tasks with execution dependencies.
     * Note that this method ensures that there cannot exists two different
     * tasks for intents with the same signature. However, this does not
     * mean that there can be more running handlers that would like to request
     * execution of this resolved nested task at the same moment. In other
     * words, this method avoids creation of duplicate tasks, but it is the
     * responsibility of the handler implementation to avoid conflicting
     * execution requests.
     * 
     * @param requestedIntent Reference to the requested intent.
     * @return Wrapper with the repair or task references.
     */
    public synchronized Wrapper provideExtendedWrapper(
        Intent requestedIntent
    ) {
        
        // Attempting to fetch an existing repair
        if (hasSignaturesEnabled()) {
            Repair fetchedRepair = fetchStoredRepair(requestedIntent);
            if (fetchedRepair != null) {
                return new Wrapper(fetchedRepair);
            }
        }
        
        // Attempting to fetch an existing task
        if (hasTasksCaching()) {
            Task fetchedTask = fetchCachedTask(requestedIntent);
            if (fetchedTask != null) {
                return new Wrapper(fetchedTask);
            }
        }
        
        // Creating a new extended task instance
        Task createdTask = createExtendedTask(requestedIntent);
        
        // Inserting this task into the cache
        if (hasTasksCaching()) {
            storeCreatedTask(createdTask);
        }
        
        // Returning the new task reference
        return new Wrapper(createdTask);
        
    }
    
    
    /**
     * Fetches the required repairs, existing or newly created extended tasks.
     * Only suitable for INVOKING_MULTIPLE and SCHEDULING_MULTIPLE approaches.
     * @param requestedIntents Container with all nested correction intents.
     * @return Required wrappers (fetched repairs, fetched or created tasks).
     */
    public synchronized List<Wrapper> provideExtendedWrappers(
        List<Intent> requestedIntents
    ) {
        
        // Preparing the wrappers list
        List<Wrapper> nestedWrappers =
            new ArrayList<Wrapper>(requestedIntents.size());
        
        // Fetching particular wrappers
        for (Intent nestedIntent : requestedIntents) {
            nestedWrappers.add(provideExtendedWrapper(nestedIntent));
        }
        
        // Returning the fetched wrappers
        return nestedWrappers;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Detects whether the execution approach belongs to the invoking family.
     * This means that the execution approach is one of the following:
     *   INVOKING_SINGLE and INVOKING_MULTIPLE.
     * @return Returns <code>true</code> for invoking approaches.
     */
    public final boolean hasInvokingApproach() {
        return
            (this.executionApproach == ExecutionApproach.INVOKING_SINGLE) ||
            (this.executionApproach == ExecutionApproach.INVOKING_MULTIPLE);
    }
    
    /**
     * Detects whether the execution approach belongs to the scheduling family.
     * This means that the execution approach is one of the following:
     *   SCHEDULING_SINGLE and SCHEDULING_MULTIPLE.
     * @return Returns <code>true</code> for scheduling approaches.
     */
    public final boolean hasSchedulingApproach() {
        return
            (this.executionApproach == ExecutionApproach.SCHEDULING_SINGLE) ||
            (this.executionApproach == ExecutionApproach.SCHEDULING_MULTIPLE);
    }
    
    
    /**
     * Detects whether the associated approach permits only 1 request at a time.
     * This means that the execution approach is one of the following:
     *   NESTING_SINGLE, INVOKING_SINGLE and SCHEDULING_SINGLE.
     * Note that single and multiple types must cover all defined approaches.
     * @see #hasMultipleApproach
     * @return Returns <code>true</code> for single type approaches.
     */
    public final boolean hasSingleApproach() {
        return
            (this.executionApproach == ExecutionApproach.NESTING_SINGLE) ||
            (this.executionApproach == ExecutionApproach.INVOKING_SINGLE) ||
            (this.executionApproach == ExecutionApproach.SCHEDULING_SINGLE);
    }
    
    /**
     * Detects whether the associated approach permits multiple requests.
     * This means that the execution approach is one of the following:
     *   INVOKING_MULTIPLE and SCHEDULING_MULTIPLE.
     * Note that single and multiple types must cover all defined approaches.
     * @see #hasSingleApproach
     * @return Returns <code>true</code> for multiple type approaches.
     */
    public final boolean hasMultipleApproach() {
        return
            (this.executionApproach == ExecutionApproach.INVOKING_MULTIPLE) ||
            (this.executionApproach == ExecutionApproach.SCHEDULING_MULTIPLE);
    }


    // -------------------------------------------------------------------------

}
