
package cz.cuni.mff.corrector.corrector;

import static cz.cuni.mff.corrector.corrector.Corrector.AlgorithmStrategy.*;
import static cz.cuni.mff.corrector.corrector.Corrector.ExecutionApproach.*;
import static cz.cuni.mff.corrector.execution.Variables.Phase;
import cz.cuni.mff.corrector.execution.Dependencies;
import cz.cuni.mff.corrector.execution.Quota;
import cz.cuni.mff.corrector.execution.Record;
import cz.cuni.mff.corrector.execution.Requests;
import cz.cuni.mff.corrector.execution.Task;
import cz.cuni.mff.corrector.execution.Variables;
import cz.cuni.mff.corrector.execution.Wrapper;
import cz.cuni.mff.corrector.graph.Edge;
import cz.cuni.mff.corrector.graph.Graph;
import cz.cuni.mff.corrector.graph.Vertex;
import cz.cuni.mff.corrector.intent.Intent;
import cz.cuni.mff.corrector.repair.Repair;
import cz.cuni.mff.corrector.testing.Report;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

/**
 * Correction handler implementation.
 * 
 * @author Martin Svoboda
 */
public class Handler implements Runnable {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Reference to the corrector instance.
     */
    protected final Corrector correctorReference;
    
    
    /**
     * Associated task representing the goal of this handler execution.
     */
    protected final Task associatedTask;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new correction handler instance.
     * @param correctorReference Reference to the corrector instance.
     * @param associatedTask Reference to the associated task.
     */
    public Handler(Corrector correctorReference, Task associatedTask) {
        
        // Storing provided references
        this.correctorReference = correctorReference;
        this.associatedTask = associatedTask;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Entry point of the particular correction handler execution.
     * 
     * This method is automatically called when the handler execution
     * is started, either directly in case of the NESTING approaches,
     * or indirectly by a created and started worker thread from the
     * pool in case of the INVOKING and SCHEDULING approaches.
     */
    public void run() {
        
        // Setting thread name
        Intent thisIntent = this.associatedTask.getCorrectionIntent();
        String uniqueLabel = thisIntent.getUniqueLabel();
        String threadName = Thread.currentThread().getName();
        Thread.currentThread().setName(threadName + ":" + uniqueLabel);
        
        /*DEBUG*///System.out.println("RUN " + thisIntent.getUniqueLabel());
        
        // Profiling the runs creation
        if (hasStatsProfiling()) {
            accessProfilingReport().increaseExecutionsCounter();
        }
        
        switch (this.correctorReference.getExecutionApproach()) {
            case NESTING_SINGLE:
                executeNestingCorrection();
                break;
            case INVOKING_SINGLE:
            case INVOKING_MULTIPLE:
                executeInvokingCorrection();
                break;
            case SCHEDULING_SINGLE:
            case SCHEDULING_MULTIPLE:
                executeSchedulingCorrection();
                break;
            default:
                throw new UnsupportedOperationException();
        }
        
        /*DEBUG*///System.out.println("END " + thisIntent.getUniqueLabel());
        
        // Restoring thread name
        Thread.currentThread().setName(threadName);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Entry point for the correction in case of the NESTING approaches.
     */
    protected void executeNestingCorrection() {
        if (hasStrategy(REFINE)) {
            performScatteredExecution();
        } else {
            performCompleteExecution();
        }
    }
    
    /**
     * Entry point for the correction in case of the INVOKING approaches.
     */
    protected void executeInvokingCorrection() {
        if (hasStrategy(REFINE)) {
            performScatteredExecution();
        } else {
            performCompleteExecution();
        }
    }
    
    /**
     * Entry point for the correction in case of the SCHEDULING approaches.
     */
    protected void executeSchedulingCorrection() {
        performScatteredExecution();
    }
    
    
    // -------------------------------------------------------------------------
    
    /**
     * Processes the complete evaluation.
     * 
     * Only relevant for the following corrector configurations:
     *   DEFAULT strategy:
     *     NESTING_SINGLE, INVOKING_SINGLE and INVOKING_MULTIPLE approaches,
     *   EXPLORE strategy:
     *     NESTING_SINGLE, INVOKING_SINGLE and INVOKING_MULTIPLE approaches.
     * 
     * The purpose of this method is to manage the entire execution.
     * Once the handler is started, the entire intent evaluation is completed.
     */
    protected void performCompleteExecution() {
        
        // Creating and registering variables
        Variables workingVars = createWorkingVariables();
        
        // Accessing required structures
        Repair intentRepair = this.associatedTask.getIntentRepair();
        
        // Initialising a multigraph instance
        Graph correctionGraph = initCorrectionGraph();
        
        // Graph exploration is now enabled
        
        // Performing the graph exploration
        processExplorationLoop();
        
        // Finding shortest paths in the graph
        if (hasStrategy(DEFAULT)) {
            processPathsSearching();
        }
        
        // Updating the terminal cost
        correctionGraph.setTermCost(workingVars.getFixedCost());
        
        // Declaring the repair completeness
        intentRepair.setCompletenessFlag();
        
        // Storing the repair in the cache
        if (this.correctorReference.hasSignaturesEnabled()) {
            this.correctorReference.storeComputedRepair(intentRepair);
        }
        
        // Announcing the execution end
        dispatchEndingNotifications();
        
        // Clearing working variables container
        workingVars.clearVariablesContainer();
        
    }
    
    
    /**
     * Processes the complete execution scattered into steps.
     * 
     * Only relevant for the following corrector configurations:
     *   DEFAULT strategy:
     *     SCHEDULING_SINGLE and SCHEDULING_MULTIPLE approaches,
     *   EXPLORE strategy:
     *     SCHEDULING_SINGLE and SCHEDULING_MULTIPLE approaches,
     *   REFINE strategy:
     *     all execution approaches.
     * 
     * The purpose of this method is to manage the entire execution.
     * 
     * Each time the handler is started, only a partial progress
     * of the entire intent evaluation is expected to be achieved.
     * 
     * Ensures that only one individual nested correction intent is requested
     * in case of the SCHEDULING_SINGLE execution approach and a set of all
     * outgoing nested correction intents from a particular explored vertex
     * are requested in case of the SCHEDULING_MULTIPLE approach.
     * 
     * Note that execution of a particular task can accidentally be requested
     * even if the task already finished its work. So, the body of this method
     * must correctly deal with these circumstances. For this purpose we can
     * easily use internal phases.
     * 
     * Also note that the primary and most significant difference between
     * the DEFAULT and EXPLORTING strategies on one hand and the REFINE
     * on the other is that the former ones creates and executes recursive
     * requests based on outgoing edges from selected reached vertices,
     * whereas in case of the latter one on ingoing edges to the promising
     * reached vertices without the knowledge they are or are not completed.
     */
    protected void performScatteredExecution() {
        
        // Fetching or creating working variables
        Variables workingVars = fetchWorkingVariables();
        
        // Accessing required structures
        Graph correctionGraph = workingVars.getCorrectionGraph();
        Repair intentRepair = this.associatedTask.getIntentRepair();
        
        // Initialising multigraph and repair
        if (workingVars.hasExecutionPhase(Phase.CREATED)) {
            // Creating a new multigraph
            correctionGraph = initCorrectionGraph();
            // Creating requests container
            this.associatedTask.createRequestsContainer();
            // Initialising iterative variables
            if (hasStrategy(REFINE)) {
                workingVars.initIterativeComponents();
            }
        }
        
        // Accessing requests container
        Requests requestsData = this.associatedTask.getRequestsContainer();
        
        // Closing the previous requests
        // Relevant only for SCHEDULING_SINGLE and SCHEDULING_MULTIPLE
        if (this.correctorReference.hasSchedulingApproach()) {
            if (requestsData.requestedTasks.size() > 0) {
                processRequestsClosing(requestsData);
            }
        }
        
        // List of requested tasks is now empty
        
        // Performing the graph exploration
        if (workingVars.hasExecutionPhase(Phase.EXPLORING)) {
            
            // Avoiding very fast notification
            synchronized (this.associatedTask) {
                
                // This synchronisation block is only important in case
                // of the scheduling approaches and may cause unnecessary
                // delays in case of the nesting and INVOKING approaches
                // of the REFINE strategy
                
                // Processing postponed intent requests
                if (hasApproach(SCHEDULING_SINGLE)) {
                    // Processing prepared edges
                    if (hasStrategy(REFINE)) {
                        if (requestsData.preparedTasks.size() > 0) {
                            processPreparedRefinements(requestsData);
                        }
                    } else {
                        if (requestsData.preparedEdges.size() > 0) {
                            processPreparedIntents(requestsData);
                        }
                    }
                    // Terminating when requested
                    if (requestsData.requestedTasks.size() > 0) {
                        return;
                    }
                }

                // Opening new nested requests
                if (hasStrategy(REFINE)) {
                    processIterativeLoop();
                } else {
                    processExplorationLoop();
                }

                // Terminating when requested
                if (this.correctorReference.hasSchedulingApproach()) {
                    if (requestsData.requestedTasks.size() > 0) {
                        return;
                    }
                }
                
            }
            
            // Reaching this point means that the execution of any nested
            // intent was requested at all, which means that there is
            // nothing more to explore in the multigraph and also nothing
            // remaing postponed, therefore, we can terminate the entire
            // phase of the multigraph construction
            
            // Finishing the entire exploration
            if (!workingVars.hasReachedVertices()) {
                workingVars.setExecutionPhase(Phase.EXPLORED);
            }
            
        }
        
        // The following code is not reachable having requested nested tasks
        // This means "requested" in the relation to the SCHEDULING approaches
        
        // Closing the refinement
        if (hasStrategy(REFINE)) {
            
            // Updating the terminal cost
            if (workingVars.getFixedCost() == null) {
                // Publishing the current estimation of the final cost
                correctionGraph.setTermCost(workingVars.getReachedCost());
            } else {
                // Publishing already resolved final correction cost
                correctionGraph.setTermCost(workingVars.getFixedCost());
            }
            
            // Testing quota availability
            if (
                workingVars.hasExecutionPhase(Phase.EXPLORING) &&
                hasExceededQuota()
            ) {
                
                // This termination block ensures that when we have already
                // used the entire granted quota and we have just finished
                // the exploration loop (regardless there are still some
                // reached vertices left or not), we terminate here to avoid
                // processing of delayed edges and the finalisation itself
                
                // Announcing the refinement end
                dispatchEndingNotifications();
                
                // Returning from this method
                return;
                
            }
            
        }
        
        // The following code is not reachable in case of attained refinement
        
        // Finding shortest paths in the graph
        if (workingVars.hasExecutionPhase(Phase.EXPLORED)) {
            
            // Processing unfinished delayed edges
            if (hasStrategy(REFINE)) {
                // Processing already prepared edges (SCHEDULING_SINGLE)
                if (requestsData.preparedTasks.size() > 0) {
                    processPreparedRefinements(requestsData);
                }
                // Terminating when requested
                if (this.correctorReference.hasSchedulingApproach()) {
                    if (requestsData.requestedTasks.size() > 0) {
                        return;
                    }
                }
                // 
                while (workingVars.hasDelayedEdges()) {
                    processDelayedEdges();
                    // Terminating when requested
                    if (this.correctorReference.hasSchedulingApproach()) {
                        if (requestsData.requestedTasks.size() > 0) {
                            return;
                        }
                    }
                }
            }
            
            // Finding shortest paths in the graph
            if (hasStrategy(DEFAULT)) {
                processPathsSearching();
            }
            
            // Updating the terminal cost
            correctionGraph.setTermCost(workingVars.getFixedCost());
            
            // Declaring the repair completeness
            intentRepair.setCompletenessFlag();
            
            // Storing the repair in the cache
            if (this.correctorReference.hasSignaturesEnabled()) {
                this.correctorReference.storeComputedRepair(intentRepair);
            }
            
            // Clearing working variables container
            workingVars.clearVariablesContainer();
            
            // Despite variables are destroied, the current phase is preserved
            
            // Finishing the entire evaluation
            workingVars.setExecutionPhase(Phase.FINISHED);
            
        }
        
        // Notifying parental tasks
        if (workingVars.hasExecutionPhase(Phase.FINISHED)) {
            // Announcing the execution end
            dispatchEndingNotifications();
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Performs or continues the multigraph exploration.
     * This method is used by the DEFAULT and EXPLORE execution approaches!
     */
    protected void processExplorationLoop() {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        Graph correctionGraph = workingVars.getCorrectionGraph();
       
        // Processing reached vertices
        while (workingVars.hasReachedVertices()) {
            
            // Selecting suitable reached vertex
            Vertex selectedVertex;
            switch (this.correctorReference.getAlgorithmStrategy()) {
                case DEFAULT:
                    // Selecting any reached vertex
                    selectedVertex = selectAnyVertex();
                    break;
                case EXPLORE:
                    // Selecting the cheapest vertex
                    selectedVertex = selectCheapestVertex();
                    // Updating the reached cost
                    workingVars.setReachedCost(
                        correctionGraph.getTracingCost(selectedVertex)
                    );
                    break;
                default:
                    throw new UnsupportedOperationException();
            }
            
            // Managing paths exploration
            if (hasStrategy(EXPLORE)) {
                
                // Detecting the interruption
                if (shouldQuitExploring(workingVars.getReachedCost())) {
                    // Clearing all reached vertices
                    workingVars.clearReachedVertices();
                    // Terminating the exploration
                    return;
                }
                
            }
            
            // Removing the vertex from the reached ones
            workingVars.removeReachedVertex(selectedVertex);
            
            // Managing paths exploration
            if (hasStrategy(EXPLORE)) {
                
                // Detecting the cost fixation
                if (shouldFixateCost(selectedVertex)) {
                    // Fixing the current cost
                    workingVars.setFixedCost(workingVars.getReachedCost());
                }
                
            }
            
            // Processing the selected vertex
            processSelectedVertex(selectedVertex);
            
            // Note that in case of the...
            // - NESTING_SINGLE approach all nested requests are called
            //   directly and successively executed and evaluated one by one.
            // - INVOKING_SINGLE and INVOKING_MULTIPLE approaches all nested
            //   requests were directly executed by workers from the pool.
            // - SCHEDULING_SINGLE and SCHEDULING_MULTIPLE approaches we need
            //   to block the processing when the execution of some nested
            //   tasks have been requested! Since the requested tasks list
            //   is used only by these approaches, the following condition
            //   would be correct even without the test.
            
            // Terminating when invoked some requests
            if (this.correctorReference.hasSchedulingApproach()) {
                synchronized (this.associatedTask) {
                    Requests requestsData = this.associatedTask.
                        getRequestsContainer();
                    if (requestsData.requestedTasks.size() > 0) {
                        return;
                    }
                }
            }
            
        }
        
    }
    
    
    /**
     * Performs or continues the multigraph exploration.
     * This method is relevant only for the REFINE execution approach!
     */
    protected void processIterativeLoop() {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
       
        // Processing reached vertices
        while (workingVars.hasReachedVertices()) {
            
            // Selecting the cheapest report
            Record selectedReport = selectPromisingReport();
            
            // Note that we could implement an optimization that would
            // process all cheapest promising reports right now, since
            // we are sure that all of them will certainly be processed
            // anyway. On the other hand, very high number of recursive
            // requests created at one time would not be suitable for
            // the INVOKING_MULTIPLE execution approach.
            
            // Accessing the selected vertex
            Vertex selectedVertex = selectedReport.associatedVertex;
            
            // Updating the reached cost
            workingVars.setReachedCost(selectedReport.generalMinimum);
            
            // Detecting vertex completeness
            Boolean isComplete = isVertexComplete(selectedReport);
            
            // Detecting the interruption
            if (shouldQuitExploring(workingVars.getReachedCost())) {
                // Clearing all reached vertices
                workingVars.clearReachedVertices();
                // Terminating the exploration
                return;
            }
            
            // Testing quota availability
            if (hasExceededQuota()) {
                return;
            }
            
            // Processing the completed vertex
            if (isComplete) {
                
                // Removing the vertex from the reached ones
                workingVars.removeReachedVertex(selectedVertex);
                
                // Detecting the cost fixation
                if (shouldFixateCost(selectedVertex)) {
                    workingVars.setFixedCost(workingVars.getReachedCost());
                }
                
                // Processing the selected vertex
                processCompleteVertex(selectedReport, selectedVertex);
                
            // Processing the incomplete vertex
            } else {
                
                // Processing the selected vertex
                processIncompleteVertex(selectedReport, selectedVertex);
                
            }
            
            // Terminating when invoked some requests
            if (this.correctorReference.hasSchedulingApproach()) {
                Requests requestsData = this.associatedTask.
                    getRequestsContainer();
                if (requestsData.requestedTasks.size() > 0) {
                    return;
                }
            }
            
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Performs or starts the further exploration of the selected vertex.
     * Only relevant for the DEFAULT and EXPLORE strategies!
     * Depending on the approach, one or more requests may be invoked.
     * On the other hand, note that no request at all might also be invoked.
     * @param selectedVertex Selected vertex to be further explored.
     */
    protected void processSelectedVertex(
        Vertex selectedVertex
    ) {
        
        // Accessing required variables
        Intent parentIntent = this.associatedTask.getCorrectionIntent();
        
        // Generating allowed nested intents
        List<Intent> nestedIntents = Intent.generateNestedIntents(
            parentIntent,
            selectedVertex.getStratumNumber(),
            selectedVertex.getAutomatonState(),
            this.correctorReference.treeGrammar,
            this.correctorReference.enabledIntents
        );
        
        // Requesting the recursive execution
        dispatchStandardRequests(selectedVertex, nestedIntents);
        
    }
    
    
    /**
     * Processes the complete vertex using the iterative strategy.
     * Only relevant for the REFINE algorithm strategy!
     * @param reportData Full report for the selected reached vertex.
     * @param selectedVertex Selected vertex to be further explored.
     */
    protected void processCompleteVertex(
        Record reportData,
        Vertex selectedVertex
    ) {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        Intent parentIntent = this.associatedTask.getCorrectionIntent();
        Graph correctionGraph = workingVars.getCorrectionGraph();
        
        // Processing edges delaying
        workingVars.addDelayedEdges(reportData.delayedEdges);
        
        // Removing unnecessary tasks
        for (Edge uselessEdge : reportData.uselessEdges) {
            workingVars.removeCachedTask(uselessEdge.getCorrectionIntent());
        }
        
        // Storing paths tracing data
        correctionGraph.
            setTracingCost(selectedVertex, reportData.tracingCost);
        correctionGraph.
            setTracingEdges(selectedVertex, reportData.tracingEdges);
        
        // Generating allowed nested intents
        List<Intent> nestedIntents = Intent.generateNestedIntents(
            parentIntent,
            selectedVertex.getStratumNumber(),
            selectedVertex.getAutomatonState(),
            this.correctorReference.treeGrammar,
            this.correctorReference.enabledIntents
        );
        
        // Exploring allowed correction intents
        for (Intent nestedIntent : nestedIntents) {
            
            // Opening the edge processing
            Edge nestedEdge = doEdgeOpening(nestedIntent);
            
            // Acquiring the nested wrapper
            Wrapper nestedWrapper;
            if (this.correctorReference.hasSingleApproach()) {
                nestedWrapper = this.correctorReference.
                    provideSimpleWrapper(nestedIntent);
            } else {
                nestedWrapper = this.correctorReference.
                    provideExtendedWrapper(nestedIntent);
            }
            
            // Accessing the wrapper items
            Repair fetchedRepair = nestedWrapper.repairReference;
            Task nestedTask = nestedWrapper.taskReference;
            
            // Assigning the repair reference
            if (fetchedRepair != null) {
                nestedEdge.setEdgeRepair(fetchedRepair);
            } else {
                nestedEdge.setEdgeRepair(nestedTask.getIntentRepair());
            }
            
            // Delaying reverse edge (might not be required but is correct)
            if (!workingVars.hasReachedVertex(nestedEdge.getTargetVertex())) {
                workingVars.addDelayedEdge(nestedEdge);
            }
            
            // Locally caching the task
            if (nestedTask != null) {
                workingVars.addCachedTask(nestedIntent, nestedTask);
            }
            
        }
        
    }
    
    
    /**
     * Processes the incomplete vertex using the iterative strategy.
     * Only relevant for the REFINE algorithm strategy!
     * @param reportData Full report for the selected reached vertex.
     * @param selectedVertex Selected vertex to be further explored.
     */
    protected void processIncompleteVertex(
        Record reportData,
        Vertex selectedVertex
    ) {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        
        // Returning on no prepared promising edges to be refined
        if (reportData.promisingEdges.isEmpty()) {
            return;
        }
        
        // Fetching the number of promising requests
        int itemsCount = reportData.promisingEdges.size();
        
        // Preparing requests containers
        List<Edge> nestedEdges = new ArrayList<Edge>(itemsCount);
        List<Task> nestedTasks = new ArrayList<Task>(itemsCount);
        List<Quota> nestedQuotas = new ArrayList<Quota>(itemsCount);
        
        // Preparing iterative requests
        for (int index = 0; index < itemsCount; index++) {
            
            // Accessing required objects
            Edge nestedEdge = reportData.promisingEdges.get(index);
            Intent nestedIntent = nestedEdge.getCorrectionIntent();
            Task nestedTask = workingVars.getCachedTask(nestedIntent);
            
            // Preparing the granted quota
            Quota nestedQuota;
            if (itemsCount == 1) {
                Integer quotaBoundary = reportData.quotaBoundary;
                if (quotaBoundary == null) {
                    quotaBoundary = reportData.closedMinimum;
                    if (reportData.closedMinimum == reportData.openMinimum) {
                        quotaBoundary++;
                    }
                }
                nestedQuota = resolveGrantedQuota(
                    nestedEdge,
                    (quotaBoundary == null) ?
                        1 : (quotaBoundary - reportData.openMinimum)
                );
            } else { // (itemsCount > 1)
                nestedQuota = resolveGrantedQuota(nestedEdge, 1);
            }
            
            // Storing the request references
            nestedEdges.add(nestedEdge);
            nestedTasks.add(nestedTask);
            nestedQuotas.add(nestedQuota);
            
        }
        
        // Dispatching requested refinements
        dispatchIterativeRequests(nestedEdges, nestedTasks, nestedQuotas);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Process one of the prepared or postponed nested intents and its edge.
     * This method is only relevant to the SCHEDULING_SINGLE approach.
     * It is called directly from the main exploration loop when a particular
     * selected vertex is being processed, or after the task execution is
     * reentered once again and some of these not yet scheduled intents
     * still exist in the container of postponed requests.
     * @param requestsData Shortcut to the requests data container.
     */
    protected void processPreparedIntents(Requests requestsData) {
        
        // Iterating over the prepared intents
        while (requestsData.preparedEdges.size() > 0) {
            
            // Selecting one postponed intent
            Edge nestedEdge = requestsData.preparedEdges.remove(0);

            // Acquiring the nested repair
            handleSchedulingSingle(requestsData, nestedEdge);
            
            // Only one request at a time
            if (requestsData.requestedTasks.size() > 0) {
                return;
            }
            
        }
        
    }
    
    
    /**
     * Process one of the prepared or postponed nested intents and its edge.
     * Only relevant to the SCHEDULING_SINGLE approach of REFINE strategy.
     * @param requestsData Shortcut to the requests data container.
     */
    protected void processPreparedRefinements(
        Requests requestsData
    ) {
        
        // Iterating over the prepared intents
        while (requestsData.preparedTasks.size() > 0) {
            
            // Selecting one postponed intent
            Task nestedTask = requestsData.preparedTasks.remove(0);
            Quota grantedQuota = requestsData.preparedQuotas.remove(0);

            // Activating the nested task
            if (!nestedTask.getIntentRepair().isCompletelyEvaluated()) {
                scheduleIterativeSimple(
                    requestsData, nestedTask, grantedQuota
                );
            }
            
            // Only one request at a time
            if (requestsData.requestedTasks.size() > 0) {
                return;
            }
            
        }
        
    }
    
    
    /**
     * Processes the delayed edges using the iterative strategy.
     * Only relevant for the REFINE strategy!
     */
    protected void processDelayedEdges() {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        Graph correctionGraph = workingVars.getCorrectionGraph();
        
        // Preparing requests containers
        List<Edge> nestedEdges = new ArrayList<Edge>();
        List<Task> nestedTasks = new ArrayList<Task>();
        List<Quota> nestedQuotas = new ArrayList<Quota>();
        
        // Iterating over delayed edges
        for (Edge delayedEdge : workingVars.getDelayedEdges()) {
            
            // Fetching required data
            boolean isCompleted = delayedEdge.getIntentRepair().
                isCompletelyEvaluated();
            int sourceCost = correctionGraph.
                getTracingCost(delayedEdge.getSourceVertex());
            int targetCost = correctionGraph.
                getTracingCost(delayedEdge.getTargetVertex());
            int edgeCost = delayedEdge.getEdgeCost();
            int inspectedCost = sourceCost + edgeCost;
            
            // Processing completed edge
            if (isCompleted) {
                // Finalising suitable edge
                if (inspectedCost == targetCost) {
                    // Updating the tracing records
                    correctionGraph.addTracingEdge(
                        delayedEdge.getTargetVertex(), delayedEdge
                    );
                }
                // Removing the cached task
                workingVars.removeCachedTask(
                    delayedEdge.getCorrectionIntent()
                );
            // Processing incomplete edge
            } else {
                // Requesting promising edge
                if (inspectedCost <= targetCost) {
                    // Preparing request data
                    Task nestedTask = workingVars.getCachedTask(
                        delayedEdge.getCorrectionIntent()
                    );
                    Quota grantedQuota = resolveGrantedQuota(
                        delayedEdge, targetCost - inspectedCost + 1
                    );
                    // Storing prepared request
                    nestedEdges.add(delayedEdge);
                    nestedTasks.add(nestedTask);
                    nestedQuotas.add(grantedQuota);
                } else {
                    // Removing the cached task
                    workingVars.removeCachedTask(
                        delayedEdge.getCorrectionIntent()
                    );
                }
            }
            
        }
        
        // Preserving the updated list
        workingVars.removeDelayedEdges();
        workingVars.addDelayedEdges(nestedEdges);
        
        // Dispatching requested refinements
        dispatchIterativeRequests(nestedEdges, nestedTasks, nestedQuotas);
        
    }
    
    
    /**
     * Closes the set of previously requested intents and their exploration.
     * Since edges and potentially also ending vertices of the previously
     * requested nested intents are already inserted into the multigraph,
     * we only need to assign these edges with newly computed repairs.
     * @param requestsData Shortcut to the requests data container.
     */
    protected void processRequestsClosing(Requests requestsData) {
        
        // Processing recorder requests
        if (!hasStrategy(REFINE)) {
            
            // Iterating over reqeusts
            int itemsCount = requestsData.requestedTasks.size();
            for (int index = 0; index < itemsCount; index++) {

                // Closing the edge processing
                doEdgeClosing(
                    requestsData.requestedEdges.get(index),
                    requestsData.requestedTasks.get(index).getIntentRepair()
                );

            }
            
        }
        
        // Clearing the requests container
        requestsData.requestedTasks.clear();
        requestsData.requestedEdges.clear();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Dispatches all requests to the standard execution processing.
     * Only relevant for the DEFAULT and EXPLORE algorithm strategies.
     * @param selectedVertex Selected vertex to be further explored.
     * @param nestedIntents List of nested intents to be processed.
     */
    public void dispatchStandardRequests(
        Vertex selectedVertex,
        List<Intent> nestedIntents
    ) {
        
        // Requesting the recursive execution
        switch (this.correctorReference.getExecutionApproach()) {
            case NESTING_SINGLE:
                handleNestingSingle(selectedVertex, nestedIntents);
                break;
            case INVOKING_SINGLE:
                handleInvokingSingle(selectedVertex, nestedIntents);
                break;
            case INVOKING_MULTIPLE:
                handleInvokingMultiple(selectedVertex, nestedIntents);
                break;
            case SCHEDULING_SINGLE:
                handleSchedulingSingle(selectedVertex, nestedIntents);
                break;
            case SCHEDULING_MULTIPLE:
                handleSchedulingMultiple(selectedVertex, nestedIntents);
                break;
            default:
                throw new UnsupportedOperationException();
            
        }
        
    }
    
    
    /**
     * Dispatches all requests to the iterative refinement processing.
     * Only relevant for the REFINE algorithm strategy.
     * @param nestedEdges List of nested edges to be executed.
     * @param nestedTasks List of associated correction tasks.
     * @param nestedQuotas List of associated assigned quotas.
     */
    public void dispatchIterativeRequests(
        List<Edge> nestedEdges,
        List<Task> nestedTasks,
        List<Quota> nestedQuotas
    ) {
        
        // Acquiring tasks execution
        switch (this.correctorReference.getExecutionApproach()) {
            case NESTING_SINGLE:
                requestNestingSingle(nestedTasks, nestedQuotas);
                break;
            case INVOKING_SINGLE:
                requestInvokingSingle(nestedTasks, nestedQuotas);
                break;
            case INVOKING_MULTIPLE:
                requestInvokingMultiple(nestedTasks, nestedQuotas);
                break;
            case SCHEDULING_SINGLE:
                requestSchedulingSingle(nestedTasks, nestedQuotas);
                break;
            case SCHEDULING_MULTIPLE:
                requestMultipleScheduling(nestedTasks, nestedQuotas);
                break;
            default:
                throw new UnsupportedOperationException();
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Processes exploration and nested evaluation of all allowed intents.
     * Only for NESTING_SINGLE approach of DEFAULT and EXPLORE strategies.
     * @param selectedVertex Selected vertex to be further explored.
     * @param nestedIntents List of nested intents to be processed.
     */
    protected void handleNestingSingle(
        Vertex selectedVertex,
        List<Intent> nestedIntents
    ) {
        
        // Exploring allowed correction intents
        for (Intent nestedIntent : nestedIntents) {
            
            // Opening the edge processing
            Edge nestedEdge = doEdgeOpening(nestedIntent);
            
            // Acquiring the nested repair
            Repair nestedRepair = acquireNestingSingle(nestedIntent);
            
            // Closing the particular request
            doEdgeClosing(nestedEdge, nestedRepair);
            
        }
        
    }
    
    
    /**
     * Processes exploration and invoked evaluation of all allowed intents.
     * Only for INVOKING_SINGLE approach of DEFAULT and EXPLORE strategies.
     * @param selectedVertex Selected vertex to be further explored.
     * @param nestedIntents List of nested intents to be processed.
     */
    protected void handleInvokingSingle(
        Vertex selectedVertex,
        List<Intent> nestedIntents
    ) {
        
        // Exploring allowed correction intents
        for (Intent nestedIntent : nestedIntents) {
            
            // Opening the edge processing
            Edge nestedEdge = doEdgeOpening(nestedIntent);
            
            // Acquiring the nested repair
            Repair nestedRepair = acquireInvokingSingle(nestedIntent);
            
            // Closing the particular request
            doEdgeClosing(nestedEdge, nestedRepair);
            
        }
        
    }
    
    
    /**
     * Processes exploration and invoked evaluation of all allowed intents.
     * Only for INVOKING_MULTIPLE approach of DEFAULT and EXPLORE strategies.
     * @param selectedVertex Selected vertex to be further explored.
     * @param nestedIntents List of nested intents to be processed.
     */
    protected void handleInvokingMultiple(
        Vertex selectedVertex,
        List<Intent> nestedIntents
    ) {
        
        // Opening the nested edges
        List<Edge> nestedEdges =
            doEdgesOpening(nestedIntents);
        
        // Acquiring all nested repairs
        List<Repair> nestedRepairs = acquireInvokingMultiple(nestedIntents);
        
        // Exploring all nested intents
        for (int index = 0; index < nestedIntents.size(); index++) {
            
            // Closing the particular request
            doEdgeClosing(
                nestedEdges.get(index),
                nestedRepairs.get(index)
            );
            
        }
        
    }
    
    
    /**
     * Processes exploration and scheduled evaluation of all allowed intents.
     * Only for SCHEDULING_SINGLE approach of DEFAULT and EXPLORE strategies.
     * Stores all nested intents into the requests container of prepared
     * intents and iterates until some (if any) of the newly prepared intents
     * is actually scheduled for its execution.
     * @param selectedVertex Selected vertex to be further explored.
     * @param nestedIntents List of nested intents to be processed.
     */
    protected void handleSchedulingSingle(
        Vertex selectedVertex,
        List<Intent> nestedIntents
    ) {
        
        // Shortcut to the requests container
        Requests requestsData = this.associatedTask.getRequestsContainer();

        // Iterating over all nested intents
        for (Intent nestedIntent : nestedIntents) {

            // Opening the edge processing
            Edge nestedEdge = doEdgeOpening(nestedIntent);

            // Marking the edge as prepared
            requestsData.preparedEdges.add(nestedEdge);

        }

        // Processing some (if any) nested prepared intent
        processPreparedIntents(requestsData);
        
    }
    
    
    /**
     * Processes exploration and scheduled evaluation of the given intent.
     * Only for SCHEDULING_SINGLE approach of DEFAULT and EXPLORE strategies.
     * @param requestsData Shortcut to the requests data container.
     * @param nestedEdge Particular postponed edge and its intent.
     */
    protected void handleSchedulingSingle(
        Requests requestsData,
        Edge nestedEdge
    ) {
        
        // Note that the corresponding edge has already been explored!
        
        // Accessing the nested intent
        Intent nestedIntent = nestedEdge.getCorrectionIntent();
        
        // Requesting the nested execution
        Repair nestedRepair =
            acquireSchedulingSingle(requestsData, nestedIntent, nestedEdge);
        
        // Note that the returned existing repair must be stored immediately!
        
        // Closing the edge exploration
        if (nestedRepair != null) {
            doEdgeClosing(nestedEdge, nestedRepair);
        }
        
    }
    
    
    /**
     * Processes exploration and scheduled evaluation of all allowed intents.
     * Only SCHEDULING_MULTIPLE approach of DEFAULT and EXPLORE strategies.
     * @param selectedVertex Selected vertex to be further explored.
     * @param nestedIntents List of nested intents to be processed.
     */
    protected void handleSchedulingMultiple(
        Vertex selectedVertex,
        List<Intent> nestedIntents
    ) {
        
        // Opening the nested edges
        List<Edge> nestedEdges = doEdgesOpening(nestedIntents);
        
        // Accessing requests container
        Requests requestsData = this.associatedTask.getRequestsContainer();
        
        // Requesting all nested executions
        List<Repair> nestedRepairs = acquireSchedulingMultiple(
            requestsData, nestedIntents, nestedEdges
        );
        
        // Note that the returned existing repair must be stored immediately!
        
        // Closing finalised edges
        for (int index = 0; index < nestedRepairs.size(); index++) {
            // Accessing intent and repair
            Repair nestedRepair = nestedRepairs.get(index);
            Edge nestedEdge = nestedEdges.get(index);
            // Closing the edge exploration
            if (nestedRepair != null) {
                doEdgeClosing(nestedEdge, nestedRepair);
            }
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Opens the exploration of a newly reached edge and its intent.
     * The corresponding edge is inserted into the multigraph and also
     * the ending vertex is inserted into it and above that into the
     * set of reached vertices in case that this ending vertex has not
     * yet been explored before.
     * @param nestedIntent Nested correction intent reference.
     * @return Reference to the newly created correction edge.
     */
    protected Edge doEdgeOpening(Intent nestedIntent) {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        Graph correctionGraph = workingVars.getCorrectionGraph();
        
        // Accessing the ending vertex from the correction intent
        Vertex endingVertex = nestedIntent.getEndingVertex();

        // Detecting the ending vertex presence in the multigraph
        if (!correctionGraph.hasVertexItem(endingVertex)) {
            // Adding a newly explored vertex into the multigraph
            correctionGraph.addNewVertex(endingVertex);
            // Adding this vertex also between reached vertices
            workingVars.addReachedVertex(endingVertex);
            // Profiling the vertex creation
            if (hasStatsProfiling()) {
                accessProfilingReport().increaseVerticesCounter();
            }
        }
        
        // Creating the edge and adding it into the multigraph
        Edge newEdge = new Edge(nestedIntent);
        correctionGraph.addNewEdge(newEdge);
        
        // Profiling the edge creation
        if (hasStatsProfiling()) {
            accessProfilingReport().increaseEdgesCounter();
        }
        
        // Returning the edge
        return newEdge;
        
    }
    
    
    /**
     * Opens the exploration of edges for the entire set of intents.
     * This method only opens edge processing one by one.
     * @see #doEdgeOpening
     * @param nestedIntents Set of nested correction intents
     * @return List of the newly created correction edges.
     * The output edges order corresponds to the input intents order.
     */
    protected List<Edge> doEdgesOpening(
        List<Intent> nestedIntents
    ) {
        
        // Preparing the edges container
        List<Edge> nestedEdges = new ArrayList<Edge>(nestedIntents.size());
        
        // Exploring all nested intents
        for (int index = 0; index < nestedIntents.size(); index++) {
            
            // Accessing the intent reference
            Intent nestedIntent = nestedIntents.get(index);
            
            // Opening the edge processing
            Edge nestedEdge = doEdgeOpening(nestedIntent);
            
            // Preserving the edge reference
            nestedEdges.add(nestedEdge);
            
        }
        
        // Returning constructed edges
        return nestedEdges;
    
    }
    
    
    /**
     * Closes the exploration of a given edge and its correction intent.
     * This procedure is called from the parental task when the nested,
     * invoked or scheduled recursive request has been completed. This
     * means that in case of the DEFAULT and EXPLORE strategies we are
     * now provided with the completely evaluated repair, whereas in case
     * of the REFINE strategy only a partial refinement was achieved
     * and we do not need to do anything.
     * @param nestedEdge Reference to the edge to be closed.
     * @param nestedRepair Reference to the repair to be associated.
     */
    protected void doEdgeClosing(
        Edge nestedEdge,
        Repair nestedRepair
    ) {
        
        // Registering the repair with the edge
        if (!hasStrategy(REFINE)) {
            nestedEdge.setEdgeRepair(nestedRepair);
        }
        
        // Managing paths exploration
        if (hasStrategy(EXPLORE)) {
            // Resolving path records
            updatePathsTracing(nestedEdge);
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Fetches the required repair or requests its nested execution.
     * Only for NESTING_SINGLE approach of DEFAULT and EXPLORE strategies.
     * The cache is considered when signatures handling is enabled.
     * @param nestedIntent Nested correction intent to be evaluated.
     * @return Reference to the created simple task.
     */
    protected Repair acquireNestingSingle(Intent nestedIntent) {
        
        // Fetching the repair or task
        Wrapper nestedWrapper =
            this.correctorReference.provideSimpleWrapper(nestedIntent);
        
        // Returning the cached repair
        if (nestedWrapper.repairReference != null) {
            return nestedWrapper.repairReference;
        }
        
        // Accessing the task reference
        Task nestedTask = nestedWrapper.taskReference;

        // Creating the handler instance
        Handler nestedHandler =
            this.correctorReference.createHandlerInstance(nestedTask);

        // Executing the nested handler
        nestedHandler.run();
        
        // Returning the task reference
        return nestedTask.getIntentRepair();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Requests the nested refinement execution for a set of tasks.
     * Relevant for the NESTING_SINGLE approach of the REFINE strategy.
     * @param nestedTasks List of tasks to be requested.
     * @param grantedQuotas List of associated granted quotas.
     */
    protected void requestNestingSingle(
        List<Task> nestedTasks,
        List<Quota> grantedQuotas
    ) {
        
        // Successively processing tasks
        for (int index = 0; index < nestedTasks.size(); index++) {
            
            // Accessing request data
            Task nestedTask = nestedTasks.get(index);
            Quota grantedQuota = grantedQuotas.get(index);
            
            // Creating the handler instance
            Handler nestedHandler =
                this.correctorReference.createHandlerInstance(nestedTask);

            // Passing the quota parameter
            nestedTask.setQuotaData(grantedQuota);

            // Executing the nested handler
            nestedHandler.run();

        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Fetches the required repair or requests its invoked execution.
     * Only for INVOKING_SINGLE approach of DEFAULT and EXPLORE strategies.
     * @param nestedIntent Nested correction intent to be evaluated.
     * @return Reference to the requested repair structure.
     */
    protected Repair acquireInvokingSingle(Intent nestedIntent) {
        
        // Fetching the repair or task
        Wrapper nestedWrapper =
            this.correctorReference.provideSimpleWrapper(nestedIntent);
        
        // Returning the cached repair
        if (nestedWrapper.repairReference != null) {
            return nestedWrapper.repairReference;
        }
        
        // Accessing the task reference
        Task nestedTask = nestedWrapper.taskReference;
        
        // Note that we do not need to use task dependencies and more
        // complicated synchronisation, since only one request can be
        // invoked at a time. This means that we do not need to be aware
        // of creating more different task instances for the same intent
        // signature, nor we have to avoid multiple invocations of the
        // same task
        
        // Creating the handler instance
        Handler nestedHandler =
            this.correctorReference.createHandlerInstance(nestedTask);
        
        // Requesting the handler execution
        this.correctorReference.requestHandlerExecution(nestedHandler);
        
        // Waiting for the nested execution
        waitCompletionSingle(nestedTask);
        
        // Returning the evaluated repair
        return nestedTask.getIntentRepair();
        
    }
    
    
    /**
     * Waits for the execution of the particular selected nested task.
     * Only for INVOKING_SINGLE approach of DEFAULT and EXPLORE strategies.
     * The passed monitor should be the task for which execution is waiting!
     * @param nestedTask Nested task for which execution should be wait.
     */
    private void waitCompletionSingle(Task nestedTask) {
        
        // Synchronising on the nested task
        synchronized (nestedTask) {
            try {
                // Avoiding spurious but possible wakeups
                while (!nestedTask.hasCompletedRapair()) {
                    nestedTask.wait();
                }
            } catch (InterruptedException ex) {
                throw new Error(
                    "Waiting for the invoked execution interrupted!"
                );
            }
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Requests the invoked refinement execution for a set of tasks.
     * Relevant for the INVOKING_SINGLE approach of the REFINE strategy.
     * @param nestedTasks List of tasks to be requested.
     * @param grantedQuotas List of associated granted quotas.
     */
    protected void requestInvokingSingle(
        List<Task> nestedTasks,
        List<Quota> grantedQuotas
    ) {
        
        // Successively processing tasks
        for (int index = 0; index < nestedTasks.size(); index++) {
            
            // Accessing the nested items
            Task nestedTask = nestedTasks.get(index);
            Quota grantedQuota = grantedQuotas.get(index);
            
            // Activating the nested task
            triggerInvokingSingle(nestedTask, grantedQuota);
            
            // Waiting for the nested execution
            waitRefinementSingle(nestedTask);
            
        }
        
    }
    
    
    /**
     * Requests the invoked refinement execution for a given task.
     * Relevant for the INVOKING_SINGLE approach of the REFINE strategy.
     * @param nestedTask Reference to the task to be requested.
     * @param grantedQuota Associated granted quota to be passed.
     */
    protected void triggerInvokingSingle(
        Task nestedTask,
        Quota grantedQuota
    ) {
        
        // Creating the handler instance
        Handler nestedHandler =
            this.correctorReference.createHandlerInstance(nestedTask);
        
        // Passing the quota parameter
        nestedTask.setQuotaData(grantedQuota);
        
        // Requesting the handler execution
        this.correctorReference.requestHandlerExecution(nestedHandler);
        
    }
    
    
    /**
     * Waits for the execution of the particular selected nested task.
     * Only relevant for the INVOKING_SINGLE approach and REFINE strategy.
     * The passed monitor should be the task for which execution is waiting!
     * @param nestedTask Nested task for which execution should be wait.
     */
    private void waitRefinementSingle(Task nestedTask) {
        
        // Synchronising on the nested task
        synchronized (nestedTask) {
            try {
                // Avoiding spurious but possible wakeups
                while (!nestedTask.hasCompletedJob()) {
                    nestedTask.wait();
                }
            } catch (InterruptedException ex) {
                throw new Error(
                    "Waiting for the invoked execution interrupted!"
                );
            }
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Fetches the required repairs or requests their invoked execution.
     * Only for INVOKING_MULTIPLE approach of DEFAULT and EXPLORE strategies.
     * The cache is considered when signatures handling is enabled.
     * @param nestedIntents Nested correction intents to be evaluated.
     * @return Reference to the requested repair structures.
     * The output order of repairs is equal to the input order of intents!
     */
    protected List<Repair> acquireInvokingMultiple(
        List<Intent> nestedIntents
    ) {
        
        // Fetching nested wrappers
        List<Wrapper> nestedWrappers =
            this.correctorReference.provideExtendedWrappers(nestedIntents);
        
        // Synchronising on the associated task
        synchronized (this.associatedTask) {
            
            // Note that this synchronisation block must overlap both the
            // following calls. Otherwise the nested thread could manage
            // to finish and notify this task before this task even manage
            // to start waiting for such notifications!
            
            // Requesting and potentially also activating nested wrappers
            boolean hasRequested = doWrappersObtaining(nestedWrappers);

            // This means that we have requested for all nested taks (if the
            // corresponding repairs are not already cached), but only those
            // tasks that were currently disbaled have been actually activated
            
            // We are also assured that the beginning of the waiting preceds
            // any notification from the nested tasks

            // Waiting for requested executions
            if (hasRequested) {
                waitUniversalMultiple();
            }
            
        }
        
        // Fetching all nested repairs
        List<Repair> outputList = doWrappersUnfolding(nestedWrappers);
        
        // Returning the nested repairs
        return outputList;
        
    }
    
    
    /**
     * Requests and optionally also activates invoked intents execution.
     * Only for INVOKING_MULTIPLE approach of DEFAULT and EXPLORE strategies.
     * Request dependencies are updated always, but nested handlers are
     * created and invoked only when the given nested tasks are currently
     * disabled, i.e. their activation has not yet been set by some other
     * concurrently running task.
     * Note that his method must be called from within the synchronisation
     * block over the calling associated task!
     * @param nestedWrappers Container with all wrappers (repairs or tasks).
     * @return Returns <code>true</code> when at least one task was requested.
     */
    private boolean doWrappersObtaining(List<Wrapper> nestedWrappers) {
        
        // Preparing the requests flag
        boolean hasRequested = false;
        
        // Synchronising on the associated task
        synchronized (this.associatedTask) {
            
            // This explicit synchronisation is necessary, because dependencies
            // of this task have to be managed in a controlled way
            
            // Requesting nested executions
            for (Wrapper nestedWrapper : nestedWrappers) {
                
                // Nested repair is already cached
                if (nestedWrapper.repairReference != null) {
                    continue;
                }
                
                // Referencing the nested task
                Task nestedTask = nestedWrapper.taskReference;
                
                // Requesting the nested execution
                synchronized (nestedTask) {
                    
                    // This explicit synchronisation is required to correctly
                    // detect and manage nested task execution state and also
                    // its execution dependencies
                    
                    // The repair is surprisingly ready
                    if (nestedTask.hasCompletedRapair()) {
                        // This test is for situations, when the nested task
                        // has already finished its execution, i.e. evaluated
                        // and published the repair, notified all its parental
                        // tasks and also changed or at least nearly changed
                        // its state to DIASBLED.
                        // We can, of course, come accross this repair at some
                        // earlier point of the nested task processing.
                        continue;
                    }
                    
                    // Otherwise the intent repair is currently not available
                    // and if the nested task is now running (it can also be
                    // disabled at all), it must be before the notifications.
                    // Therefore we can store the execution dependencies and
                    // wait for the nested task, since we will certainly be
                    // notified later on.
                    
                    // Requesting the task evaluation
                    activateInvokingStandard(nestedTask);
                    
                    // Announcing that a task has been requested
                    hasRequested = true;
                    
                    // Note also that the nested task cannot modify its
                    // dependencies container by itself, since we are
                    // currently synchronised on it
                    
                }
                
            }
            
        }
        
        // Returning the activation state
        return hasRequested;
        
    }
    
    
    /**
     * Activates the particular invoked task.
     * Only for INVOKING_MULTIPLE approach of DEFAULT and EXPLORE strategies.
     * The caller must ensure the correct synchronisation.
     * Note that the nested handler is created and activated only if it
     * is really required, but the dependencies are altered anyway.
     * @param nestedTask Nested task to be activated.
     */
    protected void activateInvokingStandard(Task nestedTask) {
        
        // Accessing execution dependencies
        Dependencies thisDeps = this.associatedTask.getExecutionDependencies();
        Dependencies nestedDeps = nestedTask.getExecutionDependencies();

        // Adding the mutual dependency
        thisDeps.addRequestedTask(nestedTask);
        nestedDeps.addRequestingTask(this.associatedTask);

        // Activating currently disabled task
        if (nestedDeps.hasStateDisabled()) {
            
            // Activating the nested task
            nestedDeps.setExecutionState(Dependencies.State.ACTIVATED);
            
            // Creating the handler instance
            Handler nestedHandler = this.correctorReference.
                createHandlerInstance(nestedTask);
            
            // Requesting the handler execution
            this.correctorReference.requestHandlerExecution(nestedHandler);
            
        }

        // Otherwise the task is already active
        
    }
    
    
    /**
     * Fetches evaluated repairs for the given set of wrappers.
     * Only for INVOKING_MULTIPLE approach of DEFAULT and EXPLORE strategies.
     * If the given repair is stored in the wrapper, it is returned directly.
     * Otherwise it is fetched from the given nested task.
     * The repairs order in the output container corresponds to the input.
     * @param nestedWrappers List of all nested wrappers to be processed.
     * @return Container with all nested repair structure references.
     * The order of repairs corresponds to the input container.
     */
    private List<Repair> doWrappersUnfolding(List<Wrapper> nestedWrappers) {
        
        // Preparing the output container
        List<Repair> outputList = new ArrayList<Repair>(nestedWrappers.size());
        
        // Fetching all nested repairs
        for (Wrapper nestedWrapper : nestedWrappers) {
            
            // Accessing the repair in the wrapper
            Repair nestedRepair = nestedWrapper.repairReference;
            
            // Accessing the repair in the task
            if (nestedRepair == null) {
                nestedRepair = nestedWrapper.taskReference.getIntentRepair();
                assert (nestedWrapper.taskReference.hasCompletedRapair());
            }
            
            // Adding the repair into the output
            outputList.add(nestedRepair);
            
        }
        
        // Returning the accessed repairs
        return outputList;
        
    }
    
    
    /**
     * Waits for the execution of the entire set of  selected nested tasks.
     * Only for INVOKING_MULTIPLE approach of DEFAULT and EXPLORE strategies.
     * The passed monitor should be the task which wants to start waiting!
     * Note that his method must be called from within the synchronisation
     * block over the calling associated task!
     */
    private void waitUniversalMultiple() {
        
        // This task is the parental one is waiting for the nested ones
        
        // Synchronising on the parental task
        synchronized (this.associatedTask) {
            try {
                // Accessing the dependencies container
                Dependencies thisDependencies =
                    this.associatedTask.getExecutionDependencies();
                // Avoiding spurious but possible wakeups
                while (thisDependencies.hasRequestedTasks()) {
                    this.associatedTask.wait();
                }
            } catch (InterruptedException ex) {
                throw new Error(
                    "Waiting for the invoked executions interrupted!"
                );
            }
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Requests the invoked refinement execution for a set of tasks.
     * Relevant for the INVOKING_MULTIPLE approach of the REFINE strategy.
     * @param nestedTasks List of tasks to be requested.
     * @param grantedQuotas List of associated granted quotas.
     */
    protected void requestInvokingMultiple(
        List<Task> nestedTasks,
        List<Quota> grantedQuotas
    ) {
        
        // Synchronising on the associated task
        synchronized (this.associatedTask) {
            
            // Initialising the requests flag
            boolean hasRequested = false;
            
            // Successively processing tasks
            for (int index = 0; index < nestedTasks.size(); index++) {
                
                // Referencing the task items
                Task nestedTask = nestedTasks.get(index);
                Quota nestedQuota = grantedQuotas.get(index);
                
                // Requesting the nested execution
                synchronized (nestedTask) {
                    
                    // Requesting the task evaluation
                    activateInvokingIterative(nestedTask, nestedQuota);
                    
                    // Announcing that a task has been requested
                    hasRequested = true;
                    
                }
                
            }

            // Waiting for requested executions
            if (hasRequested) {
                waitUniversalMultiple();
            }
            
        }
        
    }
    
    
    /**
     * Activates the particular invoked task.
     * Only for the INVOKING_MULTIPLE approach of the REFINE strategy.
     * @param nestedTask Nested task to be activated.
     * @param grantedQuota Prepared granted quota description.
     */
    protected void activateInvokingIterative(
        Task nestedTask,
        Quota grantedQuota
    ) {
        
        // Accessing execution dependencies
        Dependencies thisDeps = this.associatedTask.getExecutionDependencies();
        Dependencies nestedDeps = nestedTask.getExecutionDependencies();

        // Adding the mutual dependency
        thisDeps.addRequestedTask(nestedTask);
        nestedDeps.addRequestingTask(this.associatedTask);

        // Activating currently disabled task
        if (nestedDeps.hasStateDisabled()) {
            
            // Activating the nested task
            nestedDeps.setExecutionState(Dependencies.State.ACTIVATED);
            
            // Requesting the handler execution
            triggerInvokingSingle(nestedTask, grantedQuota);
            
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Fetches the required repair or requests its scheduled execution.
     * Only for SCHEDULING_SINGLE approach of DEFAULT and EXPLORE strategies.
     * The cache is considered when signatures handling is enabled.
     * Note that the repair reference is returned only if the repair
     * already exists in the cache and in this case the caller must
     * store the fetched repair reference immediately!
     * Otherwise the execution of the nested task is requested, the
     * requests container is updated accordingly and no repair is
     * returned at this time.
     * @param requestsData Shortcut to the requests data container.
     * @param nestedIntent Nested correction intent to be evaluated.
     * @param nestedEdge Nested edge associated with the given intent.
     * @return Reference to the existing repair or <code>null</code>.
     */
    protected Repair acquireSchedulingSingle(
        Requests requestsData,
        Intent nestedIntent,
        Edge nestedEdge
    ) {
        
        // Fetching the repair or task
        Wrapper nestedWrapper =
            this.correctorReference.provideSimpleWrapper(nestedIntent);
        
        // Returning the cached repair
        if (nestedWrapper.repairReference != null) {
            return nestedWrapper.repairReference;
            // The caller must store the repair reference immediately!
        }
        
        // Accessing the task reference
        Task nestedTask = nestedWrapper.taskReference;
        
        // Note that we do not need to use task dependencies and more
        // complicated synchronisation, since only one request can be
        // invoked at a time. This means that we do not need to be aware
        // of creating more different task instances for the same intent
        // signature, nor we have to avoid multiple invocations of the
        // same task
        
        // Activating the nested task
        scheduleStandardSimple(requestsData, nestedTask, nestedEdge);
        
        // Returning the null reference
        return null;
        
    }
    
    
    /**
     * Activates the particular scheduled task.
     * Only for SCHEDULING_SINGLE approach of DEFAULT and EXPLORE strategies.
     * @param requestsData Shortcut to the requests container.
     * @param nestedTask Nested task to be activated.
     * @param nestedEdge Reference to the associated edge.
     */
    protected void scheduleStandardSimple(
        Requests requestsData,
        Task nestedTask,
        Edge nestedEdge
    ) {
        
        // Avoiding very fast execution
        synchronized (this.associatedTask) {
            
            // Creating the handler instance
            Handler nestedHandler =
                this.correctorReference.createHandlerInstance(nestedTask);

            // Passing execution parameters
            nestedTask.setSchedulingMaster(Boolean.FALSE);
            nestedTask.setRequestingTask(this.associatedTask);

            // Updating the requests container
            requestsData.requestedTasks.add(nestedTask);
            requestsData.requestedEdges.add(nestedEdge);

            // Requesting the handler execution
            this.correctorReference.requestHandlerExecution(nestedHandler);

        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Requests the scheduled refinement execution for a set of tasks.
     * Relevant for the SCHEDULING_SINGLE approach of the REFINE strategy.
     * @param nestedTasks List of tasks to be requested.
     * @param grantedQuotas List of associated granted quotas.
     */
    protected void requestSchedulingSingle(
        List<Task> nestedTasks,
        List<Quota> grantedQuotas
    ) {
        
        // Shortcut to the requests container
        Requests requestsData = this.associatedTask.getRequestsContainer();
        
        // Marking all edges as prepared
        requestsData.preparedTasks.addAll(nestedTasks);
        requestsData.preparedQuotas.addAll(grantedQuotas);
        
        // Processing some (if any) nested prepared intent
        processPreparedRefinements(requestsData);
        
    }
    
    
    /**
     * Activates the particular scheduled task.
     * Relevant for the SCHEDULING_SINGLE approach of the REFINE strategy.
     * @param requestsData Shortcut to the requests container.
     * @param nestedTask Nested task to be activated.
     * @param grantedQuota Reference to the assigned quota.
     */
    protected void scheduleIterativeSimple(
        Requests requestsData,
        Task nestedTask,
        Quota grantedQuota
    ) {
        
        // Avoiding very fast execution
        synchronized (this.associatedTask) {
            
            // Creating the handler instance
            Handler nestedHandler =
                this.correctorReference.createHandlerInstance(nestedTask);

            // Passing execution parameters
            nestedTask.setSchedulingMaster(Boolean.FALSE);
            nestedTask.setRequestingTask(this.associatedTask);
            
            // Passing the quota parameter
            nestedTask.setQuotaData(grantedQuota);

            // Updating the requests container
            requestsData.requestedTasks.add(nestedTask);
            
            // Requesting the handler execution
            this.correctorReference.requestHandlerExecution(nestedHandler);

        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Fetches the required repairs or requests their scheduled execution.
     * Only SCHEDULING_MULTIPLE approach of DEFAULT and EXPLORE strategies.
     * The caller must store returned repair references immediately!
     * @param requestsData Shortcut to the requests data container.
     * @param nestedIntents Nested correction intents to be evaluated.
     * @param nestedEdges Nested edges associated with the intents.
     * @return List of fetched repairs or <code>null</code> references.
     * The output order of repairs is equal to the input order of intents!
     */
    protected List<Repair> acquireSchedulingMultiple(
        Requests requestsData,
        List<Intent> nestedIntents,
        List<Edge> nestedEdges
    ) {
        
        // Preparing the output list
        int itemsCount = nestedEdges.size();
        List<Repair> fetchedRepairs = new ArrayList<Repair>(itemsCount);
        
        // Synchronising on the associated task
        synchronized (this.associatedTask) {
            
            // Iterating over all provided intents
            for (int index = 0; index < itemsCount; index++) {
                
                // Accessing the nested intent
                Intent nestedIntent = nestedIntents.get(index);
                Edge nestedEdge = nestedEdges.get(index);
                
                // Fetching repair or creating task
                Wrapper nestedWrapper = this.correctorReference.
                    provideExtendedWrapper(nestedIntent);
                
                // Accessing the wrapper items
                Repair nestedRepair = nestedWrapper.repairReference;
                Task nestedTask = nestedWrapper.taskReference;
                
                // Handling fetched repair
                fetchedRepairs.add(nestedRepair);
                
                // Requesting scheduled execution
                if (
                    (nestedRepair == null) ||
                    (!nestedRepair.isCompletelyEvaluated())
                ) {
                    
                    // Requesting the nested execution
                    synchronized (nestedTask) {
                        
                        // The repair is surprisingly ready
                        if (nestedTask.hasCompletedRapair()) {
                            nestedRepair = nestedTask.getIntentRepair();
                            fetchedRepairs.set(index, nestedRepair);
                            continue;
                        }
                        
                        // Requesting the task activation
                        scheduleStandardExtended(
                            requestsData, nestedTask, nestedEdge
                        );
                        
                    }
                    
                }
                
            }
            
        }
        
        // Returning the fetched repairs
        return fetchedRepairs;
        
    }
    
    
    /**
     * Activates the particular scheduled task.
     * Only SCHEDULING_MULTIPLE approach of DEFAULT and EXPLORE strategies.
     * The caller must ensure the correct synchronisation.
     * Note that the nested handler is created and activated only if it
     * is really required, but the dependencies are altered anyway.
     * @param requestsData Shortcut to the requests container.
     * @param nestedTask Nested task to be activated.
     * @param nestedEdge Reference to the associated edge.
     */
    protected void scheduleStandardExtended(
        Requests requestsData,
        Task nestedTask,
        Edge nestedEdge
    ) {
        
        // Accessing execution dependencies
        Dependencies thisDeps = this.associatedTask.getExecutionDependencies();
        Dependencies nestedDeps = nestedTask.getExecutionDependencies();

        // Adding the mutual dependency
        thisDeps.addRequestedTask(nestedTask);
        nestedDeps.addRequestingTask(this.associatedTask);
        
        // Updating the requests container
        requestsData.requestedTasks.add(nestedTask);
        requestsData.requestedEdges.add(nestedEdge);

        // Activating currently disabled task
        if (nestedDeps.hasStateDisabled()) {
            
            // Changing the processing state
            nestedDeps.setExecutionState(Dependencies.State.ACTIVATED);
            
            // Creating the handler instance
            Handler nestedHandler =
                this.correctorReference.createHandlerInstance(nestedTask);

            // Passing execution parameters
            nestedTask.setSchedulingMaster(Boolean.FALSE);
            nestedTask.setRequestingTask(this.associatedTask);

            // Requesting the handler execution
            this.correctorReference.requestHandlerExecution(nestedHandler);
            
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Requests the scheduled refinement execution for a set of tasks.
     * Relevant for the SCHEDULING_MULTIPLE approach of the REFINE strategy.
     * @param nestedTasks List of tasks to be requested.
     * @param grantedQuotas List of associated granted quotas.
     */
    protected void requestMultipleScheduling(
        List<Task> nestedTasks,
        List<Quota> grantedQuotas
    ) {
        
        // Preparing auxiliary items
        int itemsCount = nestedTasks.size();
        Requests requestsData = this.associatedTask.getRequestsContainer();
        
        // Synchronising on the associated task
        synchronized (this.associatedTask) {
            
            // Iterating over all provided intents
            for (int index = 0; index < itemsCount; index++) {
                
                // Accessing the nested intent
                Task nestedTask = nestedTasks.get(index);
                Quota grantedQuota = grantedQuotas.get(index);
                Repair nestedRepair = nestedTask.getIntentRepair();
                
                // Requesting scheduled execution
                if (!nestedRepair.isCompletelyEvaluated()) {
                    
                    // Requesting the nested execution
                    synchronized (nestedTask) {
                        
                        // Requesting the task activation
                        scheduleIterativeExtended(
                            requestsData, nestedTask, grantedQuota
                        );
                        
                    }
                    
                }
                
            }
            
        }
        
    }
    
    
    /**
     * Activates the particular scheduled task.
     * Only relevant for SCHEDULING_MULTIPLE approach of REFINE strategy
     * @param requestsData Shortcut to the requests container.
     * @param nestedTask Nested task to be activated.
     * @param grantedQuota Reference to the assigned quota.
     */
    protected void scheduleIterativeExtended(
        Requests requestsData,
        Task nestedTask,
        Quota grantedQuota
    ) {
        
        // Accessing execution dependencies
        Dependencies thisDeps = this.associatedTask.getExecutionDependencies();
        Dependencies nestedDeps = nestedTask.getExecutionDependencies();

        // Adding the mutual dependency
        thisDeps.addRequestedTask(nestedTask);
        nestedDeps.addRequestingTask(this.associatedTask);
        
        // Updating the requests container
        requestsData.requestedTasks.add(nestedTask);

        // Activating currently disabled task
        if (nestedDeps.hasStateDisabled()) {
            
            // Changing the processing state
            nestedDeps.setExecutionState(Dependencies.State.ACTIVATED);
            
            // Creating the handler instance
            Handler nestedHandler =
                this.correctorReference.createHandlerInstance(nestedTask);
            
            // Passing execution parameters
            nestedTask.setSchedulingMaster(Boolean.FALSE);
            nestedTask.setRequestingTask(this.associatedTask);
            
            // Passing the quota parameter
            nestedTask.setQuotaData(grantedQuota);

            // Requesting the handler execution
            this.correctorReference.requestHandlerExecution(nestedHandler);
            
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Sends required notifications at the end of the handler job execution.
     * Works for all execution approaches of all algorithm strategies.
     */
    public void dispatchEndingNotifications() {
        
        // Announcing the execution end
        switch (this.correctorReference.getExecutionApproach()) {
            case NESTING_SINGLE:
                break;
            case INVOKING_SINGLE:
                notifyInvokedSingle();
                break;
            case INVOKING_MULTIPLE:
                notifyInvokedMultiple();
                break;
            case SCHEDULING_SINGLE:
                notifyScheduledSingle();
                break;
            case SCHEDULING_MULTIPLE:
                notifyScheduledMultiple();
                break;
            default:
                throw new UnsupportedOperationException();
        }
        
    }
    
    
    /**
     * Notifies the only parental task about the completed invoked execution.
     * 
     * This method can only be used by the INVOKING_SINGLE approach.
     * 
     * It should be called by the nested task at the end of its execution,
     * i.e. after it has published the evaluated repair in case of both
     * the DEFAULT and EXPLORATION strategies or after it has completed
     * one refinement step in case of the IRERATIVE strategy, in order to
     * notify the only parental task to stop its waiting stage and resume
     * its work again.
     */
    protected void notifyInvokedSingle() {
        
        // This task is the nested one and wants to notify its parent
        Task nestedTask = this.associatedTask;
        
        // Synchronising on the nested task
        synchronized (nestedTask) {
            
            // Marking the refinement flag
            nestedTask.markCompletedJob();
            // Notifying the parental handler
            nestedTask.notifyAll();
            
        }
        
    }
    
    
    /**
     * Notifies all the parental tasks about the completed invoked execution.
     * 
     * This method can only be used by the INVOKING_MULTIPLE approach.
     * 
     * It should be called by the nested task at the end of its execution,
     * i.e. after it has published the evaluated repair in case of both
     * the DEFAULT and EXPLORATION strategies or after it has completed
     * one refinement step in case of the IRERATIVE strategy, in order to
     * notify all its parental tasks (at least one must always exists)
     * to stop their waiting stage and resume their work again (which
     * can of course happen only when the particular parental task is
     * not waiting for any other requested task).
     */
    protected void notifyInvokedMultiple() {
        
        // This task is the nested one and wants to notify its parent(s)
        
        synchronized (this.associatedTask) {
            // This empty synchronisation block ensures that the procedure
            // for requesting and activating nested tasks behaves correctly.
            // Once this point is passed, the nested task (it means this one)
            // must already have stored the reference to the completely
            // evaluated repair and at the same time the stage of parental
            // tasks notification itself is ahead.
        }
        
        // Despite the access to the container of parental tasks is not
        // covered by any synchronisation, we do not need to worry. It is
        // because (according to the previous synchronisation block) the
        // repair is already cached and, therefore, no one else has any
        // reason to request this task.
        
        // Also note that the order of the following synchronisation blocks
        // cannot be reverser, although it might seem that doing it would
        // lead to increasing efficiency. This is true, but, unfortunately,
        // the correct order is very important in order to breach Coffman's
        // conditions and avoid the possibility of deadlocks!
        
        // Marking the refinement flag
        this.associatedTask.markCompletedJob();
        
        // Accessing the dependencies container
        Dependencies thisDependencies =
            this.associatedTask.getExecutionDependencies();
        
        // Iterating over all parental tasks
        while (thisDependencies.hasRequestingTasks()) {
            
            // Fecthing some parental task
            Task parentalTask = thisDependencies.getRequestingTask();
        
            // Synchronising on the parental task
            synchronized (parentalTask) {
                // Synchronising on this nested task
                synchronized (this.associatedTask) {
                    
                    // Updating mutual dependencies
                    thisDependencies.removeRequestingTask(parentalTask);
                    parentalTask.getExecutionDependencies().
                        removeRequestedTask(this.associatedTask);
                    
                    // Notifying the parental task
                    parentalTask.notifyAll();
                    
                }
            }
        
        }
        
        // Disabling the execution state
        thisDependencies.setExecutionState(Dependencies.State.DISABLED);
        
        // This state change does not need to be synchronised since when
        // we have already published the evaluated repair, no one will be
        // interrested in it
        
    }
    
    
    /**
     * Notifies the only parental task about the completed scheduled execution.
     * 
     * This method can only be used by the SCHEDULING_SINGLE approach.
     */
    protected void notifyScheduledSingle() {
        
        // Accessing the parental task
        Task parentalTask = this.associatedTask.getRequestingTask();
        
        // Notifying the parental task
        if (parentalTask != null) {
            
            // Avoiding very fast notification
            synchronized (parentalTask) {
                
                // Marking the refinement flag
                this.associatedTask.markCompletedJob();
                
                // Creating the handler instance
                Handler parentHandler = this.correctorReference.
                    createHandlerInstance(parentalTask);

                // No parameters are passed!

                // Requesting the handler execution
                this.correctorReference.requestHandlerExecution(parentHandler);
                
            }
            
        }
        
        // Terminating the starting intent
        if (this.associatedTask.getSchedulingMaster()) {
            synchronized (this.correctorReference) {
                this.correctorReference.notify();
            }
        }
        
    }
    
    
    /**
     * Notifies all the parental tasks about the completed scheduled execution.
     * 
     * This method can only be used by the SCHEDULING_MULTIPLE approach.
     */
    protected void notifyScheduledMultiple() {
        
        synchronized (this.associatedTask) { }
        
        // Marking the refinement flag
        this.associatedTask.markCompletedJob();
        
        // Accessing the dependencies container
        Dependencies thisDeps =
            this.associatedTask.getExecutionDependencies();
        
        // Iterating over all parental tasks
        while (thisDeps.hasRequestingTasks()) {
            
            // Fecthing some parental task
            Task parentalTask = thisDeps.getRequestingTask();
            
            // Synchronising on the parental task
            synchronized (parentalTask) {
                // Synchronising on the nested task
                synchronized (this.associatedTask) {
                    
                    // Accessing parental dependencies
                    Dependencies parentDeps =
                        parentalTask.getExecutionDependencies();
                    
                    // Updating mutual dependencies
                    thisDeps.removeRequestingTask(parentalTask);
                    parentDeps.removeRequestedTask(this.associatedTask);
                    
                    // Parental task can be reactivated
                    if (parentDeps.hasNoRequestedTasks()){
                        
                        // Creating the handler instance
                        Handler parentHandler = this.correctorReference.
                            createHandlerInstance(parentalTask);

                        // No parameters are passed!

                        // Requesting the handler execution
                        this.correctorReference.
                            requestHandlerExecution(parentHandler);
                    
                    }
                    
                }
            }
        
        }
        
        // In case of the starting correction intent, the previous set
        // of requesting tasks is empty. We must notify the corrector
        // thread directly.
        
        // Terminating the starting intent
        if (this.associatedTask.getSchedulingMaster()) {
            synchronized (this.correctorReference) {
                this.correctorReference.notify();
            }
        }
        
        // Disabling the execution state
        thisDeps.setExecutionState(Dependencies.State.DISABLED);
        
    }
    
    
    
    
    // =========================================================================
    // =========================================================================
    // =========================================================================
    
    
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Fetches the existing variables container or creates a new one.
     * @return Reference to the newly created or existing container.
     */
    private Variables fetchWorkingVariables() {
        
        // Attempting to access existing container
        Variables workingVars =
            (Variables)this.associatedTask.getWorkingVariables();
        
        // Creating and registering a new container
        if (workingVars == null) {
            workingVars = createWorkingVariables();
        }
        
        // Returning the container reference
        return workingVars;
        
    }
    
    /**
     * Creates a new container of working variables and registers it.
     * Note that the requested container is created always regardless
     * there might already exists another registered container!
     * @return Newly created and registered working variables.
     */
    private Variables createWorkingVariables() {
        
        // Creating and registering a new container
        Variables workingVars = new Variables();
        this.associatedTask.setWorkingVariables(workingVars);
        
        // Returning the container reference
        return workingVars;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prepares the quota object to be assigned to the given task.
     * Only relevant for the REFINE algorithm strategy.
     * @param nestedEdge Reference to the inspected edge.
     * @param allowedRise Granted growth of the edge cost.
     * @return Created quota description instance.
     */
    protected Quota resolveGrantedQuota(
        Edge nestedEdge,
        int allowedRise
    ) {
        return Quota.createLimitedQuota(
            nestedEdge.getEdgeCost() + allowedRise -
            nestedEdge.getIntentRepair().getNodeRepair().getRepairCost()
        );
    }
    
    
    /**
     * Detects whether the iterative processing can continue.
     * This means that the granted quota has not yet been fully used.
     * Only relevant for the REFINE algorithm strategy.
     * @return Returns <code>true</code> when further processing is allowed.
     */
    protected boolean hasExceededQuota() {
        return
            !this.associatedTask.getQuotaData().hasAllowedProcessing(
                this.associatedTask.getWorkingVariables().getReachedCost()
            );
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns some reached vertex from the set of reached vertices.
     * The returned vertex is not removed from the reached vertices.
     * Only suitable for the DEFAULT algorithm strategy!
     * @return Reference to some reached vertex.
     */
    protected Vertex selectAnyVertex() {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        
        // Returning tome reached vertex
        return workingVars.getReachedVertex();
        
    }
    
    
    /**
     * Returns the first reached vertex with the lowest cost.
     * The returned vertex is not removed from the reached vertices.
     * Only suitable for the EXPLORE algorithm strategy!
     * In case of the latter one, only cost estimations are used!
     * The reached vertices container in working variables is considered.
     * @return Reached vertex considered as the one with lowest cost.
     */
    protected Vertex selectCheapestVertex() {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        Graph correctionGraph = workingVars.getCorrectionGraph();
        
        // Preparing auxiliary variables
        Integer selectedCost = null;
        Vertex selectedVertex = null;
        
        // Searching for the minimal cost
        for (Vertex reachedVertex : workingVars.getReachedVertices()) {
            Integer inspectedCost =
                correctionGraph.getTracingCost(reachedVertex);
            if ((selectedCost == null) || (selectedCost > inspectedCost)) {
                selectedCost = inspectedCost;
                selectedVertex = reachedVertex;
            }
        }
        
        // Returning the selected vertex
        return selectedVertex;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Selects the most promising reached vertex by exploring their reports.
     * Note that the returned value is based on the current knowledge
     * about nested intents and their refinement progress. Therefore,
     * the caller must be aware of the fact that some other nested intents
     * might already or can be updated some time later on.
     * Only suitable for the REFINE algorithm strategy.
     * @return Full report for the most promising selected reached vertex.
     */
    protected Record selectPromisingReport() {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        
        // Initialising iteration variables
        Integer fixedCost = null;
        Record fixedReport = null;
        
        // Iterating over all reached vertices
        for (Vertex reachedVertex : workingVars.getReachedVertices()) {
            
            // Computing the selected report
            Record thisReport = resolveVertexReport(reachedVertex);
            int thisCost = thisReport.generalMinimum;
            
            // Replacing the best found candidate
            if ((fixedCost == null) || (fixedCost > thisCost)) {
                fixedCost = thisCost;
                fixedReport = thisReport;
            }
            
        }
        
        // Returning the best candidate
        return fixedReport;
        
    }
    
    
    /**
     * Computes the full report for the given reached vertex.
     * Data in the report can be used to detect completeness of the given
     * reached vertex and to further process it regardless it is considered
     * as complete or incomplete.
     * Note that tracing records about shortest paths stored directly in the
     * associated multigraph are always preserved untouched! However, tracing
     * data are precomputed and stored within the report, so when processing
     * the complete vertex they can be easily accessed and used immediately.
     * @param inspectedVertex Reached vertex to be inspected and updated.
     * @return Fully evaluated report for this particular reached vertex.
     */
    protected Record resolveVertexReport(Vertex inspectedVertex) {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        Graph correctionGraph = workingVars.getCorrectionGraph();
        
        // Creating the report object
        Record reportData = new Record(inspectedVertex);
        
        // Fetching all ingoing edges
        Set<Edge> inputEdges = correctionGraph.getIngoingEdges(inspectedVertex);
        
        // Returning if no ingoing edges
        if (inputEdges == null) {
            reportData.closedMinimum = 0;
            reportData.generalMinimum = 0;
            reportData.tracingCost = 0;
            return reportData;
        }
        
        // Preparing auxiliary containers
        int items = inputEdges.size();
        List<Edge> dataEdges = new ArrayList<Edge>(inputEdges);
        List<Boolean> dataFlags = new ArrayList<Boolean>(items);
        List<Integer> dataCosts = new ArrayList<Integer>(items);
        
        // The first ingoing edges traversal
        for (int index = 0; index < items; index++) {
            
            // Accessing the given edge
            Edge inspectedEdge = dataEdges.get(index);
            
            // Resolving the current progress
            boolean isCompleted = inspectedEdge.getIntentRepair().
                isCompletelyEvaluated();
            int inspectedCost = correctionGraph.getResolvedCost(inspectedEdge);
            
            // Storing fetched values
            dataFlags.add(isCompleted);
            dataCosts.add(inspectedCost);
            
            // Note that both these records may already have changed now!
            // It is therefore imporant to first obtain the completeness
            // state and then the cost, because when an edge declares that
            // it is already complete, this state and the cost as well
            // cannot be further changed. So we are sure that the complete
            // edge remains to be complete. On the other hand, even if we
            // come acroos some edge that we think should be open (and it
            // is no longer, or it is, but the cost increased) we cannot
            // cause any problems.
            
            // Processing completed edge
            if (isCompleted) {
                
                if (
                    (reportData.closedMinimum == null) ||
                    (reportData.closedMinimum > inspectedCost)
                ) {
                    reportData.closedMinimum = inspectedCost;
                    reportData.tracingCost = inspectedCost;
                }
                
            // Processing incomplete edge
            } else {
                
                if (
                    (reportData.openMinimum == null) ||
                    (reportData.openMinimum > inspectedCost)
                ) {
                    reportData.quotaBoundary = reportData.openMinimum;
                    reportData.openMinimum = inspectedCost;
                }
                
            }
            
        }
        
        // Selecting the overall minimum
        reportData.generalMinimum = reportData.openMinimum;
        if (reportData.closedMinimum != null) {
            if (
                (reportData.openMinimum == null) ||
                (reportData.closedMinimum < reportData.openMinimum)
            ) {
                reportData.generalMinimum = reportData.closedMinimum;
            }
        }
        
        // The second ingoing edges traversal
        for (int index = 0; index < items; index++) {
            
            // Accessing stored records
            Edge inspectedEdge = dataEdges.get(index);
            boolean isComplete = dataFlags.get(index);
            int inspectedCost = dataCosts.get(index);
            
            // Processing completed edge
            if (isComplete) {
                
                if (inspectedCost == reportData.tracingCost) {
                    reportData.tracingEdges.add(inspectedEdge);
                }
                reportData.uselessEdges.add(inspectedEdge);
            
            // Processing incomplete edge
            } else {
                
                if (reportData.tracingCost != null) {
                    if (inspectedCost == reportData.tracingCost) {
                        reportData.delayedEdges.add(inspectedEdge);
                    } else {
                        reportData.uselessEdges.add(inspectedEdge);
                    }
                }
                if (inspectedCost == reportData.openMinimum) {
                    reportData.promisingEdges.add(inspectedEdge);
                }
                
            }
            
        }
        
        // Returning the vertex report
        return reportData;
        
    }
    
    
    /**
     * Detects whether the given reached vertex can be completed.
     * A reached vertex is complete if there exists at least one closed
     * ingoing edge and all other open existing edges (if any) have higher
     * tracing cost estimations.
     * @param vertexReport Record data prepared for the inspected vertex.
     * @return Returns <code>true</code> in case the given vertex is complete.
     */
    public boolean isVertexComplete(
        Record vertexReport
    ) {
        
        // There are no ingoing edges
        if (
            (vertexReport.openMinimum == null) &&
            (vertexReport.closedMinimum == null)
        ) {
            return true;
        }
        // No ingoing edge is yet closed
        if (vertexReport.closedMinimum == null) {
            return false;
        }
        // All ingoing edges are already closed
        if (vertexReport.openMinimum == null) {
            return true;
        }
        // All open edges are worse then the best closed
        if (vertexReport.openMinimum >= vertexReport.closedMinimum) {
            return true;
        }
        // Some open edge is still perspective
        return false; 
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Detects if the final cost should be fixed during the exploration.
     * 
     * Only suitable for the EXPLORE and REFINE algorithm strategies.
     * 
     * @param selectedVertex Currently selected reached vertex.
     * @return Returns <code>true</code> if the final cost should be fixed.
     * 
     * The cost should be fixed if we (during the multigraph exploration)
     * come across the first target vertex (i.e. some vertex from the last
     * stratum with an accepting state from the corresponding automaton).
     */
    protected boolean shouldFixateCost(Vertex selectedVertex) {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        Intent parentIntent = this.associatedTask.getCorrectionIntent();
        
        // Shortcuts to vertex items
        int stratumNumber = selectedVertex.getStratumNumber();
        String automatonState = selectedVertex.getAutomatonState();
        
        // Computing the decision
        return
            (workingVars.getFixedCost() == null) &&
            (stratumNumber == parentIntent.getLastStratum()) &&
            (parentIntent.getTargetStates().contains(automatonState));
        
    }
    
    
    /**
     * Detects whether the exploration can be already terminated.
     * 
     * Only suitable for the EXPLORE and REFINE algorithm strategies.
     * 
     * @param reachedCost Current value of the reached exploration cost.
     * @return Returns <code>true</code> if the exploration should be ended.
     */
    protected boolean shouldQuitExploring(int reachedCost) {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        
        // Computing the decision
        return
            (workingVars.getFixedCost() != null) &&
            (workingVars.getFixedCost() < reachedCost);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates and initialises the correction multigraph.
     * 
     * In particular, a new correction multigraph is created and registered
     * with the working variables, then the source vertex is created, added
     * into the graph and inserted into the list of reached vertices, which
     * enables the graph exploration phase.
     * 
     * Also the execution phase is changed to EXPLORE.
     * 
     * @return Reference to the created and registered graph.
     */
    protected Graph initCorrectionGraph() {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        Intent parentIntent = this.associatedTask.getCorrectionIntent();
        Repair intentRepair = this.associatedTask.getIntentRepair();
        
        // Creating an empty correction multigraph
        Graph correctionGraph = new Graph(
            parentIntent.getTargetStates(),
            parentIntent.getSequenceLength()
        );
        
        // Registering the created graph
        workingVars.setCorrectionGraph(correctionGraph);
        intentRepair.registerSequenceRepair(correctionGraph);
        
        // Creating the source vertex
        Vertex sourceVertex = new Vertex(0, parentIntent.getSourceState());
        correctionGraph.addNewVertex(sourceVertex);
        
        // Profiling the vertex creation
        if (hasStatsProfiling()) {
            accessProfilingReport().increaseVerticesCounter();
        }
        
        // Setting shortest paths records
        if (hasStrategy(EXPLORE) || hasStrategy(REFINE)) {
            correctionGraph.setTracingCost(sourceVertex, 0);
            correctionGraph.initTracingEdges(sourceVertex);
            workingVars.setReachedCost(0);
        }
        
        // Setting the total cost estimate
        if (hasStrategy(REFINE)) {
            correctionGraph.setTermCost(0);
        }
        
        // Marking the vertex as reached
        workingVars.addReachedVertex(sourceVertex);
        
        // Upgrading the execution phase
        workingVars.setExecutionPhase(Phase.EXPLORING);
        
        // Returning the graph reference
        return correctionGraph;
        
    }
    
    
    /**
     * Finds all shortest paths in the provided and fully explored multigraph.
     * 
     * This function is only used in case of the DEFAULT strategy.
     * 
     * However, standard variables container and its components for
     * storing reached vertices and fixed cost are used.
     * Note that this function does not affect the termination cost
     * of the entire correction multigraph!
     */
    protected void processPathsSearching() {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        Intent parentIntent = this.associatedTask.getCorrectionIntent();
        Graph correctionGraph = workingVars.getCorrectionGraph();
        
        // Processing the source vertex
        Vertex sourceVertex = new Vertex(0, parentIntent.getSourceState());
        workingVars.addReachedVertex(sourceVertex);
        
        // Initialising paths records
        correctionGraph.setTracingCost(sourceVertex, 0);
        correctionGraph.initTracingEdges(sourceVertex);
        
        // Searching for the shortest paths
        while (workingVars.hasReachedVertices()) {
            
            // Selecting reached vertex with minimal cost
            Vertex inspectedVertex = selectCheapestVertex();
            workingVars.removeReachedVertex(inspectedVertex);
            Integer reachedCost =
                correctionGraph.getTracingCost(inspectedVertex);
            workingVars.setReachedCost(reachedCost);
            
            // Detecting the cost fixation
            if (shouldFixateCost(inspectedVertex)) {
                // Fixing the current cost
                workingVars.setFixedCost(reachedCost);
            }

            // Detecting the interruption
            if (shouldQuitExploring(reachedCost)) {
                // Clearing all reached vertices
                workingVars.clearReachedVertices();
                // Terminating the exploration
                break;
            }
            
            // Processing all outgoing edges
            Set<Edge> outgoingEdges =
                correctionGraph.getOutgoingEdges(inspectedVertex);
            if (outgoingEdges != null) {
                for (Edge outgoingEdge : outgoingEdges) {
                    // Reaching new vertex
                    Vertex endingVertex = outgoingEdge.getTargetVertex();
                    if (correctionGraph.getTracingCost(endingVertex) == null) {
                        workingVars.addReachedVertex(endingVertex);
                    }
                    // Resolving path records
                    updatePathsTracing(outgoingEdge);
                }
            }
            
        }
        
    }
    
    
    /**
     * Updates records about paths to the target vertex of the provided edge.
     * Note that this function does not manipulate with reached vertices.
     * @param inspectedEdge Already evaluated edge to be inspected.
     */
    protected void updatePathsTracing(Edge inspectedEdge) {
        
        // Accessing required variables
        Variables workingVars = this.associatedTask.getWorkingVariables();
        Graph correctionGraph = workingVars.getCorrectionGraph();
        
        // Accessing edge vertices
        Vertex sourceVertex = inspectedEdge.getSourceVertex();
        Vertex targetVertex = inspectedEdge.getTargetVertex();
        
        // Resolving old and new costs
        Integer sourceCost = correctionGraph.getTracingCost(sourceVertex);
        Integer targetCost = correctionGraph.getTracingCost(targetVertex);
        Integer inspectedCost = sourceCost + inspectedEdge.getEdgeCost();
        
        // Processing already reached ending vertex
        if (targetCost != null) {
            if (targetCost == inspectedCost) {
                correctionGraph.addTracingEdge(targetVertex, inspectedEdge);
            }
            if (targetCost > inspectedCost) {
                correctionGraph.setTracingCost(targetVertex, inspectedCost);
                correctionGraph.clearTracingEdges(targetVertex);
                correctionGraph.addTracingEdge(targetVertex, inspectedEdge);
            }
        
        // Processing newly reached ending vertex
        } else {
            correctionGraph.setTracingCost(targetVertex, inspectedCost);
            correctionGraph.addTracingEdge(targetVertex, inspectedEdge);
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the associated correction task.
     * @return Reference to the associated task instance.
     */
    public Task getAssociatedTask() {
        return this.associatedTask;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Tests whether the given algorithm strategy is activated.
     * @param selectedStrategy Algorithm strategy to be tested.
     * @return Returns <code>true</code> if the strategy corresponds.
     */
    protected boolean hasStrategy(
        Corrector.AlgorithmStrategy selectedStrategy
    ) {
        return (this.correctorReference.hasStrategy(selectedStrategy));
    }
    
    /**
     * Tests whether the given execution approach is activated.
     * @param selectedApproach Execution approach to be tested.
     * @return Returns <code>true</code> if the approach corresponds.
     */
    protected boolean hasApproach(
        Corrector.ExecutionApproach selectedApproach
    ) {
        return (this.correctorReference.hasApproach(selectedApproach));
    }
    
    /**
     * Tests whether the given signatures handling mode is activated.
     * @param selectedMode Signature handling mode to be tested.
     * @return Returns <code>true</code> if the mode corresponds.
     */
    protected boolean hasSignatures(
        Corrector.SignatureHandling selectedMode
    ) {
        return (this.correctorReference.hasSignatures(selectedMode));
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the shortcut to the profiling report.
     * @return Reference to the profiling report.
     */
    protected Report accessProfilingReport() {
        return this.correctorReference.profilingReport;
    }
    
    
    /**
     * Tests whether execution times profiling is enabled.
     * @return Returns <code>true</code> when time profiling active.
     */
    protected boolean hasTimesProfiling() {
        return (this.correctorReference.profilingTimes == true);
    }
    
    /**
     * Tests whether exploration statistics profiling is enabled.
     * @return Returns <code>true</code> when statistics profiling active.
     */
    protected boolean hasStatsProfiling() {
        return (this.correctorReference.profilingStats == true);
    }
    
    
    // -------------------------------------------------------------------------
    
}
