
package cz.cuni.mff.corrector.edit;

import cz.cuni.mff.corrector.instruction.Instruction;
import cz.cuni.mff.corrector.tree.Position;
import java.util.ArrayList;
import java.util.List;

/**
 * Abstract ancestor for edit operations.
 *
 * @author Martin Svoboda
 */
public abstract class Edit {

    // -------------------------------------------------------------------------


    /**
     * Position in a data tree the operation should be performed at.
     */
    protected Position basePosition;


    // -------------------------------------------------------------------------


    /**
     * Initializes the common items of the edit operation structure.
     * @param basePosition Base position in a data tree.
     */
    protected Edit(Position basePosition) {
        this.basePosition = basePosition;
    }
    

    // -------------------------------------------------------------------------


    /**
     * Returns the untyped reference to the repairing instruction.
     * @return Associated repairing instruction.
     */
    public abstract Instruction getRepairingInstruction();


    /**
     * Returns the type of the edit operation.
     * @return Enumeration type of the associated repairing instruction.
     */
    public Instruction.Type getOperationType() {
        return this.getRepairingInstruction().getInstructionType();
    }


    /**
     * Returns the cost of the edit operation.
     * @return Associated cost of the repairing instruction.
     */
    public int getOperationCost() {
        return this.getRepairingInstruction().getInstructionCost();
    }


    /**
     * Returns the base position of the edit operation.
     * @return Base position.
     */
    public Position getBasePosition() {
        return this.basePosition;
    }


    /**
     * Returns the human readable name of the operation.
     * @return Edit operation title.
     */
    public abstract String getOperationTitle();


    // -------------------------------------------------------------------------


    /**
     * Applies the modPre modification of the base position.
     * @param prefixItem Simple position to be prepended.
     */
    public void applyRootPrepending(Integer prefixItem) {
        this.basePosition.applyRootPrepending(prefixItem);
    }

    /**
     * Applies the modAlt modification of the base position.
     * @param incrementValue Value to be added to the first simple position.
     */
    public void applyRootShifting(int incrementValue) {
        this.basePosition.applyRootShifting(incrementValue);
    }

    /**
     * Applies the modCut modification of the base position.
     */
    public void applyRootCutting() {
        this.basePosition.applyRootCutting();
    }


    // -------------------------------------------------------------------------

    
    /**
     * Creates an independent clone of the edit operation.
     * Note that the reference to the instruction is preserved!
     * Only the independent clone of the position is created.
     * @return Newly created independent and identical instance.
     */
    public abstract Edit cloneIdenticalOperation();
    

    /**
     * Creates an independent clone of the edit with applied modAlt.
     * Note that the reference to the instruction is preserved!
     * Only the independent clone of the position is created.
     * @param incrementValue Shifting size to increment base positions.
     * @return Newly created independent and shifted edit operation.
     */
    public abstract Edit cloneShiftedOperation(int incrementValue);
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates an independent clone of all source operations.
     * Newly created instances are placed into a newly created container.
     * @param sourceList Input container with operations to be cloned.
     */
    public static List<Edit> cloneIdenticalSequence(
        List<Edit> sourceList
    ) {
        List<Edit> outputList = new ArrayList<Edit>();
        for (Edit oldEdit : sourceList) {
            outputList.add(oldEdit.cloneIdenticalOperation());
        }
        return outputList;
    }
    
    
    /**
     * Creates an independent clone of all source operations.
     * Newly created instances are placed into the target container.
     * @param sourceList Input container with operations to be cloned.
     * @param outputList Output container for placing created clones.
     */
    public static void cloneIdenticalSequenceAfter(
        List<Edit> sourceList,
        List<Edit> outputList
    ) {
        for (Edit oldEdit : sourceList) {
            outputList.add(oldEdit.cloneIdenticalOperation());
        }
    }
    
    
    /**
     * Creates an independent clone of all source operations.
     * The entire sequence is placed at the beginning of the output container!
     * Newly created instances are placed into the target container.
     * @param sourceList Input container with operations to be cloned.
     * @param outputList Output container for placing created clones.
     */
    public static void cloneIdenticalSequenceBefore(
        List<Edit> sourceList,
        List<Edit> outputList
    ) {
        for (int i = sourceList.size() - 1; i >= 0; i--) {
            outputList.add(0, sourceList.get(i).cloneIdenticalOperation());
        }
    }
    
    
    /**
     * Creates an independent and shifted clone of all source operations.
     * Newly created instances are placed into the target container.
     * @param sourceList Input container with operations to be cloned.
     * @param outputList Output container for placing created clones.
     * @param incrementValue Shifting size to increment base positions.
     */
    public static void cloneShiftedSequence(
        List<Edit> sourceList,
        List<Edit> outputList,
        int incrementValue
    ) {
        for (Edit oldEdit : sourceList) {
            outputList.add(oldEdit.cloneShiftedOperation(incrementValue));
        }
    }

    
    // -------------------------------------------------------------------------

}
