
package cz.cuni.mff.corrector.execution;

/**
 * Processing quota execution parameters for the iterative approach.
 * 
 * @author Martin Svoboda
 */
public class Quota {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Enumeration of both defined iterative processing modes.
     */
    public enum Mode {
        /**
         * Iterative execution goal: perform the complete correction.
         * This method is always applied on the starting intents and
         * then on all nested intents that should be evaluated fully.
         */
        UNLIMITED,
        /**
         * Iterative execution goal: only achieve further refinement.
         */
        REFINEMENT;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Selection of the iterative processing mode.
     */
    private final Mode iterativeMode;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Maximal granted boundary of the cost the can be reached.
     * This parameter is only relevant when REFINEMENT mode is active.
     */
    private final Integer grantedLimit;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new instance of the quota parameters.
     * @param iterativeMode Selection of the iterative processing mode.
     * @param grantedLimit Granted quota for iterative refinement processing.
     */
    private Quota(
        Mode iterativeMode,
        Integer grantedLimit
    ) {
        this.iterativeMode = iterativeMode;
        this.grantedLimit = grantedLimit;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new parameter instance for the limited refinement mode.
     * Note that quota should only be set during the parent-child activation.
     * @return Created quota parameter instance.
     */
    public static Quota createUnlimitedQuota() {
        return new Quota(Mode.UNLIMITED, null);
    }
    
    /**
     * Creates a new parameter instance for the limited refinement mode.
     * Note that quota should only be set during the parent-child activation.
     * @param grantedQuota Value of the granted refinement cost.
     * @return Created quota parameter instance.
     */
    public static Quota createLimitedQuota(int grantedQuota) {
        return new Quota(Mode.REFINEMENT, grantedQuota);
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the current value of limited granted quota.
     * @return Current value of limited granted quota.
     */
    public Integer getGrantedQuota() {
        return this.grantedLimit;
    }
    
    
    /**
     * Detects whether additional iterative processing is allowed.
     * @param reachedCost Current level of the achieved correction cost.
     * @return Returns <code>true</code> when further processing is permitted.
     */
    public boolean hasAllowedProcessing(int reachedCost) {
        return (this.iterativeMode == Mode.UNLIMITED) ||
            (reachedCost < this.grantedLimit);
    }
    
    
    // -------------------------------------------------------------------------
    
}
