
package cz.cuni.mff.corrector.expression;

import java.util.Map;
import java.util.Set;

/**
 * Base class for particular regular subexpressions.
 *
 * @author Martin Svoboda
 */
public abstract class Expression {
    
    // -------------------------------------------------------------------------


    /**
     * String token for opening parenthesis (must have right one character).
     */
    public static String TOKEN_MT_OPEN = "(";
    /**
     * String token for closing parenthesis (must have right one character).
     */
    public static String TOKEN_MT_CLOSE = ")";

    /**
     * String token for concatenation operator (must have right one character).
     */
    public static String TOKEN_OP_CONCAT = ".";
    /**
     * String token for union operator (must have right one character).
     */
    public static String TOKEN_OP_UNION = "|";
    /**
     * String token for Kleene star operator (must have right one character).
     */
    public static String TOKEN_OP_STAR = "*";

    /**
     * String token for empty trivial regular expression.
     * Note that in our implementation this expression acts identically
     * as TOKEN_RE_EPSILON, i.e. it corresponds to a one empty word and
     * not an empty set of words as the standard semantics suggests.
     */
    public static String TOKEN_RE_EMPTY = "#empty";
    /**
     * String token for epsilon trivial regular expression.
     */
    public static String TOKEN_RE_EPSILON = "#epsilon";

    /**
     * String token for delimiter used for marked expressions.
     */
    public static String TOKEN_MS_MARK = "#";


    // -------------------------------------------------------------------------


    /**
     * Enumeration of regular subexpression types.
     */
    public enum Type {
        /**
         * Expression type: empty expression.
         */
        EMPTY,
        /**
         * Expression type: empty word.
         */
        EPSILON,
        /**
         * Expression type: alphabet symbol.
         */
        SYMBOL,
        /**
         * Expression type: expressions concatenation.
         */
        CONCAT,
        /**
         * Expression type: expressions union.
         */
        UNION,
        /**
         * Expression type: expression iteration.
         */
        STAR;
    }


    // -------------------------------------------------------------------------


    /**
     * Creates a new expression object for a valid regular expression string.
     * @param exp String form of the regular expression.
     * @return Recursive regular expression structure.
     */
    public static Expression createExpression(String exp) {
        // Parsing trivial expressions
        if (exp.equals(TOKEN_RE_EMPTY)) {
            return new ExpressionEmpty();
        }
        if (exp.equals(TOKEN_RE_EPSILON)) {
            return new ExpressionEpsilon();
        }
        // Parsing parentheses
        if (exp.startsWith(TOKEN_MT_OPEN) && exp.endsWith(TOKEN_MT_CLOSE)) {
            return createExpression(exp.substring(1, exp.length() - 1));
        }
        // Attempting to parse union operation
        int index = parseOperand(exp, TOKEN_OP_UNION);
        if (index < exp.length()) {
            String leftOperand = exp.substring(0, index);
            Expression leftExpression = createExpression(leftOperand);
            String rightOperand = exp.substring(index + 1);
            Expression rightExpression = createExpression(rightOperand);
            return new ExpressionUnion(leftExpression, rightExpression);
        }
        // Attempting to parse concatenation operation
        index = parseOperand(exp, TOKEN_OP_CONCAT);
        if (index < exp.length()) {
            String leftOperand = exp.substring(0, index);
            Expression leftExpression = createExpression(leftOperand);
            String rightOperand = exp.substring(index + 1);
            Expression rightExpression = createExpression(rightOperand);
            return new ExpressionConcat(leftExpression, rightExpression);
        }
        // Attempting to parse Kleene star operation
        index = parseOperand(exp, TOKEN_OP_STAR);
        if (index == exp.length() - 1) {
            String leftOperand = exp.substring(0, index);
            Expression leftExpression = createExpression(leftOperand);
            return new ExpressionStar(leftExpression);
        }
        // Parsing symbol trivial expression
        return new ExpressionSymbol(exp);
    }


    /**
     * Creates a new independent copy of the particular expression.
     * @return New independent copy of the original regular expression.
     */
    public abstract Expression duplicateExpression();


    /**
     * Transforms given regular expression into marked one.
     * @param exp Original regular expression to be marked.
     * @return Marked regular expression.
     */
    public static Expression markExpression(Expression exp) {
        // Creating copy of the regular expression
        Expression marked = exp.duplicateExpression();
        // Marking symbols in the expression
        Provider provider = new Provider();
        marked.markSymbols(provider);
        return marked;
    }


    /**
     * Internal class used as a generator of numbers for marking expressions.
     */
    protected static class Provider {
        /**
         * Next value to be assigned.
         */
        int nextValue;
        /**
         * Creates a new provider instance.
         */
        public Provider() {
            resetGenerator();
        }
        /**
         * Assigns a new value.
         * @return Assigned value.
         */
        public int generateValue() {
            int thisValue = this.nextValue;
            this.nextValue++;
            return thisValue;
        }
        /**
         * Resets the generator.
         */
        public final void resetGenerator() {
            this.nextValue = 1;
        }
    }


    // -------------------------------------------------------------------------


    /**
     * Internal function for parsing operand value from the expression.
     * @param exp String form of the expression to be parsed.
     * @param sep String token of the separating operation.
     * @return Parsed operand token.
     */
    protected static int parseOperand(String exp, String sep) {
        int index = 0;
        int brackets = 0;
        String chr = null;
        do {
            chr = exp.substring(index, index + 1);
            if ((brackets == 0) && chr.equals(sep)) {
                break;
            }
            index++;
            if (chr.equals(TOKEN_MT_OPEN)) {
                brackets++;
            } else if (chr.equals(TOKEN_MT_CLOSE)) {
                brackets--;
            }
        } while (index < exp.length());
        return index;
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the type of a given regular subexpression.
     * @return Regular subexpression type.
     */
    public abstract Type getType();



    /**
     * Serializes the inductive structure of the regular expression.
     * @return Serialized regular expression.
     */
    public abstract String serializeExpression();


    /**
     * Returns the set of alphabet symbols occurring in the expression.
     * @return Set of involved symbols.
     */
    public abstract Set<String> getSymbols();


    /**
     * Marks all individual symbols of the alphabet.
     * @param provider References to the generator of number postfixes.
     */
    protected abstract void markSymbols(Provider provider);


    /**
     * Removes the marking postfix of the provided marked symbol.
     * @param markedSymbol Marked symbol to be cleaned.
     * @return Cleaned symbol without appended mark.
     * @see #markExpression(cz.cuni.mff.corrector.expression.Expression)
     */
    public static String trimMarking(String markedSymbol) {
        int position = markedSymbol.indexOf(Expression.TOKEN_MS_MARK, 1);
        return markedSymbol.substring(0, position);
    }


    // -------------------------------------------------------------------------


    /**
     * Cached value of the Empty function.
     */
    protected Boolean allowsEmpty;

    /**
     * The Empty function for Glushkov automaton construction.
     * @return Returns <code>true</code> if expression may be empty.
     */
    public abstract boolean resolveEmpty();

    /**
     * The First function for Glushkov automaton construction.
     * @return Returns the set of symbols that may appear at the beginning.
     */
    public abstract Set<String> resolveFirst();

    /**
     * The Last function for Glushkov automaton construction.
     * @return Returns the set of symbols that may appear at the end.
     */
    public abstract Set<String> resolveLast();

    /**
     * The Follow function for Glushkov automaton construction.
     * @return Returns the mapping of symbols that may be adjacent.
     */
    public abstract Map<String, Set<String>> resolveFollow();


    // -------------------------------------------------------------------------

}
