
package cz.cuni.mff.corrector.grammar;

import cz.cuni.mff.corrector.expression.Expression;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Glushkov automaton representation.
 *
 * @author Martin Svoboda
 */
public class Automaton {

    // -------------------------------------------------------------------------


    /**
     * Name of the initial state of the automaton.
     */
    public static String TOKEN_ST_INITIAL = "#initial";


    // -------------------------------------------------------------------------


    /**
     * Set of automaton states.
     */
    private Set<String> automatonStates;


    /**
     * Set of accepting states.
     */
    private Set<String> acceptingStates;


    /**
     * Encoded transition function of the automaton.
     * Mapping: source state - nonterminal symbol - target state
     */
    private Map<String, Map<String, String>> transitionRules;


    /**
     * Mapping automaton states to corresponding nonterminal symbols.
     */
    private Map<String, String> statesMapping;


    /**
     * Mapping vertices to sets of vertices reachable from the specified one.
     * Note that a state is always reachable from itself.
     */
    private Map<String, Set<String>> reachableStates;


    // -------------------------------------------------------------------------


    /**
     * Creates a new Glushkov automaton for a given marked regular expression.
     * @param markedModel Marked regular expression.
     * @see Expression#markExpression
     */
    public Automaton(Expression markedModel) {

        // Resolving states of the automaton
        Set<String> markedSymbols = markedModel.getSymbols();
        this.automatonStates = new HashSet<String>(markedSymbols);
        this.automatonStates.add(TOKEN_ST_INITIAL);

        // Creating the translation of states
        this.statesMapping = new HashMap<String, String>();
        for (String state : markedSymbols) {
            this.statesMapping.put(state, Expression.trimMarking(state));
        }

        // Resolving set of accepting states
        this.acceptingStates = markedModel.resolveLast();
        if (markedModel.resolveEmpty()) {
            this.acceptingStates.add(TOKEN_ST_INITIAL);
        }

        // Resolving set of transition rules
        this.transitionRules = new HashMap<String, Map<String, String>>();
        Set<String> first = markedModel.resolveFirst();
        Map<String, String> map = new HashMap<String, String>();
        this.transitionRules.put(TOKEN_ST_INITIAL, map);
        for (String targetState : first) {
            map.put(Expression.trimMarking(targetState), targetState);
        }
        Map<String, Set<String>> follow = markedModel.resolveFollow();
        for (String sourceState : follow.keySet()) {
            map = new HashMap<String, String>();
            this.transitionRules.put(sourceState, map);
            for (String targetState : follow.get(sourceState)) {
                map.put(Expression.trimMarking(targetState), targetState);
            }
        }

        // Resolving reachable states mapping
        this.reachableStates = new HashMap<String, Set<String>>();
        computeReachableStates();

    }


    // -------------------------------------------------------------------------


    /**
     * Returns the name of initial state.
     * @return Initial state reference.
     */
    public String getInitialState() {
        return TOKEN_ST_INITIAL;
    }

    /**
     * Returns the reference to the set of automaton states.
     * @return Set of automaton states.
     */
    public Set<String> getAutomatonStates() {
        return this.automatonStates;
    }

    /**
     * Returns the reference to the set of accepting states.
     * @return Set of accepting states.
     */
    public Set<String> getAcceptingStates() {
        return this.acceptingStates;
    }


    /**
     * Returns the set of transitions defined in a given state.
     * @param sourceState Source state of the automaton.
     * @return Map indexed by terminal label to target states.
     */
    public Map<String, String> getTransitionRules(String sourceState) {
        return this.transitionRules.get(sourceState);
    }


    /**
     * Translates the state of the automaton to nonterminal symbol.
     * @param automatonState Specified non-initial automaton state.
     * @return Corresponding nonterminal symbol.
     */
    public String resolveStateTranslation(String automatonState) {
        return this.statesMapping.get(automatonState);
    }


    // -------------------------------------------------------------------------


    /**
     * Computes the internal mapping containing reachable vertices.
     */
    private void computeReachableStates() {

        // Iterating over all automaton states
        for (String base : this.automatonStates) {

            // Initializing working variables
            Set<String> reachable = new HashSet<String>();
            reachable.add(base);
            Set<String> unprocessed = new HashSet<String>(1);
            unprocessed.add(base);

            // Processing all directly reachable states
            while (!unprocessed.isEmpty()) {
                String source = unprocessed.iterator().next();
                Map<String, String> transitions =
                    this.getTransitionRules(source);
                if (transitions != null) {
                    for (String target : transitions.values()) {
                        if (!reachable.contains(target)) {
                            reachable.add(target);
                            unprocessed.add(target);
                        }
                    }
                }
                unprocessed.remove(source);
            }
            
            // Storing set of found reachable states
            this.reachableStates.put(base, reachable);

        }

    }


    /**
     * Returns the set of vertices reachable from the specified one.
     * @param sourceState State to be inspected.
     * @return Set of vertices reachable from the inspected state.
     */
    public Set<String> getReachableStates(String sourceState) {
        return this.reachableStates.get(sourceState);
    }


    // -------------------------------------------------------------------------


    /**
     * Prints the debugging dump of the automaton.
     * @param indentPrefix Indentation string.
     */
    public void printDebugDump(String indentPrefix) {
        
        System.out.print(indentPrefix + "- automaton states:");
        for (String state : this.automatonStates) {
            System.out.print(" " + state);
        }
        System.out.println();

        System.out.print(indentPrefix + "- accepting states:");
        for (String state : this.acceptingStates) {
            System.out.print(" " + state);
        }
        System.out.println();

        System.out.print(indentPrefix + "- states mapping:");
        for (String state : this.statesMapping.keySet()) {
            System.out.print(
                " (s:" + state + "|n:" + this.statesMapping.get(state) + ")"
            );
        }
        System.out.println();

        System.out.print(indentPrefix + "- transition rules:");
        for (String source : this.transitionRules.keySet()) {
            Map<String, String> map = this.transitionRules.get(source);
            for (String nonterminal : map.keySet()) {
                System.out.print(
                    " (s:" + source + "|n:" + nonterminal + "|s:" +
                    map.get(nonterminal) + ")"
                );
            }
        }
        System.out.println();

        System.out.println(indentPrefix + "- reachable states:");
        for (String source : this.reachableStates.keySet()) {
            System.out.print(indentPrefix + "  - " + source + ":");
            for (String target : this.reachableStates.get(source)) {
                System.out.print(" " + target);
            }
            System.out.println();
        }

    }
    

    // -------------------------------------------------------------------------

}
