
package cz.cuni.mff.corrector.grammar;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Representation of a grammar context.
 *
 * @author Martin Svoboda
 */
public class Context {

    // -------------------------------------------------------------------------


    /**
     * Production rule associated with the context.
     */
    private final Rule baseRule;

    /**
     * Set of active production rules.
     */
    private final Set<Rule> activeRules;

    
    /**
     * Container of nonterminal translations based on active rules.
     * Indexing: nonterminal -> list of terminals
     */
    private final Map<String, List<String>> symbolsMap;


    // -------------------------------------------------------------------------


    /**
     * Creates a new instance of a grammar context for a given rule.
     * @param baseRule Base production rule defining the context.
     * @param grammarRef Reference to the grammar object.
     */
    public Context(Rule baseRule, Grammar grammarRef) {

        // Storing internal context fields
        this.baseRule = baseRule;

        // Initializing internal containers
        this.activeRules = new HashSet<Rule>();
        this.symbolsMap = new HashMap<String, List<String>>();

        // Selecting associated grammar rules
        Set<String> allowedNonterminals = baseRule.getAllowedNonterminals();
        for (Rule rule : grammarRef.getProductionRules()) {
            String nonterminalSymbol = rule.getNonterminalSymbol();
            String terminalSymbol = rule.getTerminalSymbol();
            if (allowedNonterminals.contains(nonterminalSymbol)) {
                
                // Storing the active rules reference
                this.activeRules.add(rule);
                
                // Accessing the terminal symbols list
                List<String> terminalsList =
                    this.symbolsMap.get(nonterminalSymbol);
                if (terminalsList == null) {
                    terminalsList = new ArrayList<String>();
                    this.symbolsMap.put(nonterminalSymbol, terminalsList);
                }
                
                // Adding the terminal into translations
                terminalsList.add(terminalSymbol);
                
            }
        }

    }


    // -------------------------------------------------------------------------


    /**
     * Returns the base production rule associated with the context.
     * @return Reference to the associated rule.
     */
    public Rule getBaseRule() {
        return this.baseRule;
    }


    /**
     * Returns the set of symbols with occurrence in the content model.
     * @return Set of nonterminals in a regular expression.
     */
    public Set<String> getAllowedNonterminals() {
        return this.baseRule.getAllowedNonterminals();
    }
    
    
    // -------------------------------------------------------------------------
    

    /**
     * Returns the Glushkov automaton for an associated regular expression.
     * @return Associated Glushkov automaton.
     */
    public Automaton getGlushkovAutomaton() {
        return this.baseRule.getGlushkovAutomaton();
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Translates the given nonterminal symbol into terminal symbols.
     * @param nonterminalSymbol Nonterminal symbol to be translated.
     * @return List of associated terminal symbols.
     */
    public List<String> resolveTerminalSymbols(String nonterminalSymbol) {
        return this.symbolsMap.get(nonterminalSymbol);
    }


    // -------------------------------------------------------------------------


    /**
     * Prints the debugging dump of the grammar context.
     * @param indentPrefix Indentation string.
     */
    public void printDebugDump(String indentPrefix) {

        System.out.println(
            indentPrefix + "- base rule: " +
            this.baseRule.getNonterminalSymbol() + " " +
            this.baseRule.getTerminalSymbol() + " " +
            this.baseRule.getContentModel()
        );

        System.out.print(indentPrefix + "- allowed nonterminals:");
        for (String symbol : this.getAllowedNonterminals()) {
            System.out.print(" " + symbol);
        }
        System.out.println();

        System.out.print(indentPrefix + "- nonterminals mapping:");
        for (String nonterminal : this.symbolsMap.keySet()) {
            for (String terminal : this.symbolsMap.get(nonterminal)) {
                System.out.print(
                    " (n:" + nonterminal + "|t:" + terminal + ")"
                );
            }
        }
        System.out.println();

        System.out.println(indentPrefix + "- glushkov automaton:");
        this.getGlushkovAutomaton().printDebugDump(indentPrefix + "  ");

    }


    // -------------------------------------------------------------------------

}
