
package cz.cuni.mff.corrector.graph;

import cz.cuni.mff.corrector.intent.Intent;
import cz.cuni.mff.corrector.repair.Repair;

/**
 * Edge of the repairing multigraph.
 * 
 * Edge instances can be used in sets as their identification is properly
 * resolved using immutable source vertex, target vertex and assigned
 * correction intent attributes. For the identification of the correction
 * intent itself the local identifier is (must be) used. As a consequence,
 * edges can be automatically identified only locally, i.e. within the given
 * multigraph context.
 *
 * @author Martin Svoboda
 */
public class Edge implements Comparable<Edge> {

    // -------------------------------------------------------------------------


    /**
     * Source vertex of the edge.
     */
    private final Vertex sourceVertex;

    /**
     * Target vertex of the edge.
     */
    private final Vertex targetVertex;


    /**
     * Associated correction intent.
     */
    private final Intent correctionIntent;


    /**
     * Fully evaluated intent repair or <code>null</code>.
     */
    private Repair intentRepair;


    // -------------------------------------------------------------------------


    /**
     * Creates a new instance of the correction edge.
     * The repair structure has to be registered manually later on.
     * Source and target vertices are automatically fetched from the intent.
     * @see #setEdgeRepair
     * @param correctionIntent Associated correction intent.
     */
    public Edge(Intent correctionIntent) {
        this.sourceVertex = correctionIntent.getInitialVertex();
        this.targetVertex = correctionIntent.getEndingVertex();
        this.correctionIntent = correctionIntent;
        this.intentRepair = null;
    }
    
    
    /**
     * Creates a new instance of the correction edge.
     * @param sourceVertex Source vertex of the edge.
     * @param targetVertex Target vertex of the edge.
     * @param correctionIntent Associated correction intent.
     * @deprecated Only for testing purposes only!
     */
    protected Edge(
        Vertex sourceVertex,
        Vertex targetVertex,
        Intent correctionIntent
    ) {
        this.sourceVertex = sourceVertex;
        this.targetVertex = targetVertex;
        this.correctionIntent = correctionIntent;
        this.intentRepair = null;
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the source vertex reference.
     * @return Source vertex of the edge.
     */
    public Vertex getSourceVertex() {
        return this.sourceVertex;
    }

    /**
     * Returns the stratum number of the source vertex reference.
     * @return Source vertex stratum number.
     */
    public int getSourceStratum() {
        return this.sourceVertex.getStratumNumber();
    }

    /**
     * Returns the automaton state of the source vertex reference.
     * @return Source vertex automaton state.
     */
    public String getSourceState() {
        return this.sourceVertex.getAutomatonState();
    }


    /**
     * Returns the target vertex reference.
     * @return Target vertex of the edge.
     */
    public Vertex getTargetVertex() {
        return this.targetVertex;
    }

    /**
     * Returns the stratum number of the target vertex reference.
     * @return Target vertex stratum number.
     */
    public int getTargetStratum() {
        return this.targetVertex.getStratumNumber();
    }

    /**
     * Returns the automaton state of the target vertex reference.
     * @return Target vertex automaton state.
     */
    public String getTargetState() {
        return this.targetVertex.getAutomatonState();
    }
    
    
    // -------------------------------------------------------------------------


    /**
     * Returns the associated correction intent.
     * @return Correction intent object reference.
     */
    public Intent getCorrectionIntent() {
        return this.correctionIntent;
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the optional reference to the nested intent repair.
     * @return Associated intent repair structure or <code>null</code>.
     */
    public Repair getIntentRepair() {
        return this.intentRepair;
    }
    
    
    /**
     * Returns the cost assigned to the edge and nested intent repair.
     * This function can only be invoked if the repair was already registered.
     * @return Fully or partially evaluated cost of the entire nested repair.
     */
    public int getEdgeCost() {
        return this.intentRepair.getTotalCost();
    }
    
    
    /**
     * Registers the provided nested intent repair structure.
     * @param intentRepair Fully or partially evaluated intent repair.
     * @see Repair#isCompletelyEvaluated
     */
    public void setEdgeRepair(Repair intentRepair) {
        this.intentRepair = intentRepair;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the edge description string.
     * @return Human readable edge serialization.
     */
    public String getSerializedString() {
        return "[" + this.sourceVertex.getSerializedString() + "-" +
            this.targetVertex.getSerializedString() + "]";
    }


    // -------------------------------------------------------------------------


    @Override
    public boolean equals(Object obj) {
        if (obj instanceof Edge) {
            Edge e = (Edge)obj;
            return (
                (this.sourceVertex.equals(e.sourceVertex)) &&
                (this.targetVertex.equals(e.targetVertex)) &&
                (this.correctionIntent == e.correctionIntent)
            );
        } else {
            return false;
        }
    }
    

    @Override
    public int hashCode() {
        int hash = 5;
        hash = 97 * hash + this.sourceVertex.hashCode();
        hash = 97 * hash + this.targetVertex.hashCode();
        hash = 97 * hash + this.correctionIntent.hashCode();
        return hash;
    }
    
    
    public int compareTo(Edge e) {
        // Comparing source vertices
        int cSource = this.sourceVertex.compareTo(e.sourceVertex);
        if (cSource != 0) {
            return cSource;
        }
        // Comparing target vertices
        int cTarget = this.targetVertex.compareTo(e.targetVertex);
        if (cTarget != 0) {
            return cTarget;
        }
        // Comparing intent signatures
        return this.correctionIntent.getGlobalSignature().compareTo(
            e.correctionIntent.getGlobalSignature()
        );
    }


    // -------------------------------------------------------------------------
    
}
