
package cz.cuni.mff.corrector.intent;

import cz.cuni.mff.corrector.edit.Edit;
import cz.cuni.mff.corrector.grammar.Context;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.grammar.Rule;
import cz.cuni.mff.corrector.graph.Vertex;
import cz.cuni.mff.corrector.instruction.InstructionAddLeaf;
import static cz.cuni.mff.corrector.intent.Intent.SEPARATOR;
import cz.cuni.mff.corrector.repair.Mend;
import cz.cuni.mff.corrector.repair.NodeRepairInsertSubtree;
import cz.cuni.mff.corrector.tree.Node;
import java.util.*;

/**
 * Implementation of the insert correction intent.
 *
 * @author Martin Svoboda
 */
public class IntentInsert extends Intent {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * List of all intent types allowed in further nesting.
     */
    private static final Set<Type> ALLOWED_INTENTS = EnumSet.of(Type.INSERT);
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new instance of the insert correction intent.
     * @param initialVertex Pair of the initial stratum and state.
     * @param endingVertex Pair of the ending stratum and state.
     * @param grammarContext Nested grammar context reference.
     * @param contextChain Optional chain of grammar contexts.
     * @param nodeLabel New label (terminal) of the inserted node.
     */
    public IntentInsert(
        Vertex initialVertex,
        Vertex endingVertex,
        Context grammarContext,
        List<Context> contextChain,
        String nodeLabel
    ) {
        
        // Constructing core components
        super(
            null,
            new NodeRepairInsertSubtree(new InstructionAddLeaf(nodeLabel)),
            initialVertex,
            endingVertex,
            new ArrayList<Node>(0),
            grammarContext,
            contextChain
        );
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    @Override
    public Type getIntentType() {
        return Type.INSERT;
    }
    
    @Override
    public Set<Type> getAllowedIntents() {
        return ALLOWED_INTENTS;
    }
    
    @Override
    protected String computeIntentSignature() {
        return Type.INSERT.fullCode + computeChainSignature();
    }
    
    @Override
    protected String computeIdentifierPostfix() {
        Rule baseRule = this.getGrammarContext().getBaseRule();
        return Type.INSERT.fullCode +
            SEPARATOR + baseRule.getNonterminalSymbol() +
            SEPARATOR + baseRule.getTerminalSymbol();
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates the set of all permitted nested insert intents.
     * @param outputSet Output container for generated intents.
     * @param parentIntent Parental correction intent reference.
     * @param initialStratum Number of the initial stratum.
     * @param initialState Name of the initial state.
     * @param treeGrammar Reference to the tree grammar.
     * @param transitions Set of transitions from the initial state.
     */
    public static void generateNestedIntents(
        List<Intent> outputSet,
        Intent parentIntent,
        int initialStratum,
        String initialState,
        Grammar treeGrammar,
        Map<String, String> transitions
    ) {
        
        // Resolving necessary prerequisites
        boolean isNestingAllowed =
            (parentIntent.getAllowedIntents().contains(Intent.Type.INSERT));
        boolean hasTransitionsDefined =
            ((transitions != null) && (!transitions.isEmpty()));
        
        // Further nesting is not possible
        if (!isNestingAllowed || !hasTransitionsDefined) {
            return;
        }
        
        // Inspecting chain of insert intents
        List<Context> superiorChain;
        if (parentIntent.getIntentType() == Type.INSERT) {
            superiorChain = parentIntent.getContextChain();
        } else {
            superiorChain = new ArrayList<Context>();
        }
        
        // Creating nested insertion intents
        for (String nonterminalSymbol : transitions.keySet()) {
            List<String> suitableTerminals = parentIntent.getGrammarContext().
                resolveTerminalSymbols(nonterminalSymbol);
            for (String terminalSymbol : suitableTerminals) {
                
                // Resolving a grammar context
                Context nestedContext = treeGrammar.getGeneralContext(
                    nonterminalSymbol, terminalSymbol
                );
                
                // Preventing repeated context in a chain
                if (superiorChain.contains(nestedContext)) {
                    continue;
                }
                
                // Appending new contexts chain
                List<Context> nestedChain =
                    new ArrayList<Context>(superiorChain);
                nestedChain.add(nestedContext);
                
                // Resolving an ending vertex
                int endingStratum = initialStratum;
                String endingState = transitions.get(nonterminalSymbol);
                
                // Creating a correction intent
                IntentInsert newIntent = new IntentInsert(
                    new Vertex(initialStratum, initialState),
                    new Vertex(endingStratum, endingState),
                    nestedContext,
                    nestedChain,
                    terminalSymbol
                );
                
                // Adding the intent into the output set
                outputSet.add(newIntent);
                
            }
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates all mends for the particular intent repair of insert type.
     * @param nodeEdit Sequence of edit operations repairing the node.
     * @param sequenceMends Set of all mends of the sequence repair.
     */
    public static void generateIntentFix(
        List<Edit> nodeEdit,
        List<Mend> sequenceMends
    ) {
        
        // Processing individual sequence mends
        for (Mend seqMend : sequenceMends) {
            
            // Adjusting edit operations from the sequence mend
            seqMend.applyRootPrepending(0);
            // Placing clones at the beginning of the original list
            Edit.cloneIdenticalSequenceBefore(
                nodeEdit, seqMend.getEditSequence()
            );
            
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
}
