
package cz.cuni.mff.corrector.intent;

import cz.cuni.mff.corrector.grammar.Context;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.grammar.Rule;
import cz.cuni.mff.corrector.graph.Vertex;
import static cz.cuni.mff.corrector.intent.Intent.SEPARATOR;
import cz.cuni.mff.corrector.repair.Mend;
import cz.cuni.mff.corrector.repair.NodeRepairRepairSubtree;
import cz.cuni.mff.corrector.tree.Node;
import java.util.*;

/**
 * Implementation of the repair correction intent.
 *
 * @author Martin Svoboda
 */
public class IntentRepair extends Intent {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * List of all intent types allowed in further nesting.
     */
    private static final Set<Type> ALLOWED_INTENTS =
        EnumSet.of(Type.INSERT, Type.DELETE, Type.REPAIR, Type.RENAME);
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new instance of the repair correction intent.
     * @param baseNode Optional associated base node reference.
     * @param initialVertex Pair of the initial stratum and state.
     * @param endingVertex Pair of the ending stratum and state.
     * @param nodeSequence Sequence of nested nodes to be processed.
     * @param grammarContext Nested grammar context reference.
     */
    public IntentRepair(
        Node baseNode,
        Vertex initialVertex,
        Vertex endingVertex,
        List<Node> nodeSequence,
        Context grammarContext
    ) {
        
        // Constructing core components
        super(
            baseNode,
            new NodeRepairRepairSubtree(),
            initialVertex,
            endingVertex,
            nodeSequence,
            grammarContext,
            null
        );
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    @Override
    public Type getIntentType() {
        return Type.REPAIR;
    }
    
    @Override
    public Set<Type> getAllowedIntents() {
        return ALLOWED_INTENTS;
    }
    
    @Override
    protected String computeIntentSignature() {
        Rule rule = this.getGrammarContext().getBaseRule();
        return Type.REPAIR.fullCode +
            SEPARATOR + getBaseNode() +
            SEPARATOR + rule.getNonterminalSymbol() +
            SEPARATOR + rule.getTerminalSymbol();
    }
    
    @Override
    protected String computeIdentifierPostfix() {
        Rule baseRule = this.getGrammarContext().getBaseRule();
        return Type.REPAIR.fullCode +
            SEPARATOR + baseRule.getNonterminalSymbol() +
            SEPARATOR + baseRule.getTerminalSymbol();
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates the set of all permitted nested repair intents.
     * @param outputSet Output container for generated intents.
     * @param parentIntent Parental correction intent reference.
     * @param initialStratum Number of the initial stratum.
     * @param initialState Name of the initial state.
     * @param treeGrammar Reference to the tree grammar.
     * @param transitions Set of transitions from the initial state.
     */
    public static void generateNestedIntents(
        List<Intent> outputSet,
        Intent parentIntent,
        int initialStratum,
        String initialState,
        Grammar treeGrammar,
        Map<String, String> transitions
    ) {
        
        // Resolving necessary prerequisites
        boolean isNestingAllowed =
            (parentIntent.getAllowedIntents().contains(Intent.Type.REPAIR));
        boolean isLastStratum =
            (initialStratum == parentIntent.getNodesSequence().size());
        boolean hasTransitionsDefined =
            ((transitions != null) && (!transitions.isEmpty()));
        
        // Further nesting is not possible
        if (!isNestingAllowed || isLastStratum || !hasTransitionsDefined) {
            return;
        }
        
        // Referencing the following node
        Node inspectedNode = parentIntent.nodeSequence.get(initialStratum);
        String nodeLabel = inspectedNode.getNodeLabel();
        
        // Creating nested repair intents
        for (String nonterminalSymbol : transitions.keySet()) {
            List<String> suitableTerminals = parentIntent.getGrammarContext().
                resolveTerminalSymbols(nonterminalSymbol);
            for (String terminalSymbol : suitableTerminals) {
                
                // Processing only when relevant
                if (!nodeLabel.equals(terminalSymbol)) {
                    continue;
                }
                
                // Resolving a grammar context
                Context nestedContext = treeGrammar.getGeneralContext(
                    nonterminalSymbol, terminalSymbol
                );
                
                // Resolving an ending vertex
                int endingStratum = initialStratum + 1;
                String endingState = transitions.get(nonterminalSymbol);
                
                // Creating a nested correction intent
                IntentRepair newIntent = new IntentRepair(
                    inspectedNode,
                    new Vertex(initialStratum, initialState),
                    new Vertex(endingStratum, endingState),
                    inspectedNode.getChildNodes(),
                    nestedContext
                );

                // Adding the intent into the output set
                outputSet.add(newIntent);
                
            }
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates all mends for the particular intent repair of repair type.
     * @param sequenceMends Set of all mends of the sequence repair.
     */
    public static void generateIntentFix(
        List<Mend> sequenceMends
    ) {
        
        // Processing individual sequence mends
        for (Mend seqMend : sequenceMends) {
            
            // Adjusting edit operations from the sequence mend
            seqMend.applyRootPrepending(0);
            
        }
        
    }
    
    
    // -------------------------------------------------------------------------

}
