
package cz.cuni.mff.corrector.repair;

import cz.cuni.mff.corrector.edit.Edit;
import cz.cuni.mff.corrector.graph.Edge;
import cz.cuni.mff.corrector.graph.Graph;
import cz.cuni.mff.corrector.graph.Vertex;
import cz.cuni.mff.corrector.intent.*;
import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * Representation of the correction intent repair.
 * 
 * @author Martin Svoboda
 */
public class Repair {

    // -------------------------------------------------------------------------

    
    /**
     * Reference to the associated correction intent.
     */
    private final Intent correctionIntent;
    

    /**
     * Reference to the node repair.
     */
    protected final NodeRepair nodeRepair;
    
    /**
     * Sequence repair reference.
     */
    private SequenceRepair sequenceRepair;
    
    
    /**
     * State of the fully intent repair evaluation.
     * This value is set manually and only the user is responsible for its semantic.
     * However, if the graph is provided during the construction, it is set to <code>true</code>.
     * @see #setCompletenessFlag
     */
    protected boolean isEvaluated;


    // -------------------------------------------------------------------------


    /**
     * Creates a new instance of the intent repair structure.
     * The evaluation completeness flag is always set to <code>false</code>.
     * @param associatedIntent Reference to the associated intent.
     * @param correctionGraph Reference to the prepared multigraph.
     */
    public Repair(
        Intent associatedIntent,
        Graph correctionGraph
    ) {
        this.correctionIntent = associatedIntent;
        this.nodeRepair = associatedIntent.getNodeRepair();
        this.sequenceRepair =
            new SequenceRepair(
                correctionGraph, associatedIntent.getIntentType()
            );
        this.isEvaluated = false;
    }
    
    
    /**
     * Creates a new instance of the intent repair structure.
     * The evaluation completeness flag is always set to <code>false</code>.
     * The omitted sequence repair (multigraph) can be registered later on.
     * @param associatedIntent Reference to the associated intent.
     */
    public Repair(
        Intent associatedIntent
    ) {
        this.correctionIntent = associatedIntent;
        this.nodeRepair = associatedIntent.getNodeRepair();
        this.sequenceRepair = null;
        this.isEvaluated = false;
    }
    
    
    /**
     * Creates a new instance of the intent repair structure.
     * @param associatedIntent Reference to the associated intent.
     * @param nodeRepair Reference to the prepared node repair.
     * @param sequenceRepair Reference to the prepared sequence repair.
     * @deprecated Only for testing purposes only!
     */
    protected Repair(
        Intent associatedIntent,
        NodeRepair nodeRepair,
        SequenceRepair sequenceRepair
    ) {
        this.correctionIntent = associatedIntent;
        this.nodeRepair = nodeRepair;
        this.sequenceRepair = sequenceRepair;
        this.isEvaluated = false;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Registers the provided multigraph with this repair.
     * The sequence repair encapsulating the graph is created automatically.
     * The evaluation completeness flag is kept untouched.
     * @param correctionGraph Reference to the prepared multigraph.
     */
    public void registerSequenceRepair(
        Graph correctionGraph
    ) {
        this.sequenceRepair =
            new SequenceRepair(
                correctionGraph, this.correctionIntent.getIntentType()
            );
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the reference to the associated correction intent.
     * IMPORTANT NOTICE: Note that when signature handling is enabled,
     * this correction intent reference is fully valid and makes sense
     * only within the task that actually generated this repair structure
     * and especially not from other tasks that would only like to wait
     * until this repair is also available for them!
     * @return Associated correction intent structure reference.
     */
    public Intent getCorrectionIntent() {
        return this.correctionIntent;
    }
    
    
    /**
     * Returns the associated node repair.
     * @return Reference to the node repair.
     */
    public NodeRepair getNodeRepair() {
        return this.nodeRepair;
    }
    
    
    /**
     * Returns the reference to the sequence repair.
     * Note that this reference is invalid if not yet registered.
     * @return Associated sequence repair structure.
     */
    public SequenceRepair getSequenceRepair() {
        return this.sequenceRepair;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the value of the flag of fully evaluated sequence repair.
     * Note that the repair is considered as complete if its corresponding
     * task declared this fact. It means that no online detection is
     * undertaken at this moment, only the current value of stored flag
     * is accessed and returned.
     * @return The current value of the sequence repair evaluation flag.
     */
    public boolean isCompletelyEvaluated() {
        return this.isEvaluated;
    }
    
    /**
     * Sets the repair evaluation flag as activated (to <code>true</code>).
     */
    public void setCompletenessFlag() {
        this.isEvaluated = true;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the overall intent repair cost.
     * This method can be invoked even if the internal graph is not yet
     * defined (assigned to the internal sequence repair structure). In
     * such case only the node repair cost is considered, or even if this
     * node repair is not defined, <code>0</code> is transparently returned.
     * @return Overall correction intent repairing cost.
     */
    public int getTotalCost() {
        int overallCost = 0;
        if (this.nodeRepair != null) {
            overallCost += this.nodeRepair.getRepairCost();
        }
        if (this.sequenceRepair != null) {
            overallCost += this.sequenceRepair.getRepairCost();
        }
        return overallCost;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generated fix for the entire intent repair.
     * This reference is equal to null if the fix was not yet generated.
     * @see #getIntentFix
     * @see #generateIntentFix
     */
    private List<List<Edit>> intentFix = null;
    
    
    /**
     * Generated set of mends for the entire intent repair.
     * This reference is equal to null if mends were not yet generated.
     * @see #getIntentMends
     * @see #generateIntentMends
     */
    private List<Mend> intentMends = null;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the set of all found repairing edit sequences.
     * Note that internal structures are provided outside!
     * Individual edit operations and even their lists must not be modified!
     * This function may only be called on aggregated repairs.
     * @return List of found repairing edit sequences.
     */
    public List<List<Edit>> getIntentFix() {
        // Invoking the fix generation
        if (this.intentFix == null) {
            this.intentFix = generateIntentFix();
        }
        // Returning the fix reference
        return this.intentFix;
    }
    
    
    /**
     * Generates the fix structure for the entire intent repair.
     * @return Newly generated intent repair fix.
     */
    private List<List<Edit>> generateIntentFix() {
        
        // Assuring the generation of mends
        getIntentMends();
        
        // Generating the list of edit sequences
        List<List<Edit>> outputList = new ArrayList<List<Edit>>();
        for (Mend mend : this.intentMends) {
            outputList.add(mend.getEditSequence());
        }
        
        // Returning the prepared fix
        return outputList;
        
    }


    /**
     * Generates the set of all mends for the intent repair.
     * This procedure directly modifies the provided container.
     * @param sequenceMends Particular working set of sequence mends.
     */
    private void generateIntentFix(List<Mend> sequenceMends) {

        // Switching particular intent types
        switch (this.getSequenceRepair().getIntentType()) {
            case CORRECT:
                IntentCorrect.generateIntentFix(
                    sequenceMends
                );
                break;
            case INSERT:
                IntentInsert.generateIntentFix(
                    this.nodeRepair.generateNodeFix(),
                    sequenceMends
                );
                break;
            case DELETE:
                IntentDelete.generateIntentFix(
                    this.nodeRepair.generateNodeFix(),
                    sequenceMends
                );
                break;
            case REPAIR:
                IntentRepair.generateIntentFix(
                    sequenceMends
                );
                break;
            case RENAME:
                IntentRename.generateIntentFix(
                    this.nodeRepair.generateNodeFix(),
                    sequenceMends
                );
                break;
            default:
                throw new UnsupportedOperationException();
        }

    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the set of all mends for the intent repair.
     * If this set is not already created, it will be generated automatically.
     * The caller must not manipulate with the provided list or its items!
     * @return Fetched or newly generated set of mend structures.
     */
    public List<Mend> getIntentMends() {
        // Invoking the mends generation
        if (this.intentMends == null) {
            this.intentMends = generateIntentMends();
        }
        // Returning the mends reference
        return this.intentMends;
    }
    
    
    /**
     * Transforms the intent repair into all possible edit sequences.
     * @return Newly generated set of mend structures.
     */
    private List<Mend> generateIntentMends() {

        // Fetching the set of mends for the sequence
        List<Mend> sequenceMends = this.sequenceRepair.generateSequenceFix();

        // Encapsulating sequence mends to intent mends
        this.generateIntentFix(sequenceMends);
        
        // Returning the adjusted intent mends
        return sequenceMends;

    }
    
    
    /**
     * Exports generated intent fix into an external file.
     * @param fileName Name of the output file to be generated.
     * @return Returns <code>true</code> on success.
     */
    public boolean exportIntentFix(String fileName) {
        
        try {
            
            // Accessing the intent fix
            List<List<Edit>> intentFix = getIntentFix();
            
            // Preparing output file
            DataOutputStream outputStream =
                new DataOutputStream(new FileOutputStream(fileName));
            
            // Processing generated mends
            int mendNumber = 0;
            for (List<Edit> mend : intentFix) {
                mendNumber++;
                outputStream.writeBytes("Repair " + mendNumber + ":" + EOL);
                for (Edit edit : mend) {
                    outputStream.writeBytes("  " + edit.getOperationTitle() + EOL);
                }
            }
            
            // Closing the output stream
            outputStream.close();
            
            // Returning successful flag
            return true;
            
        } catch (IOException ex) {
            return false;
        }
            
    }
    
    /**
     * Default line separator.
     */
    private static final String EOL = System.getProperty("line.separator");

    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prints the debugging dump of the repair intent structure.
     * @see #printDebugDump(java.lang.String)
     */
    public void printDebugDump() {
        printDebugDump("");
    }
    
    /**
     * Prints the debugging dump of the repair intent structure.
     * Output pattern:
     *   Correction intent signature 1
     *     Vertex - Nested Intent Signature - Vertex - ... - Vertex
     *     Shortest path 2
     *     ...
     *   Correction intent signature 2
     *   Correction intent signature 3
     *   ...
     * Prints basic information about this and all other (even recursively)
     * nested correction intents. Using the signature distinction, each
     * signature distinct intent is printed right once.
     * Processed are all those correction intents that are nested in this
     * one and that are incident with the shortest paths.
     * @param indentPrefix Optional initial output indentation string.
     */
    public void printDebugDump(String indentPrefix) {
        
        // Initialising the working container
        Map<String, Repair> collectedRepairs = new TreeMap<String, Repair>();
        
        // Adding this repair into the requests
        collectedRepairs.put(this.correctionIntent.getGlobalSignature(), this);
        
        // Recursively gathering nested repairs
        gatherInvolvedRepairs(collectedRepairs);
        
        // Printing the debugging dump
        for (
            Map.Entry<String, Repair> nestedEntry :
            collectedRepairs.entrySet()
        ) {
            String nestedSignature = nestedEntry.getKey();
            Repair nestedRepair = nestedEntry.getValue();
            System.out.println(nestedSignature);
            nestedRepair.printPathsDump(indentPrefix + "  ");
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Collects all nested intent repairs that are involved in shortest paths.
     * @param nestedRepairs Output container for gathered nested repairs.
     */
    protected void gatherInvolvedRepairs(
        Map<String, Repair> nestedRepairs
    ) {
        
        // Fetching the shortest paths
        List<List<Edge>> shortestPaths = this.sequenceRepair.
            getCorrectionGraph().fetchShortestPaths();
        
        // Processing fetched paths
        gatherInvolvedRepairs(nestedRepairs, shortestPaths);
        
    }
    
    /**
     * Collects all nested intent repairs that are involved in shortest paths.
     * This method is internal and serves only for derived particular repairs!
     * @param nestedRepairs Output container for active nested repairs.
     * @param shortestPaths Shortest paths to be printed and processed.
     */
    protected void gatherInvolvedRepairs(
        Map<String, Repair> nestedRepairs,
        List<List<Edge>> shortestPaths
    ) {
        
        // Processing fetched paths
        for (List<Edge> selectedPath : shortestPaths) {
            
            // Processing all edges of a path
            for (Edge currentEdge : selectedPath) {
                
                Repair nestedRepair = currentEdge.getIntentRepair();
                Intent nestedIntent = currentEdge.getCorrectionIntent();
                String nestedSignature = nestedIntent.getGlobalSignature();
                
                // Adding the found nested repair
                if (!nestedRepairs.containsKey(nestedSignature)) {
                    nestedRepairs.put(nestedSignature, nestedRepair);
                    nestedRepair.gatherInvolvedRepairs(nestedRepairs);
                }
                
            }
            
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prints the list of all shortest paths of the given repair.
     * @param indentPrefix Output indentation string.
     */
    protected void printPathsDump(String indentPrefix) {
        
        // Fetching the shortest paths
        List<List<Edge>> shortestPaths = this.sequenceRepair.
            getCorrectionGraph().fetchShortestPaths();
        
        // Processing fetched paths
        printPathsDump(indentPrefix, shortestPaths);
        
    }
    
    /**
     * Prints the list of all provided paths.
     * This method is internal and serves only for derived particular repairs!
     * @param indentPrefix Output indentation string.
     * @param shortestPaths Shortest paths to be printed.
     */
    protected void printPathsDump(
        String indentPrefix,
        List<List<Edge>> shortestPaths
    ) {
        
        // Sorting the shortest paths
        Collections.sort(
            shortestPaths,
            this.sequenceRepair.getCorrectionGraph().pathsComparator
        );
        
        // Processing fetched paths
        for (List<Edge> selectedPath : shortestPaths) {
            
            // Printing the source vertex
            if (selectedPath.size() > 0) {
                Edge firstEdge = selectedPath.iterator().next();
                Vertex sourceVertex = firstEdge.getSourceVertex();
                System.out.print(
                    indentPrefix +
                    sourceVertex.getSerializedString()
                );
            }
            
            // Printing edge by edge details
            for (Edge currentEdge : selectedPath) {
                Vertex targetVertex = currentEdge.getTargetVertex();
                Intent nestedIntent = currentEdge.getCorrectionIntent();
                System.out.print(
                    " " + nestedIntent.getGlobalSignature() + " " +
                    targetVertex.getSerializedString()
                );
            }
            
            // Finalising the path output
            if (selectedPath.size() > 0) {
                System.out.println();
            }
            
        }
        
    }
    
    
    /**
     * Prints the dump of computed fixes, i.e. all paths and their operations.
     */
    public void printFixesDump() {
        
        // Assuring the generation of mends
        getIntentMends();
        
        // Printing edit operations on paths
        for (Mend mend : this.intentMends) {
            for (Edit edit : mend.getEditSequence()) {
                System.out.print(edit.getOperationTitle() + " ");
            }
            System.out.println();
        }
        
    }
    

    // -------------------------------------------------------------------------

}
