
package cz.cuni.mff.corrector.repair;

import cz.cuni.mff.corrector.edit.Edit;
import cz.cuni.mff.corrector.graph.Edge;
import cz.cuni.mff.corrector.graph.Graph;
import cz.cuni.mff.corrector.intent.Intent;
import java.util.ArrayList;
import java.util.List;

/**
 * Abstract sequence repair encapsulation for correction multigraph.
 * 
 * Note that a sequence repair structure always contains a valid reference
 * to an associated correction multigraph, however, a given intent repair
 * structure may not yet have defined reference to any sequence repair
 * (when the given correction multigraph not yet exists).
 *
 * @author Martin Svoboda
 */
public class SequenceRepair {

    // -------------------------------------------------------------------------
    
    
    /**
     * Associated correction multigraph reference.
     */
    private final Graph correctionGraph;
    
    
    /**
     * Particular type of the associated correction intent.
     */
    private final Intent.Type intentType;


    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new instance of the sequence repair.
     * @param correctionGraph Reference to the correction multigraph.
     * @param intentType Particular type of associated intent.
     */
    public SequenceRepair(
        Graph correctionGraph,
        Intent.Type intentType
    ) {
        this.correctionGraph = correctionGraph;
        this.intentType = intentType;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the associated correction multigraph.
     * @return Reference to the associated multigraph.
     */
    public Graph getCorrectionGraph() {
        return this.correctionGraph;
    }
    
    
    /**
     * Returns the overall sequence repair cost from the multigraph.
     * @return Sequence repair terminal cost.
     */
    public int getRepairCost() {
        return this.correctionGraph.getTermCost();
    }
    
    
    /**
     * Returns the type of the associated correction intent.
     * @return Associated intent type enumeration constant.
     */
    public Intent.Type getIntentType() {
        return this.intentType;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Transforms the sequence repair into all possible edit sequences.
     * @return Generated set of mend structures.
     */
    public List<Mend> generateSequenceFix() {

        // Fetching all repairing shortest paths
        List<List<Edge>> pathsSet = this.correctionGraph.fetchShortestPaths();

        // Preparing output container for composed mends
        List<Mend> outputContainer = new ArrayList<Mend>();

        // Iterating over all find repairing paths
        for (List<Edge> path : pathsSet) {

            // Preparing working sequence of nested mends
            List<List<Mend>> nestedMends = new ArrayList<List<Mend>>();

            // Processing eges on the particular path
            for (Edge edge : path) {
                // Fetching mends from the nested intent
                Repair edgeRepair = edge.getIntentRepair();
                List<Mend> edgeMends = edgeRepair.getIntentMends();
                // Adding the fetched set into the working list
                nestedMends.add(edgeMends);
            }
            
            /*
             * Note that the list of nested mends may contain duplicated
             * references to the same mend structures! This is because
             * the path may contain edges with the same intent signatures
             * and, thus, generated mend structures.
             */
            
            // Producing the Cartesian product of mends
            this.combineNestedMends(
                outputContainer,
                nestedMends
            );
            
        }
        
        // Returning the set of all found sequence mends
        return outputContainer;
        
    }
    
    
    /**
     * Finds all mend combinations for a given sequence.
     * @param outputContainer Output container for generated mends.
     * @param pathMends List of original mend sets to be combined.
     * Cloned edit operations are used.
     */
    protected void combineNestedMends(
        List<Mend> outputContainer,
        List<List<Mend>> pathMends
    ) {
        this.combineNestedMends(
            outputContainer,
            pathMends,
            0,
            new ArrayList<Edit>(),
            0
        );
    }
    
    
    /**
     * Finds all mend combinations for a given sequence.
     * @param outputContainer Output container for generated mends.
     * @param pathMends List of mend sets to be combined.
     * @param recursionIndex Control variable for recursive processing.
     * @param workingArray Working sequence for mend combinations.
     * @param workingShift Running value of shifting size.
     */
    private void combineNestedMends(
        List<Mend> outputContainer,
        List<List<Mend>> pathMends,
        int recursionIndex,
        List<Edit> workingSequence,
        int workingShift
    ) {
        
        // Terminating the recursive processing
        if (recursionIndex == pathMends.size()) {
            
            // Encapsulating a newly constructed sequence mend
            Mend newMend = new Mend(
                new ArrayList<Edit>(
                    Edit.cloneIdenticalSequence(workingSequence)
                ),
                resolveShiftingSize(workingShift)
            );
            // Adding this new mend into the output container
            outputContainer.add(newMend);
            
        // Recursively exploring the next edge
        } else {
            
            // Referencing mends of the given edge
            List<Mend> mendsSet = pathMends.get(recursionIndex);
            
            // Storing the initial pointer to the working list
            int formerSize = workingSequence.size();
            
            // Processing all particular mends
            for (Mend mend : mendsSet) {
                // Appending fetched, cloned and shifted edit operations
                Edit.cloneShiftedSequence(
                    mend.getEditSequence(), workingSequence, workingShift
                );
                // Recursively continuing generation process
                combineNestedMends(
                    outputContainer,
                    pathMends,
                    recursionIndex + 1,
                    workingSequence,
                    workingShift + mend.getShiftingSize()
                );
                // Cleaning the working sequence to the former state
                int currentSize = workingSequence.size();
                for (int i = currentSize - 1; i >= formerSize ; i--) {
                    workingSequence.remove(i);
                }
                
            }
            
        }
        
    }
    
    
    /**
     * Resolves the shifting size depending on the parent intent type.
     * @param finalShifting Reached local shifting of the sequence.
     * @return Final shifting with respect to the parent intent behavior.
     */
    private int resolveShiftingSize(int finalShifting) {
        switch (this.intentType) {
            case INSERT:
            case REPAIR:
            case RENAME:
                return 1;
            case DELETE:
                return 0;
            case CORRECT:
                return 0;
            default:
                throw new UnsupportedOperationException();
        }
    }
    
    
    // -------------------------------------------------------------------------

}
