
package cz.cuni.mff.corrector.testing;

import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.tree.Tree;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * Ancestor for particular experimental evaluation analyses.
 * 
 * @author Martin Svoboda
 */
public class Analysis {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Relative path from the home folder to testing documents.
     */
    public static final String PATH = "profiling/";
    
    
    /**
     * End of line characters used in outputs of analyses.
     */
    public static final String EOL = System.getProperty("line.separator");
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Provides the instance of the required evaluation data tree instance.
     * @param analysisName Analysis name to determine directory.
     * @param grammarId Identifier of the associated grammar.
     * @param nodesCount Number of nodes to reference the required document.
     * @param serialNumber Serial number to reference the required document.
     * @return Parsed and created data tree instance.
     */
    protected Tree acquireDataTree(
        String analysisName,
        int grammarId,
        int nodesCount,
        int serialNumber
    ) {
        
        // Accessing the input XML document
        String fileName = PATH + "trees-" + analysisName + "/" +
            "Tree-" + grammarId + "-" +
            nodesCount + "-" + serialNumber + ".xml";
        
        // Printing progress notification
        System.out.println("Loading data tree [" + fileName + "]");
        
        // Parsing the input XML document
        Tree treeObject = Tree.createFromXmlFile(fileName);
        
        // Returning the data tree object
        return treeObject;
        
    }
    
    
    /**
     * Provides the instance of the required evaluation grammar instance.
     * @param grammarId Identifier of the requested grammar.
     * @return Parsed and ready to use grammar instance.
     */
    protected Grammar acquireGrammarInstance(int grammarId) {
        
        // Accessing the input grammar document
        String fileName = PATH + "grammars/" + "Grammar-" + grammarId + ".txt";
        
        // Printing progress notification
        System.out.println("Loading grammar [" + fileName + "]");
        
        // Parsing the grammar document
        Grammar grammarObject = Grammar.createFromDataFile(fileName);
        
        // Returning the prepared grammar object
        return grammarObject;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Exports the provided analysis output into the required text file.
     * @param outputName Core part of the analysis output name.
     * @param stringBuilder Textual output of the performed analysis. 
     */
    protected void exportComputedAnalysis(
        String outputName,
        StringBuilder stringBuilder
    ) {
        
        // Preparing the output file name
        String fileName = PATH + "analyses/" + outputName + ".txt";
        
        // Exporting the output content
        try {
            FileOutputStream fos = new FileOutputStream(fileName);
            DataOutputStream dos = new DataOutputStream(fos);
            dos.writeBytes(stringBuilder.toString());
            dos.close();
        } catch (FileNotFoundException e) {
            System.out.println("Unable to open output file");
            throw new Error();
        } catch (IOException e) {
            System.out.println("Unable to write analysis output");
            throw new Error();
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
}
