
package cz.cuni.mff.corrector.testing;

import cz.cuni.mff.corrector.corrector.Corrector;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.tree.Tree;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * Analysis dealing with task calls in case of the REFINE strategy.
 * 
 * @author Martin Svoboda
 */
public class AnalysisCalls extends Analysis {
    
    // -------------------------------------------------------------------------
    
    
    // Analysis configuration attributes
    private final String analysisName = "calls";
    private final int grammarId = 0;
    private final int sizeBase = 1000;
    private final int sizeIncrement = 1000;
    private final int sizeVariants = 10;
    private final int treeInstances = 10;
    private final int averageFanout = 5;

    // Algorithm strategies selection
    private final Corrector.AlgorithmStrategy strategy =
        Corrector.AlgorithmStrategy.REFINE;

    // Corrector configuration attributes
    private final Corrector.ExecutionApproach approach =
        Corrector.ExecutionApproach.SCHEDULING_MULTIPLE;
    
    // Signature handling selection
    Corrector.SignatureHandling signature =
        Corrector.SignatureHandling.ENABLED;
    
    // Enabled profiling selection
    private final boolean profileTimes = false;
    private final boolean profileStats = true;
    
    
    // -------------------------------------------------------------------------
    
    
    public void run() {
        
        // Opening analysis notifications
        System.out.println("Calls Profiling - Starting Analysis");
        
        // Acquiring grammar instance
        Grammar grammarObject = acquireGrammarInstance(grammarId);
        
        // Preparing the range of trees
        List<Integer> treeSizes = new ArrayList<Integer>();
        int currentSize = sizeBase;
        for (int s = 0; s < sizeVariants; s++, currentSize += sizeIncrement) {
            treeSizes.add(currentSize);
        }
        
        // Acquiring data tree instances
        Map<Integer, List<Tree>> dataTrees = new TreeMap<Integer, List<Tree>>();
        for (int s : treeSizes) {
            List<Tree> instances = new ArrayList<Tree>(treeInstances);
            for (int t = 0; t < treeInstances; t++) {
                instances.add(acquireDataTree(analysisName, grammarId, s, t));
            }
            dataTrees.put(s, instances);
        }
        
        String configLabel =
            Corrector.composeAlgorithmCode(strategy, approach, signature);
        System.out.println(
            "Starting configuration [" + configLabel + "]"
        );
        
        // Preparing container for final reports
        // Indexing: size -> report
        Map<Integer, Report> analysisReports = new TreeMap<Integer, Report>();
        
        // Computing analysis reports
        computeAnalysisReports(
            grammarObject, dataTrees, treeSizes, analysisReports
        );
        
        // Generating output table
        exportOutputTable(treeSizes, analysisReports);
        
        // Closing analysis notifications
        System.out.println("Calls Profiling - Analysis Finished");
        System.out.println();
        System.out.println();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Performs the analysis and computes reports.
     * @param grammarObject Reference to the associated grammar.
     * @param dataTrees Container with loaded data trees.
     * @param treeSizes List of considered tree sizes.
     * @param analysisReports Output container for computed reports.
     */
    private void computeAnalysisReports(
        Grammar grammarObject,
        Map<Integer, List<Tree>> dataTrees,
        List<Integer> treeSizes,
        Map<Integer, Report> analysisReports
    ) {
        
        // Iterating over tree sizes
        for (int s : treeSizes) {
            
            System.out.println("Processing data trees [" + s + "]");
            
            // Preparing container for instance reports
            List<Report> tReports = new ArrayList<Report>();

            // Iterating over tree instances
            for (int t = 0; t < treeInstances; t++) {
                System.out.println(
                    "Correcting data tree [" + s + "-" + t + "]"
                );
                // Invoking data tree correction
                Corrector ci = Corrector.createCorrectorInstance(
                    dataTrees.get(s).get(t), grammarObject,
                    strategy, approach, signature,
                    profileTimes, profileStats
                );
                // Fetching the profiling report
                tReports.add(ci.getProfilingReport());
            }

            // Aggregating reports over instances
            System.out.println("Aggregating reports");
            Report aReport = Report.createAggregatedReport(tReports);
            analysisReports.put(s, aReport);

        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Exports the computed reports into the table type output.
     * @param treeSizes List of considered tree sizes.
     * @param analysisReports Container with computed reports.
     */
    private void exportOutputTable(
        List<Integer> treeSizes,
        Map<Integer, Report> analysisReports
    ) {
        
        // Preparing the intents table output
        String outputName = "analysis-calls-table";
        StringBuilder sb = new StringBuilder();
        
        System.out.println("Computing output [" + outputName + "]");
        
        // Output columns:
        //   For each tree size (number of nodes) one column
        // Output rows:
        //   Overall number of created tasks
        //   Overall number of explored edges
        //   Overall number of task executions
        
        sb.append("NdsCnt");
        for (int s : treeSizes) {
            sb.append("\t");
            sb.append(s);
        }
        sb.append(Analysis.EOL);
        
        // Printing overall task numbers
        sb.append("TskCnt");
        for (int s : treeSizes) {
            sb.append("\t");
            sb.append(analysisReports.get(s).getTaskCounter());
        }
        sb.append(Analysis.EOL);
        
        // Printing explored edges numbers
        sb.append("EdgCnt");
        for (int s : treeSizes) {
            sb.append("\t");
            sb.append(analysisReports.get(s).getEdgesCounter());
        }
        sb.append(Analysis.EOL);
        
        // Printing execution calls numbers
        sb.append("ExeCnt");
        for (int s : treeSizes) {
            sb.append("\t");
            sb.append(analysisReports.get(s).getExecutionsCounter());
        }
        sb.append(Analysis.EOL);
        
        System.out.println("Exporting output [" + outputName + "]");
        
        // Exporting the analysis output
        exportComputedAnalysis(outputName, sb);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prepares all the required testing data trees.
     */
    public void generateDataTrees() {
        System.out.println("Calls Profiling - Generating Trees");
        Generator g = new Generator();
        int currentSize = this.sizeBase;
        for (
            int s = 0;
            s < this.sizeVariants;
            s++, currentSize += this.sizeBase
        ) {
            g.generateDataTrees(
                analysisName,
                this.treeInstances,
                currentSize,
                this.averageFanout
            );
        }
    }
    
    
    // -------------------------------------------------------------------------
    
}
