
package cz.cuni.mff.corrector.testing;

import cz.cuni.mff.corrector.corrector.Corrector;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.tree.Tree;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;

/**
 * Analysis dealing with average size of explored multigraphs.
 * 
 * @author Martin Svoboda
 */
public class AnalysisGraphs extends Analysis {
    
    // -------------------------------------------------------------------------
    
    
    // Analysis configuration attributes
    private final String analysisName = "graphs";
    private final int grammarId = 0;
    private final int sizeBase = 1000;
    private final int treeInstances = 10;
    private final int averageFanout = 5;

    // Algorithm strategies selection
    private final EnumSet<Corrector.AlgorithmStrategy> strategies =
        EnumSet.of(
            Corrector.AlgorithmStrategy.DEFAULT,
            Corrector.AlgorithmStrategy.EXPLORE,
            Corrector.AlgorithmStrategy.REFINE
        );

    // Corrector configuration attributes
    private final Corrector.ExecutionApproach approach =
        Corrector.ExecutionApproach.SCHEDULING_MULTIPLE;
    
    // Signature handling selection
    Corrector.SignatureHandling signature =
        Corrector.SignatureHandling.ENABLED;
    
    // Enabled profiling selection
    private final boolean profileTimes = false;
    private final boolean profileStats = true;
    
    
    // -------------------------------------------------------------------------
    
    
    public void run() {
        
        // Opening analysis notifications
        System.out.println("Graphs Profiling - Starting Analysis");
        
        // Acquiring grammar instance
        Grammar grammarObject = acquireGrammarInstance(grammarId);
        
        // Acquiring data tree instances
        List<Tree> dataTrees = new ArrayList<Tree>(treeInstances);
        for (int t = 0; t < treeInstances; t++) {
            dataTrees.add(
                acquireDataTree(analysisName, grammarId, sizeBase, t)
            );
        }
        
        // Preparing container for final reports
        // Indexing: strategy -> report
        Map<Corrector.AlgorithmStrategy, Report> analysisReports =
            new EnumMap<Corrector.AlgorithmStrategy, Report>
                (Corrector.AlgorithmStrategy.class);
        
        // Computing analysis reports
        computeAnalysisReports(grammarObject, dataTrees, analysisReports);
        
        // Generating output table
        exportOutputTable(analysisReports);
        
        // Closing analysis notifications
        System.out.println("Graphs Profiling - Analysis Finished");
        System.out.println();
        System.out.println();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Performs the analysis and computes reports.
     * @param grammarObject Reference to the associated grammar.
     * @param dataTrees Container with loaded data trees.
     * @param analysisReports Output container for computed reports.
     */
    private void computeAnalysisReports(
        Grammar grammarObject,
        List<Tree> dataTrees,
        Map<Corrector.AlgorithmStrategy, Report> analysisReports
    ) {
        
        // Iterating over algorithm strategies
        for (Corrector.AlgorithmStrategy strategy : strategies) {
            
            String configLabel =
                Corrector.composeAlgorithmCode(strategy, approach, signature);
            System.out.println(
                "Starting configuration [" + configLabel + "]"
            );

            // Preparing container for instance reports
            List<Report> tReports = new ArrayList<Report>();

            // Iterating over tree instances
            for (int t = 0; t < treeInstances; t++) {
                System.out.println(
                    "Correcting data tree [" + t +
                    "] config [" + configLabel + "]"
                );
                // Invoking data tree correction
                Corrector ci = Corrector.createCorrectorInstance(
                    dataTrees.get(t), grammarObject,
                    strategy, approach, signature,
                    profileTimes, profileStats
                );
                // Fetching the profiling report
                tReports.add(ci.getProfilingReport());
            }

            // Aggregating reports over instances
            System.out.println("Aggregating reports");
            Report aReport = Report.createAggregatedReport(tReports);
            analysisReports.put(strategy, aReport);

        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Exports the computed reports into the table type output.
     * @param analysisReports Container with computed reports.
     */
    private void exportOutputTable(
        Map<Corrector.AlgorithmStrategy, Report> analysisReports
    ) {
        
        // Preparing the intents table output
        String outputName = "analysis-graphs-table";
        StringBuilder sb = new StringBuilder();
        
        System.out.println("Computing output [" + outputName + "]");
        
        // Output columns:
        //   Algorithm strategy code
        //   Overall number of tasks
        //   Explored vertices: overall number over all tasks, average on task
        //   Explored edges: overall number over all tasks, average on task
        
        sb.append("CorStg\t");
        sb.append("TskCnt\t");
        sb.append("VtcCnt\t");
        sb.append("VtcAvg\t");
        sb.append("EdgCnt\t");
        sb.append("EdgAvf");
        sb.append(Analysis.EOL);
        
        // Iterating over strategies and modes
        for (Corrector.AlgorithmStrategy strategy : strategies) {

            Report aReport = analysisReports.get(strategy);

            sb.append(strategy.codeName);
            sb.append("\t");
            
            sb.append(aReport.getTaskCounter());
            sb.append("\t");
            
            sb.append(aReport.getVerticesCounter());
            sb.append("\t");
            BigDecimal averageVertices = new BigDecimal(
                (double)aReport.getVerticesCounter()/
                (double)aReport.getTaskCounter()
            );
            averageVertices = averageVertices.setScale(1, RoundingMode.HALF_UP);
            sb.append(String.format("%s", averageVertices.doubleValue()));
            sb.append("\t");
            
            sb.append(aReport.getEdgesCounter());
            sb.append("\t");
            BigDecimal averageEdes = new BigDecimal(
                (double)aReport.getEdgesCounter() /
                (double)aReport.getTaskCounter()
            );
            averageEdes = averageEdes.setScale(1, RoundingMode.HALF_UP);
            sb.append(String.format("%s", averageEdes.doubleValue()));
            
            sb.append(Analysis.EOL);

        }
        
        System.out.println("Exporting output [" + outputName + "]");
        
        // Exporting the analysis output
        exportComputedAnalysis(outputName, sb);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prepares all the required testing data trees.
     */
    public void generateDataTrees() {
        System.out.println("Graphs Profiling - Generating Trees");
        Generator g = new Generator();
        g.generateDataTrees(
            analysisName,
            this.treeInstances,
            this.sizeBase,
            this.averageFanout
        );
    }
    
    
    // -------------------------------------------------------------------------
    
}
