
package cz.cuni.mff.corrector.testing;

import cz.cuni.mff.corrector.corrector.Corrector;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.tree.Tree;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * Analysis of execution times using the RFN-N1-E configuration.
 * 
 * @author Martin Svoboda
 */
public class AnalysisScaling extends Analysis {
    
    // -------------------------------------------------------------------------
    
    
    // Analysis configuration attributes
    private final String analysisName = "scaling";
    private final int grammarId = 0;
    private final int sizeBase = 10000;
    private final int sizeIncrement = 10000;
    private final int sizeVariants = 10;
    private final int treeInstances = 20;
    private final int repeatExtra = 5;
    private final int repeatCount = 5;
    private final int averageFanout = 12;

    // Algorithm strategy selection
    private final Corrector.AlgorithmStrategy strategy =
        Corrector.AlgorithmStrategy.REFINE;
    
    // Execution approach selection
    private final Corrector.ExecutionApproach approach =
        Corrector.ExecutionApproach.NESTING_SINGLE;
    
    // Signature handling selection
    private final Corrector.SignatureHandling signature =
        Corrector.SignatureHandling.ENABLED;
    
    // Enabled profiling selection
    private final boolean profileTimes = true;
    private final boolean profileStats = false;
    
    
    // -------------------------------------------------------------------------
    
    
    public void run() {
        
        // Opening analysis notifications
        System.out.println("Scaling Profiling - Starting Analysis");
        
        // Acquiring grammar instance
        Grammar grammarObject = acquireGrammarInstance(grammarId);
        
        // Preparing the range of trees
        List<Integer> treeSizes = new ArrayList<Integer>();
        int currentSize = sizeBase;
        for (int s = 0; s < sizeVariants; s++, currentSize += sizeIncrement) {
            treeSizes.add(currentSize);
        }
        
        // Preparing container for final reports
        // Indexing: size -> report
        Map<Integer, Report> analysisReports = new TreeMap<Integer, Report>();
        
        // Computing analysis reports
        computeAnalysisReports(grammarObject, treeSizes, analysisReports);
        
        // Generating output table
        exportOutputTable(treeSizes, analysisReports);
        
        // Generating output figure
        exportOutputFigure(treeSizes, analysisReports);
        
        // Closing analysis notifications
        System.out.println("Scaling Profiling - Analysis Finished");
        System.out.println();
        System.out.println();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Performs the analysis and grammar reports.
     * @param grammarObject Reference to the associated grammar.
     * @param treeSizes List of considered tree sizes.
     * @param analysisReports Output container for computed reports.
     */
    private void computeAnalysisReports(
        Grammar grammarObject,
        List<Integer> treeSizes,
        Map<Integer, Report> analysisReports
    ) {
        
        String configLabel = Corrector.composeAlgorithmCode(
            strategy, approach, signature
        );
        System.out.println(
            "Starting configuration [" + configLabel + "]"
        );

        // Iterating over tree sizes
        for (int s : treeSizes) {

            System.out.println(
                "Processing tree size [" + s + "]"
            );

            // Preparing container for instance reports
            List<Report> tReports = new ArrayList<Report>();

            // Iterating over tree instances
            for (int t = 0; t < treeInstances; t++) {
                
                Tree dataTree = acquireDataTree(analysisName, grammarId, s, t);

                System.out.println(
                    "Processing data tree [" + s + "-" + t + "]"
                );

                // Preparing container over repetitions
                List<Report> rReports = new ArrayList<Report>();

                // Repeating the tree correction
                for (int r = 0; r < repeatCount + repeatExtra; r++) {
                    System.out.println(
                        "Correcting data tree [" +
                        s + "-" + t + "] config [" +
                        configLabel + "] " +
                        ((r < repeatExtra) ? "warming" : "attempt") +
                        " [" + r + "]"
                    );
                    // Invoking data tree correction
                    Corrector ci = Corrector.createCorrectorInstance(
                        dataTree, grammarObject,
                        strategy, approach, signature,
                        profileTimes, profileStats
                    );
                    // Fetching the profiling report
                    if (r >= repeatExtra) {
                        rReports.add(ci.getProfilingReport());
                    }
                }

                // Aggregating reports over repetitions
                System.out.println(
                    "Aggregating data tree reports [" +
                    s + "-" + t + "]"
                );
                Report aReport = Report.createAggregatedReport(rReports);
                tReports.add(aReport);
            }

            // Aggregating reports over instances
            System.out.println(
                "Aggregating tree size reports [" + s + "]"
            );
            Report aReport = Report.createAggregatedReport(tReports);
            analysisReports.put(s, aReport);

        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Exports the computed reports into the table type output.
     * @param treeSizes List of considered tree sizes.
     * @param analysisReports Container with computed reports.
     */
    private void exportOutputTable(
        List<Integer> treeSizes,
        Map<Integer, Report> analysisReports
    ) {
        
        // Preparing the intents table output
        String outputName = "analysis-scaling-table";
        StringBuilder sb = new StringBuilder();
        
        System.out.println("Computing output [" + outputName + "]");
        
        // Output columns:
        //   For each tree size (number of nodes) one column
        // Output rows:
        //   Only one row for RFN-N1-E configuration
        // Cell values:
        //   Execution times in miliseconds
        
        for (int size : treeSizes) {
            sb.append(size);
            sb.append("\t");
        }
        sb.deleteCharAt(sb.length() - 1);
        sb.append(Analysis.EOL);

        // Iterating over tree sizes
        for (int size : treeSizes) {
            Report aReport = analysisReports.get(size);
            sb.append(aReport.getExecutionTime());
            sb.append("\t");
        }
        sb.deleteCharAt(sb.length() - 1);
        sb.append(Analysis.EOL);
        
        System.out.println("Exporting output [" + outputName + "]");
        
        // Exporting the analysis output
        exportComputedAnalysis(outputName, sb);
        
    }
    
    
    /**
     * Exports the computed reports into the figure type output.
     * @param treeSizes List of considered tree sizes.
     * @param analysisReports Container with computed reports.
     */
    private void exportOutputFigure(
        List<Integer> treeSizes,
        Map<Integer, Report> analysisReports
    ) {
        
        // Preparing the intents table output
        String outputName = "analysis-scaling-figure";
        StringBuilder sb = new StringBuilder();
        
        System.out.println("Computing output [" + outputName + "]");
        
        sb.append("NdsCnt\t");
        sb.append(strategy.codeName);
        sb.append(approach.codeName);
        sb.append(signature.codeName);
        sb.append(Analysis.EOL);
        
        // Iterating over tree sizes
        for (int size : treeSizes) {
            sb.append(size);
            Report aReport = analysisReports.get(size);
            sb.append("\t");
            sb.append(aReport.getExecutionTime());
            sb.append(Analysis.EOL);
        }
        
        System.out.println("Exporting output [" + outputName + "]");
        
        // Exporting the analysis output
        exportComputedAnalysis(outputName, sb);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prepares all the required testing data trees.
     */
    public void generateDataTrees() {
        System.out.println("Scaling Profiling - Generating Trees");
        Generator g = new Generator();
        int currentSize = this.sizeBase;
        for (
            int s = 0;
            s < this.sizeVariants;
            s++, currentSize += this.sizeBase
        ) {
            g.generateDataTrees(
                analysisName,
                this.treeInstances,
                currentSize,
                this.averageFanout
            );
        }
    }
    
    
    // -------------------------------------------------------------------------
    
}
