
package cz.cuni.mff.corrector.testing;

import cz.cuni.mff.corrector.corrector.Corrector;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.intent.Intent;
import cz.cuni.mff.corrector.tree.Tree;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Analysis dealing with task numbers and distribution.
 * 
 * @author Martin Svoboda
 */
public class AnalysisTasks extends Analysis {
    
    // -------------------------------------------------------------------------
    
    
    // Analysis configuration attributes
    private final String analysisName = "tasks";
    private final int grammarId = 0;
    private final int sizeBase = 100;
    private final int treeInstances = 10;
    private final int averageFanout = 4;

    // Algorithm strategies selection
    private final EnumSet<Corrector.AlgorithmStrategy> strategies =
        EnumSet.of(
            Corrector.AlgorithmStrategy.DEFAULT,
            Corrector.AlgorithmStrategy.EXPLORE,
            Corrector.AlgorithmStrategy.REFINE
        );

    // Corrector configuration attributes
    private final Corrector.ExecutionApproach approach =
        Corrector.ExecutionApproach.SCHEDULING_MULTIPLE;
    
    // Signature handling selection
    private final EnumSet<Corrector.SignatureHandling> signatures =
        EnumSet.of(
            Corrector.SignatureHandling.DISABLED,
            Corrector.SignatureHandling.ENABLED
        );
    
    // Enabled profiling selection
    private final boolean profileTimes = false;
    private final boolean profileStats = true;
    
    
    // -------------------------------------------------------------------------
    
    
    public void run() {
        
        // Opening analysis notifications
        System.out.println("Tasks Profiling - Starting Analysis");
        
        // Acquiring grammar instance
        Grammar grammarObject = acquireGrammarInstance(grammarId);
        
        // Acquiring data tree instances
        List<Tree> dataTrees = new ArrayList<Tree>(treeInstances);
        for (int t = 0; t < treeInstances; t++) {
            dataTrees.add(
                acquireDataTree(analysisName, grammarId, sizeBase, t)
            );
        }
        
        // Preparing container for final reports
        // Indexing: strategy -> signature -> report
        TasksReportsContainer analysisReports =
            new TasksReportsContainer(strategies);
        
        // Computing analysis reports
        computeAnalysisReports(grammarObject, dataTrees, analysisReports);
        
        // Generating output table
        exportOutputTable(analysisReports);
        
        // Closing analysis notifications
        System.out.println("Tasks Profiling - Analysis Finished");
        System.out.println();
        System.out.println();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Performs the analysis and computes reports.
     * @param grammarObject Reference to the associated grammar.
     * @param dataTrees Container with loaded data trees.
     * @param analysisReports Output container for computed reports.
     */
    private void computeAnalysisReports(
        Grammar grammarObject,
        List<Tree> dataTrees,
        TasksReportsContainer analysisReports
    ) {
        
        // Iterating over strategies and modes
        for (Corrector.AlgorithmStrategy strategy : strategies) {
            for (Corrector.SignatureHandling signature : signatures) {
                
                String configLabel = Corrector.composeAlgorithmCode(
                    strategy, approach, signature
                );
                System.out.println(
                    "Starting configuration [" + configLabel + "]"
                );
        
                // Preparing container for instance reports
                List<Report> tReports = new ArrayList<Report>();

                // Iterating over tree instances
                for (int t = 0; t < treeInstances; t++) {
                    System.out.println(
                        "Correcting data tree [" + t +
                        "] config [" + configLabel + "]"
                    );
                    // Invoking data tree correction
                    Corrector ci = Corrector.createCorrectorInstance(
                        dataTrees.get(t), grammarObject,
                        strategy, approach, signature,
                        profileTimes, profileStats
                    );
                    // Fetching the profiling report
                    tReports.add(ci.getProfilingReport());
                }

                // Aggregating reports over instances
                System.out.println("Aggregating reports");
                Report aReport = Report.createAggregatedReport(tReports);
                analysisReports.get().get(strategy).put(signature, aReport);
                
            }
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Exports the computed reports into the table type output.
     * @param analysisReports Container with computed reports.
     */
    private void exportOutputTable(
        TasksReportsContainer analysisReports
    ) {
        
        // Preparing the intents table output
        String outputName = "analysis-tasks-table";
        StringBuilder sb = new StringBuilder();
        
        System.out.println("Computing output [" + outputName + "]");
        
        // Output columns:
        //   Algorithm strategy code
        //   Signature handling mode
        //   Overall number of tasks over all types of intents
        //   Nuber of tasks over particular types of intents
        
        sb.append("CorStg\t");
        sb.append("CorSig\t");
        sb.append("TskAll");
        for (Intent.Type intentType : Intent.Type.values()) {
            sb.append("\t");
            sb.append("Tsk");
            sb.append(intentType.fullCode);
        }
        sb.append(Analysis.EOL);
        
        // Iterating over strategies and modes
        for (Corrector.AlgorithmStrategy strategy : strategies) {
            for (Corrector.SignatureHandling signature : signatures) {
                
                Report aReport = analysisReports.get().get(strategy).get(signature);
                
                sb.append(strategy.codeName);
                sb.append("\t");
                sb.append(signature.codeName);
                sb.append("\t");
                sb.append(aReport.getTaskCounter());
                for (Intent.Type intentType : Intent.Type.values()) {
                    sb.append("\t");
                    sb.append(aReport.getTaskCounters(intentType));
                }
                sb.append(Analysis.EOL);
                
            }
        }
        
        System.out.println("Exporting output [" + outputName + "]");
        
        // Exporting the analysis output
        exportComputedAnalysis(outputName, sb);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prepares all the required testing data trees.
     */
    public void generateDataTrees() {
        System.out.println("Tasks Profiling - Generating Trees");
        Generator g = new Generator();
        g.generateDataTrees(
            analysisName,
            this.treeInstances,
            this.sizeBase,
            this.averageFanout
        );
    }
    
    
    // -------------------------------------------------------------------------
    
}


// =============================================================================


/**
 * Internal container for storing the analysis reports.
 * @author Martin Svoboda
 */
class TasksReportsContainer {
    
    // -------------------------------------------------------------------------
    
    /**
     * Internal container for storing computed reports.
     * Indexing: strategy -> approach -> size -> report
     */
    Map<
        Corrector.AlgorithmStrategy,
        Map<Corrector.SignatureHandling, Report>
    > reportsMap;
    
    // -------------------------------------------------------------------------
    
    /**
     * Creates a new instance of the reports container.
     * @param strategies Set of all considered algorithm strategies.
     */
    public TasksReportsContainer(Set<Corrector.AlgorithmStrategy> strategies) {
        this.reportsMap =
            new EnumMap<
                Corrector.AlgorithmStrategy,
                Map<Corrector.SignatureHandling, Report>
            >(Corrector.AlgorithmStrategy.class);
        for (Corrector.AlgorithmStrategy strategy : strategies) {
            this.reportsMap.put(
                strategy,
                new EnumMap<Corrector.SignatureHandling, Report>(
                    Corrector.SignatureHandling.class
                )
            );
        }
    }
    
    // -------------------------------------------------------------------------
    
    /**
     * Returns the reference to the reports container.
     * @return Reference to the internal reports map.
     */
    public Map<
        Corrector.AlgorithmStrategy,
        Map<Corrector.SignatureHandling, Report>
    >
    get() {
        return this.reportsMap;
    }
    
    // -------------------------------------------------------------------------
    
}
