
package cz.cuni.mff.corrector.testing;

import cz.cuni.mff.corrector.corrector.Corrector;
import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.tree.Tree;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * Analysis of execution times depending on configuration and sizes.
 * 
 * @author Martin Svoboda
 */
public class AnalysisTimes extends Analysis {
    
    // -------------------------------------------------------------------------
    
    
    // Analysis configuration attributes
    private final String analysisName = "times";
    private final int grammarId = 0;
    private final int sizeBase = 1000;
    private final int sizeIncrement = 1000;
    private final int sizeVariants = 10;
    private final int treeInstances = 10;
    private final int repeatExtra = 5;
    private final int repeatCount = 10;
    private final int averageFanout = 7;

    // Algorithm strategies selection
    private final EnumSet<Corrector.AlgorithmStrategy> strategies =
        EnumSet.of(
            //Corrector.AlgorithmStrategy.DEFAULT,
            Corrector.AlgorithmStrategy.EXPLORE,
            Corrector.AlgorithmStrategy.REFINE
        );
    
    // Execution approaches selection
    private final EnumSet<Corrector.ExecutionApproach> approaches =
        EnumSet.of(
            Corrector.ExecutionApproach.NESTING_SINGLE,
            //Corrector.ExecutionApproach.INVOKING_SINGLE,
            //Corrector.ExecutionApproach.SCHEDULING_SINGLE,
            Corrector.ExecutionApproach.INVOKING_MULTIPLE,
            Corrector.ExecutionApproach.SCHEDULING_MULTIPLE
        );
    
    // Signature handling selection
    private final Corrector.SignatureHandling signature =
        Corrector.SignatureHandling.ENABLED;
    
    // Enabled profiling selection
    private final boolean profileTimes = true;
    private final boolean profileStats = false;
    
    
    // -------------------------------------------------------------------------
    
    
    public void run() {
        
        // Opening analysis notifications
        System.out.println("Times Profiling - Starting Analysis");
        
        // Acquiring grammar instance
        Grammar grammarObject = acquireGrammarInstance(grammarId);
        
        // Preparing the range of trees
        List<Integer> treeSizes = new ArrayList<Integer>();
        int currentSize = sizeBase;
        for (int s = 0; s < sizeVariants; s++, currentSize += sizeIncrement) {
            treeSizes.add(currentSize);
        }
        
        // Preparing container for final reports
        // Indexing: strategy -> approach -> size -> report
        TimesReportsContainer analysisReports = new TimesReportsContainer();
        
        // Computing analysis reports
        computeAnalysisReports(grammarObject, treeSizes, analysisReports);
        
        // Generating output table
        exportOutputTable(treeSizes, analysisReports);
        
        // Generating output figure
        exportOutputFigure(treeSizes, analysisReports);
        
        // Closing analysis notifications
        System.out.println("Times Profiling - Analysis Finished");
        System.out.println();
        System.out.println();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Performs the analysis and grammarObject reports.
     * @param grammarObject Reference to the associated grammar.
     * @param treeSizes List of considered tree sizes.
     * @param analysisReports Output container for computed reports.
     */
    private void computeAnalysisReports(
        Grammar grammarObject,
        List<Integer> treeSizes,
        TimesReportsContainer analysisReports
    ) {
        
        // Iterating over strategies and approaches
        for (Corrector.AlgorithmStrategy strategy : strategies) {
            Map<Corrector.ExecutionApproach, Map<Integer, Report>> sReports =
                new EnumMap<Corrector.ExecutionApproach, Map<Integer, Report>>
                (Corrector.ExecutionApproach.class);
            analysisReports.get().put(strategy, sReports);
            
            for (Corrector.ExecutionApproach approach : approaches) {
                Map<Integer, Report> pReports = new TreeMap<Integer, Report>();
                sReports.put(approach, pReports);
                
                String configLabel = Corrector.composeAlgorithmCode(
                    strategy, approach, signature
                );
                System.out.println(
                    "Starting configuration [" + configLabel + "]"
                );
                
                // Iterating over tree sizes
                for (int s : treeSizes) {
                    
                    System.out.println(
                        "Processing tree size [" + s + "]"
                    );
                    
                    // Preparing container for instance reports
                    List<Report> tReports = new ArrayList<Report>();

                    // Iterating over tree instances
                    for (int t = 0; t < treeInstances; t++) {
                        
                        System.out.println(
                            "Processing data tree [" + s + "-" + t + "]"
                        );
                        
                        Tree dataTree =
                            acquireDataTree(analysisName, grammarId, s, t);

                        // Preparing container over repetitions
                        List<Report> rReports = new ArrayList<Report>();

                        // Repeating the tree correction
                        for (int r = 0; r < repeatCount + repeatExtra; r++) {
                            System.out.println(
                                "Correcting data tree [" +
                                s + "-" + t + "] config [" +
                                configLabel + "] " +
                                ((r < repeatExtra) ? "warming" : "attempt") +
                                " [" + r + "]"
                            );
                            // Invoking data tree correction
                            Corrector ci = Corrector.createCorrectorInstance(
                                dataTree, grammarObject,
                                strategy, approach, signature,
                                profileTimes, profileStats
                            );
                            // Fetching the profiling report
                            if (r >= repeatExtra) {
                                rReports.add(ci.getProfilingReport());
                            }
                        }

                        // Aggregating reports over repetitions
                        System.out.println(
                            "Aggregating data tree reports [" +
                            s + "-" + t + "]"
                        );
                        Report aReport =
                            Report.createAggregatedReport(rReports);
                        tReports.add(aReport);
                    }
                    
                    // Aggregating reports over instances
                    System.out.println(
                        "Aggregating tree size reports [" + s + "]"
                    );
                    Report aReport = Report.createAggregatedReport(tReports);
                    pReports.put(s, aReport);
                    
                }
                
            }
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Exports the computed reports into the table type output.
     * @param treeSizes List of considered tree sizes.
     * @param analysisReports Container with computed reports.
     */
    private void exportOutputTable(
        List<Integer> treeSizes,
        TimesReportsContainer analysisReports
    ) {
        
        // Preparing the intents table output
        String outputName = "analysis-times-table";
        StringBuilder sb = new StringBuilder();
        
        System.out.println("Computing output [" + outputName + "]");
        
        // Output columns:
        //   Algorithm strategy code
        //   Execution approach code
        //   For each tree size (number of nodes) one column
        // Output rows:
        //   For each configuration (strategy, approach) one row
        // Cell values:
        //   Execution times in miliseconds
        
        sb.append("CorStg\t");
        sb.append("CorApp");
        for (int size : treeSizes) {
            sb.append("\t");
            sb.append(size);
        }
        sb.append(Analysis.EOL);
        
        // Iterating over strategies and approaches
        for (Corrector.AlgorithmStrategy strategy : strategies) {
            for (Corrector.ExecutionApproach approach : approaches) {
                
                sb.append(strategy.codeName);
                sb.append("\t");
                sb.append(approach.codeName);
                
                // Iterating over tree sizes
                for (int size : treeSizes) {
                        
                    Report aReport = analysisReports.get().
                        get(strategy).get(approach).get(size);

                    sb.append("\t");
                    sb.append(aReport.getExecutionTime());
                    
                }
                
                sb.append(Analysis.EOL);
                
            }
        }
        
        System.out.println("Exporting output [" + outputName + "]");
        
        // Exporting the analysis output
        exportComputedAnalysis(outputName, sb);
        
    }
    
    
    /**
     * Exports the computed reports into the figure type output.
     * @param treeSizes List of considered tree sizes.
     * @param anaysisReports Container with computed reports.
     */
    private void exportOutputFigure(
        List<Integer> treeSizes,
        TimesReportsContainer anaysisReports
    ) {
        
        // Preparing the intents table output
        String outputName = "analysis-times-figure";
        StringBuilder sb = new StringBuilder();
        
        System.out.println("Computing output [" + outputName + "]");
        
        sb.append("NdsCnt");
        for (Corrector.AlgorithmStrategy strategy : strategies) {
            for (Corrector.ExecutionApproach approach : approaches) {
                sb.append("\t");
                sb.append(strategy.codeName);
                sb.append(approach.codeName);
                sb.append(signature.codeName);
            }
        }
        sb.append(Analysis.EOL);
        
        // Iterating over tree sizes
        for (int size : treeSizes) {
            
            sb.append(size);

            // Iterating over strategies and approaches
            for (Corrector.AlgorithmStrategy strategy : strategies) {
                for (Corrector.ExecutionApproach approach : approaches) {

                    Report aReport =
                    anaysisReports.get().get(strategy).get(approach).get(size);

                    sb.append("\t");
                    sb.append(aReport.getExecutionTime());
                    
                }
            }
            
            sb.append(Analysis.EOL);

        }
        
        System.out.println("Exporting output [" + outputName + "]");
        
        // Exporting the analysis output
        exportComputedAnalysis(outputName, sb);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Prepares all the required testing data trees.
     */
    public void generateDataTrees() {
        System.out.println("Times Profiling - Generating Trees");
        Generator g = new Generator();
        int currentSize = this.sizeBase;
        for (
            int s = 0;
            s < this.sizeVariants;
            s++, currentSize += this.sizeBase
        ) {
            g.generateDataTrees(
                analysisName,
                this.treeInstances,
                currentSize,
                this.averageFanout
            );
        }
    }
    
    
    // -------------------------------------------------------------------------
    
}


// =============================================================================


/**
 * Internal container for storing the analysis reports.
 * @author Martin Svoboda
 */
class TimesReportsContainer {
    
    // -------------------------------------------------------------------------
    
    /**
     * Internal container for storing computed reports.
     * Indexing: strategy -> approach -> size -> report
     */
    Map<
        Corrector.AlgorithmStrategy,
        Map<Corrector.ExecutionApproach, Map<Integer, Report>>
    > reportsMap;
    
    // -------------------------------------------------------------------------
    
    /**
     * Creates a new instance of the reports container.
     * Only the first layer of the container is created.
     */
    public TimesReportsContainer() {
        this.reportsMap =
            new EnumMap<
                Corrector.AlgorithmStrategy,
                Map<Corrector.ExecutionApproach, Map<Integer, Report>>
            >(Corrector.AlgorithmStrategy.class);
    }
    
    // -------------------------------------------------------------------------
    
    /**
     * Returns the reference to the reports container.
     * @return Reference to the internal reports map.
     */
    public Map<
        Corrector.AlgorithmStrategy,
        Map<Corrector.ExecutionApproach, Map<Integer, Report>>
    >
    get() {
        return this.reportsMap;
    }
    
    // -------------------------------------------------------------------------
    
}
