
package cz.cuni.mff.corrector.testing;

import cz.cuni.mff.corrector.tree.Node;
import cz.cuni.mff.corrector.tree.NodeElement;
import cz.cuni.mff.corrector.tree.Stats;
import cz.cuni.mff.corrector.tree.Tree;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

/**
 * Generator of sample XML documents used for the experimental evaluation.
 * 
 * @author Martin Svoboda
 */
public class Generator {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Relative path from the home folder to testing documents.
     */
    public static final String PATH = "profiling/";
    
    
    /**
     * XML prologue string used when exporting generated trees.
     */
    public static final String PROLOGUE =
        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Fixed single type tree grammar for data trees generation.
     * This generator is not able to produce any other documents.
     */
    private static final String grammarDefinition =
        "A B\n" +
        "A a C.DA*\n" +
        "B b DB*\n" +
        "C c #empty\n" +
        "DA d C*\n" +
        "DB d A|B|C\n";
    
    
    /**
     * Identifier of the associated grammar.
     */
    private static final int grammarId = 0;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Initial seed for the generator of random values.
     * This seed is applied when each 
     */
    private static final long randomSeed = 123456;
    
    
    /**
     * Instance of the generator of random values.
     */
    private static final Random randomGenerator = new Random();
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates the requested number of data trees.
     * When called, the generator of random numbers is always initialised
     * with the configured seed value, which causes that the same results
     * will be provided having requested the same configuration.
     * Generated trees will be exported and stored in the target directory.
     * @param analysisName Analysis name to determine directory.
     * @param treeInstances Number of data trees to be generated.
     * @param requiredNodes Number of nodes each tree to have.
     * @param averageFanout Recommended fanout for nodes DA_d, DB_d, C_c.
     */
    public void generateDataTrees(
        String analysisName,
        int treeInstances,
        int requiredNodes,
        int averageFanout
    ) {
        
        // Initialising the random generator
        randomGenerator.setSeed(randomSeed);
        
        // Preparing statistics container
        List<Stats> treeStats = new ArrayList<Stats>(treeInstances);
        
        String baseName = grammarId + "-" + requiredNodes;
        
        // Generating requested data trees
        for (int i = 0; i < treeInstances; i++) {
            
            String treeName = baseName + "-" + i;
            System.out.println("Generating data tree [" + treeName + "]");
            
            // Generating a new data tree
            Tree treeObject = genereteNode_root(requiredNodes, averageFanout);
            
            System.out.println("Exporting data tree [" + treeName + "]");
            
            // Exporting the data tree
            String fileName = PATH + "trees-" + analysisName +
                "/Tree-" + treeName + ".xml";
            treeObject.exportToXmlFile(fileName, PROLOGUE);
            
            // Storing the tree statistics
            treeStats.add(treeObject.getTreeStatistics());
            
        }
        
        // Merging data tree statistics
        Stats averageStats = Stats.mergeTreeStatistics(treeStats);
        
        // Exporting data tree statistics
        exportTreeStatistics(analysisName, baseName, averageStats, treeStats);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates the root node of a new data tree.
     * Despite there are two options to choose, node B_b is always selected.
     * The required number of nodes must be set to at least 2.
     * @param requiredNodes Number of nodes the tree should have.
     * @param averageFanout Recommended fanout for specific nodes.
     * @return Newly generated data tree having the required size.
     */
    private Tree genereteNode_root(
        int requiredNodes,
        int averageFanout
    ) {
        
        // Generating B_b as the root node
        Node node_B_b = generateNode_B_b(requiredNodes - 1, averageFanout);
        
        // Returning the root node reference
        return new Tree(node_B_b);
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates a new B_b data tree node and recursively its content.
     * @param requiredNodes Number of nodes the tree should have.
     * Note that this number does not contain the subtree root node.
     * @param averageFanout Recommended fanout for specific nodes.
     * @return Root node of the entire generated subtree.
     */
    private Node generateNode_B_b(
        int requiredNodes,
        int averageFanout
    ) {
        
        // Creating the requested B_b node
        NodeElement parentNode = new NodeElement("b");
        
        // Determining number of DB_d nodes
        int nestedQuota = 2;
        int siblingNodes =
            resolveSiblingsCount(requiredNodes, averageFanout - 1, nestedQuota);
        
        // Computing the base allocation
        int baseAllocation =
            appointBaseAllocation(requiredNodes, siblingNodes, nestedQuota);
        int remainingNodes = requiredNodes - (siblingNodes * baseAllocation);
        
        // Distributing the unused nodes
        Map<Integer, Integer> extraAllocations =
            new HashMap<Integer, Integer>(siblingNodes);
        distributeExtraNodes(extraAllocations, remainingNodes, siblingNodes);
        
        // Generating particular DB_d nodes
        for (int i = 0; i < siblingNodes; i++) {
            // Computing the node request
            int requestedNodes =
                resolveRequiredNodes(baseAllocation, extraAllocations, i);
            // Recursively generating the node
            Node nestedNode = generateNode_DB_d(requestedNodes, averageFanout);
            // Incorporating the node into the tree
            parentNode.addChildNode(nestedNode);
        }
        
        // Returning the parent node
        return parentNode;
        
    }
    
    
    /**
     * Generates a new A_a data tree node and recursively its content.
     * @param requiredNodes Number of nodes the tree should have.
     * Note that this number does not contain the subtree root node.
     * @param averageFanout Recommended fanout for specific nodes.
     * @return Root node of the entire generated subtree.
     */
    private Node generateNode_A_a(
        int requiredNodes,
        int averageFanout
    ) {
        
        // Creating the requested A_a node
        NodeElement parentNode = new NodeElement("a");
        
        // Generating the first C_c node
        Node firstNode = generateNode_C_c();
        parentNode.addChildNode(firstNode);
        int availableNodes = requiredNodes - 1;
        
        // Determining number of DA_d nodes
        int nestedQuota = 1;
        int siblingNodes =
            resolveSiblingsCount(availableNodes, averageFanout, nestedQuota);
        
        // Computing the base allocation
        int baseAllocation =
            appointBaseAllocation(availableNodes, siblingNodes, nestedQuota);
        int remainingNodes = availableNodes - (siblingNodes * baseAllocation);
        
        /*DEBUG*/if (siblingNodes <= 0) return parentNode;
                
        // Distributing the unused nodes
        Map<Integer, Integer> extraAllocations =
            new HashMap<Integer, Integer>(siblingNodes);
        distributeExtraNodes(extraAllocations, remainingNodes, siblingNodes);
        
        // Generating particular DA_d nodes
        for (int i = 0; i < siblingNodes; i++) {
            // Computing the node request
            int requestedNodes =
                resolveRequiredNodes(baseAllocation, extraAllocations, i);
            // Recursively generating the node
            Node nestedNode = generateNode_DA_d(requestedNodes);
            // Incorporating the node into the tree
            parentNode.addChildNode(nestedNode);
        }
        
        // Returning the parent node
        return parentNode;
        
    }
    
    
    /**
     * Generates a new C_c data tree node and its empty content.
     * @return Root node of the entire generated subtree.
     */
    private Node generateNode_C_c() {
        
        // Creating the requested C_c node
        NodeElement parentNode = new NodeElement("c");
        
        // Returning the parent node
        return parentNode;
        
    }
    
    
    /**
     * Generates a new DA_d data tree node and recursively its content.
     * @param requiredNodes Number of nodes the tree should have.
     * Note that this number does not contain the subtree root node.
     * @param averageFanout Recommended fanout for specific nodes.
     * @return Root node of the entire generated subtree.
     */
    private Node generateNode_DA_d(
        int requiredNodes
    ) {
        
        // Creating the requested A_a node
        NodeElement parentNode = new NodeElement("d");
        
        // Generating all nested C_c nodes
        for (int i = 0; i < requiredNodes; i++) {
            Node nestedNode = generateNode_C_c();
            parentNode.addChildNode(nestedNode);
        }
        
        // Returning the parent node
        return parentNode;
        
    }
    
    
    /**
     * Generates a new DB_d data tree node and recursively its content.
     * @param requiredNodes Number of nodes the tree should have.
     * Note that this number does not contain the subtree root node.
     * @param averageFanout Recommended fanout for specific nodes.
     * @return Root node of the entire generated subtree.
     */
    private Node generateNode_DB_d(
        int requiredNodes,
        int averageFanout
    ) {
        
        // Creating the requested DB_d node
        NodeElement parentNode = new NodeElement("d");
        
        // Generating one of the allowed child nodes
        Node nestedNode;
        if (requiredNodes == 1) {
            // Randomly selecting between C_c and B_b
            if (randomGenerator.nextBoolean()) {
                // Generating a new C_c node
                nestedNode = generateNode_C_c();
            } else {
                // Generating a new B_b node
                nestedNode = generateNode_B_b(requiredNodes - 1, averageFanout);
            }
        } else {
            // Randomly selecting between B_b and A_a
            if (requiredNodes > averageFanout) {
                // Generating a new B_b node
                nestedNode = generateNode_B_b(requiredNodes - 1, averageFanout);
            } else {
                // Generating a new A_a node
                nestedNode = generateNode_A_a(requiredNodes - 1, averageFanout);
            }
        }
        parentNode.addChildNode(nestedNode);
        
        // Returning the parent node
        return parentNode;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Assigns the standard number of sibling nodes near the requested base.
     * @param requiredNodes Overall number of available nodes.
     * @param baseFanout Average expected number of nodes.
     * @param nestedQuota Minimal number of nodes required by subtrees.
     * @return Appointed number that is close to the given base.
     */
    private int resolveSiblingsCount(
        int requiredNodes,
        int baseFanout,
        int nestedQuota
    ) {
        
        // Number of nodes below/above the base
        int allowedSpan = baseFanout / 5;
        if (allowedSpan == 0) allowedSpan = 1;
        // Number of determined sibling nodes
        int siblingNodes = baseFanout +
            randomGenerator.nextInt(1 + allowedSpan * 2) - allowedSpan;
        if (siblingNodes < 0) siblingNodes = 1;
        // Reflecting the available nodes
        if (siblingNodes * nestedQuota > requiredNodes) {
            siblingNodes = requiredNodes / nestedQuota;
        }
        // Returning the determined fanout
        return siblingNodes;
        
    }
    
    
    /**
     * Computes the base allocation for a set of sibling nodes.
     * @param requiredNodes Overall number of available nodes.
     * @param siblingNodes Adjusted number of sibling nodes.
     * @param nestedQuota Minimal number of nodes required by subtrees.
     * @return Resolved standard base allocation of nodes per sibling.
     */
    private int appointBaseAllocation(
        int requiredNodes,
        int siblingNodes,
        int nestedQuota
    ) {
        int baseAllocation = requiredNodes / (siblingNodes + nestedQuota);
        if (baseAllocation < nestedQuota) {
            baseAllocation = nestedQuota;
        }
        return baseAllocation;
    }
    
    
    /**
     * Distributes the remaining nodes between 
     * @param extraAllocations Output map to store distributed allocations.
     * @param remainingNodes Number of nodes remaining to be allocated.
     * @param siblingNodes Number of sibling nodes receiving these allocations.
     */
    private void distributeExtraNodes(
        Map<Integer, Integer> extraAllocations,
        int remainingNodes,
        int siblingNodes
    ) {
        while (remainingNodes > 0) {
            int siblingIndex = randomGenerator.nextInt(siblingNodes);
            Integer currentExtra = extraAllocations.get(siblingIndex);
            int newExtra = randomGenerator.nextInt(remainingNodes) + 1;
            extraAllocations.put(
                siblingIndex,
                (currentExtra == null) ? newExtra : currentExtra + newExtra
            );
            remainingNodes -= newExtra;
        }
    }
    
    
    /**
     * Computes the number of nodes that should be requested recursively.
     * @param baseAllocation Base number of allocated nodes.
     * @param extraAllocations Map of assigned extra allocations.
     * @param inspectedItem Currently inspected sibling node.
     * @return Number of total assigned node allocation.
     */
    private int resolveRequiredNodes(
        int baseAllocation,
        Map<Integer, Integer> extraAllocations,
        int inspectedItem
    ) {
        Integer extraAllocation = extraAllocations.get(inspectedItem);
        if (extraAllocation == null) {
            extraAllocation = 0;
        }
        return baseAllocation + extraAllocation - 1;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Exports the given average statistics into the external file.
     * @param analysisName Analysis name to determine directory.
     * @param baseName Core part of the output document name.
     * @param averageStats Average statistics to be exported.
     * @param treeStats Particular data tree statistics.
     */
    public void exportTreeStatistics(
        String analysisName,
        String baseName,
        Stats averageStats,
        List<Stats> treeStats
    ) {
        
        // Preparing the output file name
        String fileName = PATH + "trees/" +
            "Trees-" + analysisName + "-" + baseName + ".txt";
        
        // Output columns:
        //   Overall number of nodes in a tree
        //   Minimal/maximal detected fanout
        //   Minimal/maximal detected depth
        // Output rows:
        //   Average statistics over all data trees
        //   All particular data tree statistics
        
        // Composing the output content
        StringBuilder stringBuilder = new StringBuilder();
        
        stringBuilder.append("NdsCnt\t");
        stringBuilder.append("FanMin\t");
        stringBuilder.append("FanMax\t");
        stringBuilder.append("DptMin\t");
        stringBuilder.append("DptMax");
        stringBuilder.append(Analysis.EOL);
        
        // Exporting the average statistics
        exportTreeStatistics(stringBuilder, averageStats);
        stringBuilder.append(Analysis.EOL);
        // Exporting all particular statistics
        for (Stats statsItem : treeStats) {
            exportTreeStatistics(stringBuilder, statsItem);
        }
        
        // Exporting the output content
        try {
            FileOutputStream fos = new FileOutputStream(fileName);
            DataOutputStream dos = new DataOutputStream(fos);
            dos.writeBytes(stringBuilder.toString());
            dos.close();
        } catch (FileNotFoundException e) {
            System.out.println("Unable to open output file");
            throw new Error();
        } catch (IOException e) {
            System.out.println("Unable to write statistics data");
            throw new Error();
        }
        
    }
    
    
    /**
     * Exports the particular data tree statistics into the output.
     * @param stringBuilder Reference to the output string builder.
     * @param inspectedStats Particular tree statistics to be processed.
     */
    public void exportTreeStatistics(
        StringBuilder stringBuilder,
        Stats inspectedStats
    ) {
        
        stringBuilder.append(inspectedStats.treeNodes);
        stringBuilder.append("\t");
        stringBuilder.append(inspectedStats.fanoutMin);
        stringBuilder.append("\t");
        stringBuilder.append(inspectedStats.fanoutMax);
        stringBuilder.append("\t");
        stringBuilder.append(inspectedStats.depthMin);
        stringBuilder.append("\t");
        stringBuilder.append(inspectedStats.depthMax);
        stringBuilder.append(Analysis.EOL);
        
    }
    
    
    // -------------------------------------------------------------------------
    
}
