
package cz.cuni.mff.corrector.testing;

import cz.cuni.mff.corrector.intent.Intent;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;

/**
 * Representation of profiling results computed over one corrector usage.
 * 
 * @author Martin Svoboda
 */
public class Report {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Overall execution time in miliseconds.
     */
    private long executionTime;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Overall number of explored edges.
     */
    private long edgeCounter;
    
    /**
     * Overall number of explored vertices.
     */
    private long vertexCounter;
    
    
    /**
     * Overall number of evaluated intents.
     * When signature handling is enabled, this number corresponds to the
     * number of unique signatures, but not to the number of explored edges!
     * When disabled, it corresponds to the number of explored edges.
     * This means the this counter should be incremented each and every
     * time a new distinct task is created, i.e. when a new intent is
     * requested to be evaluated.
     */
    private long taskCounter;
    
    /**
     * Overall number of evaluated intents split according to their types.
     */
    private final Map<Intent.Type, Long> taskCounters;
    
    
    /**
     * Overall number of executed runs of tasks.
     */
    private long executionsCounter;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new instance of a profiling report.
     */
    public Report() {
        
        this.executionTime = 0;
        this.edgeCounter = 0;
        this.vertexCounter = 0;
        this.taskCounter = 0;
        this.taskCounters = new EnumMap<Intent.Type, Long>(Intent.Type.class);
        for (Intent.Type intentType : Intent.Type.values()) {
            this.taskCounters.put(intentType, new Long(0));
        }
        this.executionsCounter = 0;
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the overall execution time.
     * @return Execution time in miliseconds.
     */
    public long getExecutionTime() {
        return this.executionTime;
    }
    
    
    /**
     * Sets the overall execution time.
     * @param executionTime Execution time in miliseconds to be set.
     */
    public synchronized void setExecutionTime(long executionTime) {
        this.executionTime = executionTime;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the value of the overall number of explored edges.
     * @return Number of explored edges.
     */
    public long getEdgesCounter() {
        return this.edgeCounter;
    }
    
    /**
     * Increases the current value of explored edges count by 1.
     */
    public synchronized void increaseEdgesCounter() {
        this.edgeCounter++;
    }
    
    
    /**
     * Returns the value of the overall number of explored vertices.
     * @return Number of explored vertices.
     */
    public long getVerticesCounter() {
        return this.vertexCounter;
    }
    
    /**
     * Increases the current value of explored vertices count by 1.
     */
    public synchronized void increaseVerticesCounter() {
        this.vertexCounter++;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Returns the value of the overall number of all required tasks.
     * @return Number of distinct intent signatures.
     */
    public long getTaskCounter() {
        return this.taskCounter;
    }
    
    /**
     * Returns the numbers of required tasks separated by intent types.
     * @return Map of task counters indexed by intent types.
     */
    public Map<Intent.Type, Long> getTaskCounters() {
        return this.taskCounters;
    }
    
    /**
     * Returns the numbers of required tasks for the given intent type.
     * @param intentType Associated intent type selection.
     * @return Task counter associated with the given type.
     */
    public Long getTaskCounters(Intent.Type intentType) {
        return this.taskCounters.get(intentType);
    }
    
    /**
     * Increases the current value of required tasks count by 1.
     * This method updates the overall counter as well as the specific one!
     * @param intentType Associated intent type selection.
     */
    public synchronized void increaseTaskCounters(Intent.Type intentType) {
        this.taskCounter++;
        this.taskCounters.put(
            intentType,
            this.taskCounters.get(intentType) + 1
        );
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Sets the value of the overall number of executed runs of tasks.
     */
    /**
     * Returns the value of the overall number of executed runs of tasks.
     * Note that this counter only works when signature handling is enabled!
     * @return Number of executed runs of tasks.
     */
    public long getExecutionsCounter() {
        return this.executionsCounter;
    }
    
    /**
     * Increases the current value of executed runs of tasks count by 1.
     */
    public synchronized void increaseExecutionsCounter() {
        this.executionsCounter++;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new and averaged report over the set of input reports.
     * Note that the implementation can be easy only because all the
     * internal components are numbers that can be summed up and later
     * on easily divided in order to obtain the overall average values
     * which are to goal of this method.
     * @param reportsList List of input reports to be aggregated.
     * @return Created aggregated report instance.
     */
    public static Report createAggregatedReport(List<Report> reportsList) {
        
        // Preparing the output report
        Report finalReport = new Report();
        
        // Iterating over reports and summing values of their attributes
        for (Report oneReport : reportsList) {
            finalReport.executionTime += oneReport.executionTime;
            finalReport.edgeCounter += oneReport.edgeCounter;
            finalReport.vertexCounter += oneReport.vertexCounter;
            finalReport.taskCounter += oneReport.taskCounter;
            for (Intent.Type intentType : Intent.Type.values()) {
                finalReport.taskCounters.put(
                    intentType,
                    finalReport.taskCounters.get(intentType) +
                        oneReport.taskCounters.get(intentType)
                );
            }
            finalReport.executionsCounter += oneReport.executionsCounter;
        }
        
        // Detecting the number of input reports
        int reportsCount = reportsList.size();
        
        // Computing the overall average values
        finalReport.executionTime /= reportsCount;
        finalReport.edgeCounter /= reportsCount;
        finalReport.vertexCounter /= reportsCount;
        finalReport.taskCounter /= reportsCount;
        for (Intent.Type intentType : Intent.Type.values()) {
            finalReport.taskCounters.put(
                intentType,
                finalReport.taskCounters.get(intentType) / reportsCount
            );
        }
        finalReport.executionsCounter /= reportsCount;
        
        // Returning the created aggregated report
        return finalReport;
        
    }
    
    
    // -------------------------------------------------------------------------
    
}
