
package cz.cuni.mff.corrector.tree;

import java.util.ArrayList;
import java.util.List;

/**
 * Representation of an element node in a data tree.
 *
 * @author Martin Svoboda
 */
public class NodeElement extends Node {

    // -------------------------------------------------------------------------


    /**
     * Element name.
     */
    private String labelName;

    /**
     * Ordered list of child nodes.
     */
    private List<Node> childNodes;


    // -------------------------------------------------------------------------


    /**
     * Creates a new instance of an element node.
     * @param labelName Element name.
     */
    public NodeElement(String labelName) {
        this.labelName = labelName;
        this.childNodes = new ArrayList<Node>();
    }


    // -------------------------------------------------------------------------


    /**
     * Returns the child number at a given relative position.
     * @param index Number of the relative position inside a node.
     * @return Requested child node.
     */
    public Node getChildNode(int index) {
        if (this.childNodes.size() > index) {
            return this.childNodes.get(index);
        } else {
            return null;
        }
    }


    public List<Node> getChildNodes() {
        return this.childNodes;
    }


    /**
     * Returns the specified subsequence of child nodes.
     * @param indexFrom Left inclusive boundary.
     * @param indexTo Right exclusive boundary.
     * @return Requested range of child nodes.
     */
    public List<Node> getChildNodes(int indexFrom, int indexTo) {
        return this.childNodes.subList(indexFrom, indexTo);
    }


    // -------------------------------------------------------------------------


    /**
     * Adds a new node as the last child node of this parent node.
     * @param childNode Reference to a new child node.
     */
    public void addChildNode(Node childNode) {
        this.childNodes.add(childNode);
    }


    // -------------------------------------------------------------------------


    public Type getNodeType() {
        return Type.NODE_ELEMENT;
    }


    public String getNodeLabel() {
        return this.labelName;
    }


    public int countTreeNodes() {
        int counter = 1;
        for (Node childNode : this.childNodes) {
            counter += childNode.countTreeNodes();
        }
        return counter;
    }
    
    
    public Stats getTreeStatistics() {
        
        // Node with no child nodes
        if (this.childNodes.isEmpty()) {
            return new Stats(1, 0, 0, 1, 1);
        }
        
        //Fetching nested statistics
        List<Stats> nestedStats = new ArrayList<Stats>();
        for (Node childNode : this.childNodes) {
            nestedStats.add(childNode.getTreeStatistics());
        }
        
        // Composing nested statistics
        Stats outputStats = Stats.mergeNestedStatistics(nestedStats);
        
        // Reflecting this node features
        outputStats.treeNodes += 1;
        if (
            (outputStats.fanoutMin > childNodes.size()) ||
            (outputStats.fanoutMin == 0)
        ) {
            outputStats.fanoutMin = childNodes.size();
        }
        if (
            (outputStats.fanoutMax < childNodes.size()) ||
            (outputStats.fanoutMax == 0)
        ) {
            outputStats.fanoutMax = childNodes.size();
        }
        outputStats.depthMin += 1;
        outputStats.depthMax += 1;
        
        // Returning aggregated statistics
        return outputStats;
        
    }


    // -------------------------------------------------------------------------


    public void serializeNode(StringBuilder builder) {
        if (this.childNodes.isEmpty()) {
            builder.append(Node.TOKEN_TG_OPEN_START);
            builder.append(this.getNodeLabel());
            builder.append(Node.TOKEN_TG_CLOSE_EMPTY);
        } else {
            builder.append(Node.TOKEN_TG_OPEN_START);
            builder.append(this.getNodeLabel());
            builder.append(Node.TOKEN_TG_CLOSE);
            for (Node childNode : this.childNodes) {
                childNode.serializeNode(builder);
            }
            builder.append(Node.TOKEN_TG_OPEN_END);
            builder.append(this.getNodeLabel());
            builder.append(Node.TOKEN_TG_CLOSE);
        }
    }


    public void serializeNode(StringBuilder builder, int level) {
        if (this.childNodes.isEmpty()) {
            Node.generateIndent(builder, level);
            builder.append(Node.TOKEN_TG_OPEN_START);
            builder.append(this.getNodeLabel());
            builder.append(Node.TOKEN_TG_CLOSE_EMPTY);
            builder.append(Node.TOKEN_SE_EOL);
        } else if (
            (this.childNodes.size() == 1) &&
            (this.getChildNode(0).getNodeType() == Type.NODE_DATA)
        ) {
            Node.generateIndent(builder, level);
            builder.append(Node.TOKEN_TG_OPEN_START);
            builder.append(this.getNodeLabel());
            builder.append(Node.TOKEN_TG_CLOSE);
            this.getChildNode(0).serializeNode(builder);
            builder.append(Node.TOKEN_TG_OPEN_END);
            builder.append(this.getNodeLabel());
            builder.append(Node.TOKEN_TG_CLOSE);
            builder.append(Node.TOKEN_SE_EOL);
        } else {
            Node.generateIndent(builder, level);
            builder.append(Node.TOKEN_TG_OPEN_START);
            builder.append(this.getNodeLabel());
            builder.append(Node.TOKEN_TG_CLOSE);
            builder.append(Node.TOKEN_SE_EOL);
            for (Node childNode : this.childNodes) {
                childNode.serializeNode(builder, level + 1);
            }
            Node.generateIndent(builder, level);
            builder.append(Node.TOKEN_TG_OPEN_END);
            builder.append(this.getNodeLabel());
            builder.append(Node.TOKEN_TG_CLOSE);
            builder.append(Node.TOKEN_SE_EOL);
        }
    }


    // -------------------------------------------------------------------------

}
