
package cz.cuni.mff.corrector.tree;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.List;

/**
 * Container for basic statistics about particular data trees.
 * 
 * @author Martin Svoboda
 */
public class Stats {
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Overall number of nodes.
     */
    public int treeNodes;
    
    
    /**
     * Minimal detected fanout.
     */
    public int fanoutMin;
    
    /**
     * Maximal detected fanout.
     */
    public int fanoutMax;
    
    
    /**
     * Minimal detected depth.
     */
    public int depthMin;
    
    /**
     * Maximal detected depth.
     */
    public int depthMax;
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Creates a new instance of data tree statistics container.
     * @param treeNodes Overall number of nodes.
     * @param fanoutMin Minimal detected fanout.
     * @param fanoutMax Maximal detected fanout.
     * @param depthMin Minimal detected depth.
     * @param depthMax Maximal detected depth.
     */
    public Stats(
        int treeNodes,
        int fanoutMin,
        int fanoutMax,
        int depthMin,
        int depthMax
    ) {
        this.treeNodes = treeNodes;
        this.fanoutMin = fanoutMin;
        this.fanoutMax = fanoutMax;
        this.depthMin = depthMin;
        this.depthMax = depthMax;
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Merges statistics from nested child nodes.
     * Note that at least one data tree statistic need to be provided.
     * @param nestedStats List of fetched nested statistics.
     * @return Computed aggregated statistics.
     */
    public static Stats mergeNestedStatistics(List<Stats> nestedStats) {
        
        // Composing nested statistics
        Stats outputStats = null;
        for (Stats nestedItem : nestedStats) {
            if (outputStats == null) {
                // Setting the first child as a base
                outputStats = nestedItem;
            } else {
                // Reflecting the remaining child nodes
                outputStats.treeNodes += nestedItem.treeNodes;
                if (
                    (outputStats.fanoutMin > nestedItem.fanoutMin) &&
                    (nestedItem.fanoutMin > 0)
                ) {
                    outputStats.fanoutMin = nestedItem.fanoutMin;
                }
                if (outputStats.fanoutMax < nestedItem.fanoutMax) {
                    outputStats.fanoutMax = nestedItem.fanoutMax;
                }
                if (outputStats.depthMin > nestedItem.depthMin) {
                    outputStats.depthMin = nestedItem.depthMin;
                }
                if (outputStats.depthMax < nestedItem.depthMax) {
                    outputStats.depthMax = nestedItem.depthMax;
                }
            }
        }
        
        // Returning the merged stats
        return outputStats;
        
    }
    
    
    /**
     * Merges statistics from several data trees and computes their averages.
     * Note that at least one data tree statistic need to be provided.
     * @param treeStats List of data tree statistics.
     * @return Computed aggregated statistics.
     */
    public static Stats mergeTreeStatistics(List<Stats> treeStats) {
        
        // Composing provided statistics
        Stats outputStats = new Stats(0, 0, 0, 0, 0);
        for (Stats statsItem : treeStats) {
            outputStats.treeNodes += statsItem.treeNodes;
            outputStats.fanoutMin += statsItem.fanoutMin;
            outputStats.fanoutMax += statsItem.fanoutMax;
            outputStats.depthMin += statsItem.depthMin;
            outputStats.depthMax += statsItem.depthMax;
        }
        
        // Average computation not required
        if (treeStats.size() == 1) {
            return outputStats;
        }
        
        // Computing the overall averages
        BigDecimal x;
        x = (new BigDecimal(
            (double)outputStats.treeNodes / (double)treeStats.size()
        )).setScale(0, RoundingMode.HALF_UP);
        outputStats.treeNodes = x.intValue();
        x = (new BigDecimal(
            (double)outputStats.fanoutMin / (double)treeStats.size()
        )).setScale(0, RoundingMode.HALF_UP);
        outputStats.fanoutMin = x.intValue();
        x = (new BigDecimal(
            (double)outputStats.fanoutMax / (double)treeStats.size()
        )).setScale(0, RoundingMode.HALF_UP);
        outputStats.fanoutMax = x.intValue();
        x = (new BigDecimal(
            (double)outputStats.depthMin / (double)treeStats.size()
        )).setScale(0, RoundingMode.HALF_UP);
        outputStats.depthMin = x.intValue();
        x = (new BigDecimal(
            (double)outputStats.depthMax / (double)treeStats.size()
        )).setScale(0, RoundingMode.HALF_UP);
        outputStats.depthMax = x.intValue();
        
        // Returning the merged stats
        return outputStats;
        
    }
    
    
    // -------------------------------------------------------------------------
    
}
