
package cz.cuni.mff.corrector.tree;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import org.xml.sax.*;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * Implementation of SAX parser for building data trees for XML documents.
 * 
 * @author Martin Svoboda
 */
public class XmlParser implements ContentHandler, ErrorHandler {

    // -------------------------------------------------------------------------


    /**
     * Reference to the parsed root node.
     */
    private Node rootNode;


    /**
     * Parsing locator instance.
     */
    private Locator saxLocator;

    /**
     * Stack of nodes
     */
    private List<NodeElement> nodesStack;

    /**
     * Buffer for storing partially parsed values of data nodes.
     */
    private StringBuilder valueBuilder;


    // -------------------------------------------------------------------------


    /**
     * Creates a new and empty parser object.
     */
    private XmlParser() {
        this.rootNode = null;
        this.nodesStack = new ArrayList<NodeElement>();
        this.valueBuilder = null;
    }
    

    // -------------------------------------------------------------------------


    /**
     * Parses provided XML document and returns its recursive node structure.
     * @param fileName Path of an XML document to be parsed.
     * @return Reference to the root node element.
     */
    public static Node parseXmlDocument(String fileName) {
        try {
            XmlParser builder = new XmlParser();
            XMLReader parser = XMLReaderFactory.createXMLReader();
            InputSource source = new InputSource(fileName);
            parser.setContentHandler(builder);
            parser.setErrorHandler(builder);
            parser.parse(source);
            return builder.rootNode;
        } catch (SAXException exception) {
            throw new Error(exception);
        } catch (IOException exception) {
            throw new Error(exception);
        }
    }
    

    // -------------------------------------------------------------------------

    
    public void setDocumentLocator(Locator locator) {
        this.saxLocator = locator;
    }


    // -------------------------------------------------------------------------

    
    public void startDocument()
    throws SAXException { }
    
    public void endDocument()
    throws SAXException { }

    public void startPrefixMapping(String prefix, String uri)
    throws SAXException { }

    public void endPrefixMapping(String prefix)
    throws SAXException { }


    // -------------------------------------------------------------------------


    public void startElement(
        String uri,
        String localName,
        String qName,
        Attributes atts
    ) throws SAXException {
        // Processing potentially existing data value
        if (this.valueBuilder != null) {
            createDataNode();
        }
        // Creating a new node object
        NodeElement nodeElem = new NodeElement(localName);
        // Adding a node to the top of the stack
        this.nodesStack.add(nodeElem);
    }

    public void endElement(
        String uri,
        String localName,
        String qName
    ) throws SAXException {
        // Processing potentially existing data value
        if (this.valueBuilder != null) {
            createDataNode();
        }
        // Removing the node from the stack
        int lastItem = this.nodesStack.size() - 1;
        NodeElement closedNode = this.nodesStack.get(lastItem);
        this.nodesStack.remove(lastItem);
        // Inserting node as the last child of parent
        if (lastItem > 0) {
            this.nodesStack.get(lastItem - 1).addChildNode(closedNode);
        // Otherwise storing the root node reference
        } else {
            this.rootNode = closedNode;
        }
    }


    // -------------------------------------------------------------------------


    public void characters(char[] ch, int start, int length)
    throws SAXException {
        // Preparing builder if not prepared
        if (this.valueBuilder == null) {
            this.valueBuilder = new StringBuilder();
        }
        // Storing the parsed value into the builder
        this.valueBuilder.append(ch, start, length);
    }


    /**
     * Internal function for creating a new data node for buffered value.
     * Warning: values with only white spaces are completely ignored!
     */
    private void createDataNode() {
        String dataString = this.valueBuilder.toString();
        dataString = dataString.trim();
        if (dataString.isEmpty()) {
            this.valueBuilder = null;
            return;
        }
        NodeData dataNode = new NodeData(dataString);
        this.nodesStack.get(this.nodesStack.size() - 1).addChildNode(dataNode);
        this.valueBuilder = null;
    }


    // -------------------------------------------------------------------------


    public void ignorableWhitespace(char[] ch, int start, int length)
    throws SAXException { }

    public void processingInstruction(String target, String data)
    throws SAXException { }

    public void skippedEntity(String name)
    throws SAXException { }


    // -------------------------------------------------------------------------


    public void warning(SAXParseException exception)
    throws SAXException { }

    public void error(SAXParseException exception)
    throws SAXException { }

    public void fatalError(SAXParseException exception)
    throws SAXException { }


    // -------------------------------------------------------------------------

}
