
package cz.cuni.mff.corrector.corrector;

import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.repair.Repair;
import cz.cuni.mff.corrector.tree.Tree;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Map;
import org.junit.Test;

/**
 * Testing class for the universal corrector implementation.
 *
 * @author Martin Svoboda
 */
public class CorrectorTest {

    // -------------------------------------------------------------------------


    /**
     * XML prologue string.
     */
    public static String PROLOGUE =
        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";

    /**
     * Relative path from the home folder to testing XML documents.
     */
    public static String PATH = "test/cz/cuni/mff/corrector/data/";
    
    
    // -------------------------------------------------------------------------


    /**
     * Tests the universal correction algorithm.
     */
    @Test
    public void testUniversalCorrector() {
        
        // Selecting input to be processed
        int lowerBoundary = 0;
        int upperBoundary = 5;
        
        // Selecting algorithm strategies       
        EnumSet<Corrector.AlgorithmStrategy> strategies =
            EnumSet.of(
                Corrector.AlgorithmStrategy.DEFAULT,
                Corrector.AlgorithmStrategy.EXPLORE,
                Corrector.AlgorithmStrategy.REFINE
            );
        
        // Selecting execution approaches
        EnumSet<Corrector.ExecutionApproach> approaches =
            EnumSet.of(
                Corrector.ExecutionApproach.NESTING_SINGLE,
                Corrector.ExecutionApproach.INVOKING_SINGLE,
                Corrector.ExecutionApproach.INVOKING_MULTIPLE,
                Corrector.ExecutionApproach.SCHEDULING_SINGLE,
                Corrector.ExecutionApproach.SCHEDULING_MULTIPLE
            );
        
        // Selecting signature modes
        EnumSet<Corrector.SignatureHandling> signatures =
            EnumSet.of(
                Corrector.SignatureHandling.DISABLED,
                Corrector.SignatureHandling.ENABLED
            );
        
        // Preparing input containers
        Map<Integer, Grammar> grammars = new HashMap<Integer, Grammar>();
        Map<Integer, Tree> trees = new HashMap<Integer, Tree>();
        
        // Loading trees and grammars
        for (int i = lowerBoundary; i <= upperBoundary; i++) {
            // Loading and parsing grammar
            String grammarFile = PATH + "Grammar-" + i + ".txt";
            grammars.put(i, Grammar.createFromDataFile(grammarFile));
            // Loading and parsing document
            String xmlFile = PATH + "Tree-" + i + ".xml";
            trees.put(i, Tree.createFromXmlFile(xmlFile));
        }
        
        // Iterating over strategies, approaches, modes
        for (Corrector.AlgorithmStrategy strategy : strategies) {
            for (Corrector.ExecutionApproach approach : approaches) {
                for (Corrector.SignatureHandling signature : signatures) {
                    
                    // Iterating over data trees
                    for (int i = lowerBoundary; i <= upperBoundary; i++) {
                        // Invoking data tree correction
                        correct(
                            strategy, approach, signature,
                            i, grammars.get(i), trees.get(i)
                        );
                    }
                    
                }
            }
        }
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Performs the correction according to the provided configuration.
     * @param strategy Selected algorithm strategy.
     * @param approach Selected execution approach.
     * @param signature Selecting signature handling mode.
     * @param index Index number of the tree to be processed.
     * @param grammar Parsed grammar to be used for correction.
     * @param tree Data tree to be corrected.
     */
    private void correct(
        Corrector.AlgorithmStrategy strategy,
        Corrector.ExecutionApproach approach,
        Corrector.SignatureHandling signature,
        int index,
        Grammar grammar,
        Tree tree
    ) {
        
        // Composing configuration code
        String algorithmCode =
            Corrector.composeAlgorithmCode(strategy, approach, signature);
        
        // Opening the correction output
        System.out.println(
            "UNIVERSAL CORRECTOR ALGORITHM :: " + algorithmCode + " :: " +
            "DATA TREE " + index
        );
        System.out.println();
        
        // Creating corrector instance
        Corrector ci =
            Corrector.createCorrectorInstance(
                tree, grammar,
                strategy, approach, signature,
                false, false
            );
        
        // Accessing computed repair
        Repair treeRepair = ci.getTreeRepair();
        
        // Printing edit sequences
        //treeRepair.printFixesDump();
        
        // Printing repair description
        treeRepair.printDebugDump();
        
        System.out.println();
        System.out.println();
        
    }
    
    
    // -------------------------------------------------------------------------

}
