
package cz.cuni.mff.corrector.graph;

import cz.cuni.mff.corrector.grammar.Grammar;
import cz.cuni.mff.corrector.instruction.InstructionRemoveLeaf;
import cz.cuni.mff.corrector.intent.Intent;
import cz.cuni.mff.corrector.intent.IntentCorrect;
import cz.cuni.mff.corrector.intent.IntentDelete;
import cz.cuni.mff.corrector.repair.NodeRepair;
import cz.cuni.mff.corrector.repair.NodeRepairDeleteSubtree;
import cz.cuni.mff.corrector.repair.Repair;
import cz.cuni.mff.corrector.repair.SequenceRepair;
import cz.cuni.mff.corrector.tree.NodeElement;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import static org.junit.Assert.*;
import org.junit.Test;

/**
 * Basic tests for graph structures.
 * 
 * @author Martin Svoboda
 */
public class GraphTest {
    
    // -------------------------------------------------------------------------


    /**
     * Tests the bahaviour of graph construction methods.
     */
    @Test
    public void testGraphConstruction() {
        
        System.out.println("testGraphConstruction()\n");
        
        /* Graph structure:
         *   0A -> 1A -> 2A -> 3A
         *                  -> 3B
         *                  -> 3B
         *                  -> 3C
         *      -> 1B -> 2B -> 3B
         */
        
        Set<String> targetStates = new HashSet<String>();
        targetStates.add("3A");
        targetStates.add("3B");
        int strataNumber = 3;
        
        Graph g = new Graph(targetStates, strataNumber);
        
        Vertex v0a = new Vertex(0, "0A");
        g.addNewVertex(v0a);
        assertSame(v0a, g.getVertexItem(0, "0A"));
        assertEquals(1, g.getVerticesCount());
        
        Vertex v1a = new Vertex(1, "1A");
        g.addNewVertex(v1a);
        assertSame(v1a, g.getVertexItem(1, "1A"));
        assertEquals(2, g.getVerticesCount());
        
        Intent i = new IntentDelete(null, v0a, v1a, null, null);
        
        Edge e0a1a = new Edge(v0a, v1a, i);
        g.addNewEdge(e0a1a);
        Set<Edge> oe0a = g.getOutgoingEdges(v0a);
        assertEquals(1, oe0a.size());
        assertTrue(oe0a.contains(e0a1a));
        Set<Edge> ie1a = g.getIngoingEdges(v1a);
        assertEquals(1, ie1a.size());
        assertTrue(ie1a.contains(e0a1a));
        assertEquals(1, g.getEdgesCount());
        
        Vertex v2a = new Vertex(2, "2A");
        g.addNewVertex(v2a);
        assertSame(v2a, g.getVertexItem(2, "2A"));
        assertEquals(3, g.getVerticesCount());
        
        Edge e1a2a = new Edge(v1a, v2a, i);
        g.addNewEdge(e1a2a);
        assertEquals(2, g.getEdgesCount());
        
        Vertex v3a = new Vertex(3, "3A");
        g.addNewVertex(v3a);
        assertSame(v3a, g.getVertexItem(3, "3A"));
        assertEquals(4, g.getVerticesCount());
        Edge e2a3a = new Edge(v2a, v3a, i);
        g.addNewEdge(e2a3a);
        assertEquals(3, g.getEdgesCount());
        
        Grammar grammar = Grammar.createFromDataString("Root\nRoot root #empty");
        IntentDelete i2a3b1 =
            new IntentDelete(new NodeElement("X"), new Vertex(2, "2A"), new Vertex(3, "2A"), null, grammar.getEmptyContext());
        IntentDelete i2a3b2 =
            new IntentDelete(new NodeElement("Y"), new Vertex(2, "2A"), new Vertex(3, "2A"), null, grammar.getEmptyContext());
        
        Vertex v3b = new Vertex(3, "3B");
        g.addNewVertex(v3b);
        assertSame(v3b, g.getVertexItem(3, "3B"));
        assertEquals(5, g.getVerticesCount());
        Edge e2a3b1 = new Edge(v2a, v3b, i2a3b1);
        g.addNewEdge(e2a3b1);
        assertEquals(4, g.getEdgesCount());
        Edge e2a3b2 = new Edge(v2a, v3b, i2a3b2);
        g.addNewEdge(e2a3b2);
        assertEquals(5, g.getEdgesCount());
        
        Vertex v3c = new Vertex(3, "3C");
        g.addNewVertex(v3c);
        assertSame(v3c, g.getVertexItem(3, "3C"));
        assertEquals(6, g.getVerticesCount());
        Edge e2a3c = new Edge(v2a, v3c, i);
        g.addNewEdge(e2a3c);
        assertEquals(6, g.getEdgesCount());
        
        Set<Edge> oe2a = g.getOutgoingEdges(v2a);
        assertEquals(4, oe2a.size());
        assertTrue(oe2a.contains(e2a3a));
        assertTrue(oe2a.contains(e2a3b1));
        assertTrue(oe2a.contains(e2a3b2));
        assertTrue(oe2a.contains(e2a3c));
        
        Vertex v1b = new Vertex(1, "1B");
        g.addNewVertex(v1b);
        Edge e0a1b = new Edge(v0a, v1b, i);
        g.addNewEdge(e0a1b);
        
        Vertex v2b = new Vertex(2, "2B");
        g.addNewVertex(v2b);
        Edge e1b2b = new Edge(v1b, v2b, i);
        g.addNewEdge(e1b2b);
        
        Edge e2b3b = new Edge(v2b, v3b, i);
        g.addNewEdge(e2b3b);

        Set<Vertex> gv = g.getGraphVertices();
        assertEquals(8, gv.size());
        Set<Edge> ge = g.getGraphEdges();
        assertEquals(9, ge.size());
        
        Set<String> gt = g.getTargetStates();
        assertEquals(2, gt.size());
        assertTrue(gt.contains("3A"));
        assertTrue(gt.contains("3B"));
        Vertex gtva = g.getTargetVertex("3A");
        assertEquals(3, gtva.getStratumNumber());
        assertEquals("3A", gtva.getAutomatonState());
        
        g.printDebugDump("");
        System.out.println();
        
    }
    
    
    /**
     * Tests the capabilities of managing shortest paths.
     * @see #generateSampleGraph()
     */
    @Test
    public void testShortestPaths() {
        
        System.out.println("testShortestPaths()\n");
        
        Graph g = generateSampleGraph();
        
        Vertex v0a = g.getVertexItem(0, "0A");
        assertNotNull(g.getTracingCost(v0a));
        assertEquals(0, (int)g.getTracingCost(v0a));
        assertNotNull(g.getTracingEdges(v0a));
        assertEquals(0, g.getTracingEdges(v0a).size());
        
        Vertex v1a = g.getVertexItem(1, "1A");
        Edge e0a1a1 = g.getIngoingEdges(v1a).iterator().next();
        assertEquals(2, (int)g.getTracingCost(v1a));
        assertEquals(1, g.getTracingEdges(v1a).size());
        assertTrue(g.getTracingEdges(v1a).contains(e0a1a1));
        
        Vertex v3b = g.getVertexItem(3, "3B");
        Set<Edge> ie3b = g.getIngoingEdges(v3b);
        assertEquals(3, (int)g.getTracingCost(v3b));
        assertEquals(3, g.getTracingEdges(v3b).size());
        assertTrue(g.getTracingEdges(v3b).containsAll(ie3b));
        
        assertEquals(3, g.getTermCost());
        
        List<List<Edge>> gsp = g.fetchShortestPaths();
        assertEquals(3, gsp.size());
        for (List<Edge> list : gsp) {
            assertEquals(3, list.size());
            assertEquals(0, list.get(0).getSourceStratum());
            assertEquals("0A", list.get(0).getSourceState());
            assertEquals(3, list.get(2).getTargetStratum());
            assertEquals("3B", list.get(2).getTargetState());
        }

        g.printDebugDump("");
        System.out.println();
        
    }
    
    
    // -------------------------------------------------------------------------
    
    
    /**
     * Generates a sample testing graph.
     * 
     * Graph structure:
     *   0A -2-> 1A -0-> 2A -5-> 3A
     *                      -1-> 3B
     *                      -1-> 3B
     *                      -1-> 3C
     *      -0-> 1B -1-> 2B -2-> 3B
     * Final vertices
     *   3A, 3B
     * Terminal cost:
     *   3
     * Shortest paths:
     *   0A -> 1A -> 2A -> 3B
     *   0A -> 1A -> 2A -> 3B
     *   0A -> 1B -> 2B -> 3B
     */
    public static Graph generateSampleGraph() {
        
        Set<String> targetStates = new HashSet<String>();
        targetStates.add("3A");
        targetStates.add("3B");
        int strataNumber = 3;
        
        Graph g = new Graph(targetStates, strataNumber);
        
        Grammar grammar = Grammar.createFromDataString("Root\nRoot root #empty");
        
        NodeElement nex = new NodeElement("X");
        InstructionRemoveLeaf irlx = new InstructionRemoveLeaf();
        NodeRepair nrx = new NodeRepairDeleteSubtree(irlx);
        SequenceRepair srax = new SequenceRepair(g, Intent.Type.DELETE);
        
        IntentDelete ix =
            new IntentDelete(nex, new Vertex(2, "2A"), new Vertex(3, "2A"), null, grammar.getEmptyContext());
        IntentDelete i2a3b1 =
            new IntentDelete(nex, new Vertex(2, "2A"), new Vertex(3, "2A"), null, grammar.getEmptyContext());
        IntentDelete i2a3b2 =
            new IntentDelete(nex, new Vertex(2, "2A"), new Vertex(3, "2A"), null, grammar.getEmptyContext());
        
        Repair r0 = new Repair(null, nrx, srax) {
            @Override
            public int getTotalCost() { return 0; }
        };
        Repair r1 = new Repair(null, nrx, srax) {
            @Override
            public int getTotalCost() { return 1; }
        };
        Repair r2 = new Repair(null, nrx, srax) {
            @Override
            public int getTotalCost() { return 2; }
        };
        Repair r5 = new Repair(null, nrx, srax) {
            @Override
            public int getTotalCost() { return 5; }
        };
        
        Vertex v0a = new Vertex(0, "0A");
        g.addNewVertex(v0a);
        Vertex v1a = new Vertex(1, "1A");
        g.addNewVertex(v1a);
        Vertex v1b = new Vertex(1, "1B");
        g.addNewVertex(v1b);
        Vertex v2a = new Vertex(2, "2A");
        g.addNewVertex(v2a);
        Vertex v2b = new Vertex(2, "2B");
        g.addNewVertex(v2b);
        Vertex v3a = new Vertex(3, "3A");
        g.addNewVertex(v3a);
        Vertex v3b = new Vertex(3, "3B");
        g.addNewVertex(v3b);
        Vertex v3c = new Vertex(3, "3C");
        g.addNewVertex(v3c);
        
        Edge e0a1a1 = new Edge(v0a, v1a, ix);
        g.addNewEdge(e0a1a1);
        Edge e0a1b1 = new Edge(v0a, v1b, ix);
        g.addNewEdge(e0a1b1);
        Edge e1a2a1 = new Edge(v1a, v2a, ix);
        g.addNewEdge(e1a2a1);
        Edge e1b2b1 = new Edge(v1b, v2b, ix);
        g.addNewEdge(e1b2b1);
        Edge e2a3a1 = new Edge(v2a, v3a, ix);
        g.addNewEdge(e2a3a1);
        Edge e2a3b1 = new Edge(v2a, v3b, i2a3b1);
        g.addNewEdge(e2a3b1);
        Edge e2a3b2 = new Edge(v2a, v3b, i2a3b2);
        g.addNewEdge(e2a3b2);
        Edge e2a3c1 = new Edge(v2a, v3c, ix);
        g.addNewEdge(e2a3c1);
        Edge e2b3b1 = new Edge(v2b, v3b, ix);
        g.addNewEdge(e2b3b1);
        
        e0a1a1.setEdgeRepair(r2);
        e0a1b1.setEdgeRepair(r0);
        e1a2a1.setEdgeRepair(r0);
        e1b2b1.setEdgeRepair(r1);
        e2a3a1.setEdgeRepair(r5);
        e2a3b1.setEdgeRepair(r1);
        e2a3b2.setEdgeRepair(r1);
        e2a3c1.setEdgeRepair(r1);
        e2b3b1.setEdgeRepair(r2);
        
        Set<Edge> pe0a = new HashSet<Edge>();
        g.setTracingCost(v0a, 0);
        g.setTracingEdges(v0a, pe0a);
        
        Set<Edge> pe1a = new HashSet<Edge>();
        pe1a.add(e0a1a1);
        g.setTracingCost(v1a, e0a1a1.getEdgeCost());
        g.setTracingEdges(v1a, pe1a);
        
        Set<Edge> pe1b = new HashSet<Edge>();
        pe1b.add(e0a1b1);
        g.setTracingCost(v1b, e0a1b1.getEdgeCost());
        g.setTracingEdges(v1b, pe1b);
        
        Set<Edge> pe2a = new HashSet<Edge>();
        pe2a.add(e1a2a1);
        g.setTracingCost(v2a, g.getTracingCost(v1a) + e1a2a1.getEdgeCost());
        g.setTracingEdges(v2a, pe2a);
        
        Set<Edge> pe2b = new HashSet<Edge>();
        pe2b.add(e1b2b1);
        g.setTracingCost(v2b, g.getTracingCost(v1b) + e1b2b1.getEdgeCost());
        g.setTracingEdges(v2b, pe2b);
        
        Set<Edge> pe3a = new HashSet<Edge>();
        pe3a.add(e2a3a1);
        g.setTracingCost(v3a, g.getTracingCost(v2a) + e2a3a1.getEdgeCost());
        g.setTracingEdges(v3a, pe3a);
        
        Set<Edge> pe3b = new HashSet<Edge>();
        pe3b.add(e2a3b1);
        pe3b.add(e2a3b2);
        pe3b.add(e2b3b1);
        g.setTracingCost(v3b, g.getTracingCost(v2a) + 1);
        g.setTracingEdges(v3b, pe3b);
        
        Set<Edge> pe3c = new HashSet<Edge>();
        pe3c.add(e2a3c1);
        g.setTracingCost(v3c, g.getTracingCost(v2a) + e2a3c1.getEdgeCost());
        g.setTracingEdges(v3c, pe3c);
        
        g.setTermCost(3);
        assertEquals(3, g.getTermCost());

        return g;
        
    }
    
    
    // -------------------------------------------------------------------------
    
}
